/*
  Reads graphs as list of adjacency lists ("&" is comment) format.
  This works, but could use some polishing.
*/

/* ----------------------------------------------------------------- */

#pragma once

#include <vector>
#include <iostream>
#include <cstdio>
#include <cstring>
#include <cstdlib>

#include "utils/string_utils.hpp"
#include "utils/exceptions.hpp"
#include "utils/string_utils.hpp"
#include "utils/instance.hpp"

namespace io {

  // read the graph g from the instream "in"
  Instance* read_adj_list(std::istream& in, const unsigned num_paths, const unsigned num_cycles, const unsigned num_objects)
  {
    DEBUG1(std::cout << "reading adjacency list format"<< std::endl);
    unsigned num_vertices = 0;
    unsigned no_lines = 0;
    std::string in_line; // buffer for the next line
    Instance* I = new Instance(num_paths, num_cycles, num_objects);

    std::vector<ScafVertex> verts;

    try{
      while(std::getline(in, in_line)) {
        ++no_lines;
        
        if(num_vertices == 0){
          // read number of vertices
          if(std::sscanf(in_line.c_str(), "%u", &num_vertices) != 1) throw except::bad_syntax(no_lines, "failed to read #vertices");
          if ( ( num_vertices == 0 ) ) throw except::bad_syntax(no_lines, "no vertices");
          verts.reserve(num_vertices);
          for(unsigned vi = 0; vi < num_vertices; ++vi) verts.push_back(I->add_vertex(std::to_string(vi), vi));
          // if v is even, then add a matching edge to v+1
          //for(unsigned vi = 0; vi < num_vertices; vi+=2)
          //  I->add_matching_edge(verts[vi], verts[vi+1]);
        } else {
          switch (in_line[0]) {
            case '&':                  /* skip lines with comments */
            case '\n':                 /* skip empty lines   */
            case '\0':                 /* skip empty lines at the end of file */
              break;
            default:  // format: <vertex id>[ \t]*<space separated adj list of <vertex>,<weight>>
              try{
                unsigned u, weight;
                const unsigned v = read_single_number(in_line);
                while(!in_line.empty()) {
                  // skip all delimeters
                  skip_to(in_line, "0123456789");
                  // if we read everything we could, we're good for the next line, otherwise, skip to the next non-delimeter
                  if(in_line.empty()) break;
                  // read the next vertex
                  u = read_single_number(in_line);
                  // check vertex bounds
                  if(u < num_vertices){
                    // if next is a comma, read the edge weight
                    if(in_line[0] == ','){
                      // skip to the next weight
                      skip_to(in_line, "0123456789");
                      if(in_line.empty()) throw except::bad_syntax(no_lines, "indicated weight not found");
                      // read the weight
                      weight = read_single_number(in_line);
                    } else weight = 0; // per default the weight of an edge is 0
                  } else throw except::bad_syntax(no_lines, "vertex number out of bounds");
                  I->add_edge(verts[u], verts[v], ScafEdgeProperty(weight, 0, 1, ""));
                } // while
              } catch(std::invalid_argument& ex){
                // first characted has to be a digit to read a vertex
                throw except::bad_syntax(no_lines, "unexpected format, expected vertex");
              }
          } // switch
        } // if
      } // while
      if ( in.eof() == 0 ) throw except::bad_syntax(no_lines, "expected EOF");
      if ( no_lines == 0 ) throw except::read_error(no_lines, "input stream is empty");
    } catch(except::bad_syntax& ex){
      std::cout << "Syntax error in line "<<ex.line_no<<": "<<ex.what()<<" - file not in adjacency list format!"<<std::endl;
      delete I;
      return NULL;
    } catch(except::read_error& ex){
      std::cout << "Error reading after "<<ex.line_no<<" lines: "<<ex.what()<<" - file not in adjacency list format!"<<std::endl;
      delete I;
      return NULL;
    }
#warning TODO: support insert size and other information in DIMACS format
    /* Thank God! all is done */
    return I;
  } // function


  // write the graph g to the outstream "out"
  void write_adj_list_graph(std::ostream& out, const Instance& I){
    // write # vertices
    const RawScaffoldGraph& g = I.get_graph();
    out << boost::num_vertices(g) << std::endl << '&' << std::endl;

    // adjacency lists. To this end, iterate over all vertices & print their adjacent vertices one by one
    for(ScafVIterRange ur = boost::vertices(g); ur.first != ur.second; ++ur.first){
      const ScafVertex& u = *ur.first;
      // output vertex name + some spaces
      out << g[u].name << '\t';
      // for each incident edge output the target and its weight if the index of the target is larger than our own
      for(ScafOEdgeIterRange er = boost::out_edges(u, g); er.first != er.second; ++er.first){
        // only output edges from smaller to larger index
        const ScafVertex& v = boost::target(*er.first, g); 
        if(g[u].index < g[v].index)
          out << g[v].name << ',' << g[*er.first].weight << ' ';
      } // for
      // close the line
      out << std::endl;
    }
  } // function
} // namespace

