/*
  Reads coloring problem in DIMACS format.
  This works, but could use some polishing.
*/

/* ----------------------------------------------------------------- */

#pragma once

#include <vector>
#include <iostream>
#include <cstdio>
#include <cstring>
#include <cstdlib>


#include <boost/graph/graph_traits.hpp>

#include "utils/exceptions.hpp"
#include "utils/instance.hpp"

namespace io {

  // read the graph g from the instream "in"
  Instance* read_dimacs_graph(std::istream& in, const unsigned num_paths, const unsigned num_cycles, const unsigned num_objects)
  {
    DEBUG1(std::cout << "reading DIMACS format"<<std::endl);
    const uint P_FIELDS = 2; // number of fields expected in a "p" line
    const uint E_FIELDS = 2; // number of fields expected in an "e" line
    unsigned no_lines = 0; // number of lines read
    unsigned no_elines = 0; // number of lines defining edges
    unsigned num_vertices, num_edges; // number of vertices & edges in the graph
    std::string in_line; // buffer for the next line
    Instance* I = new Instance(num_paths, num_cycles, num_objects);

    std::vector<ScafVertex> verts;

    try{
      while(std::getline(in, in_line)) {
        ++no_lines;
        switch (in_line[0]) {
          case 'c':                  /* skip lines with comments */
          case '\n':                 /* skip empty lines   */
          case '\0':                 /* skip empty lines at the end of file */
            break;
          case 'p':   /* contains "edge NUMBER_NODES NUMBER_EDGES */
            if(std::sscanf(in_line.c_str(), "%*c %*s %u %u", &num_vertices, &num_edges) != P_FIELDS) 
              throw except::bad_syntax(no_lines, "unexpected p-line format");
            verts.reserve(num_vertices);
            for(unsigned vi = 0; vi < num_vertices; ++vi) verts.push_back(I->add_vertex(std::to_string(vi), vi));
            break;
          case 'e':                    /* edge description */
            unsigned u,v; // store the indices of the incident vertices
            if(no_elines >= num_edges) throw except::bad_syntax(no_lines,"too many edges specified or no p line found");
            if(std::sscanf(in_line.c_str(),"%*c %u %u", &u, &v) != E_FIELDS) 
              throw except::bad_syntax(no_lines, "unexpected e-line format");
            --u; // index from 0, not 1
            --v;
            if((u < num_vertices) && (v < num_vertices)){
              ScafEdge e;
              bool success;
              boost::tie(e, success) = I->add_edge(verts[u], verts[v]);
              assert(success);
            } else throw except::bad_syntax(no_lines, "vertex index out of bounds");
            ++no_elines;
            break;
          default:  /* unknown type of line */
            throw except::bad_syntax(no_lines, "unrecognized line type");
        } // switch
      } // while
      if ( in.eof() == 0 ) throw except::bad_syntax(no_lines, "read error"); 
      if ( no_lines == 0 ) throw except::read_error(no_lines, "input stream is empty");
      if ( no_elines < num_edges ) throw except::bad_syntax(no_lines, "did not read enough edges");
    } catch(except::bad_syntax& ex){
      std::cout << "Syntax error in line "<<ex.line_no<<": "<<ex.what()<< " - file not in DIMACS format"<<std::endl;
      delete I;
      return NULL;
    } catch(except::read_error& ex){
      std::cout << "Error reading after "<<ex.line_no<<" lines: "<<ex.what()<< " - file not in DIMACS format"<<std::endl;
      delete I;
      return NULL;
    }
#warning TODO: support insert size and other information in DIMACS format
    /* Thank God! all is done */
    return I;
  } // function


  // write the graph g to the outstream "out"
  void write_dimacs_graph(std::ostream& out, const Instance& I){
    // write p line
    const RawScaffoldGraph& g = I.get_graph();
    out << "p edge "<<boost::num_vertices(g)<<" "<<boost::num_edges(g)<<std::endl;

    // write e lines. To this end, iterate over all vertices & print their adjacent vertices one by one
    for(ScafVIterRange ur = boost::vertices(g); ur.first != ur.second; ++ur.first){
      const ScafVertex& u = *ur.first;
      const size_t u_idx = g[u].index;
      for(ScafAdjIterRange vr = boost::adjacent_vertices(u, g); vr.first != vr.second; ++vr.first){
        // only output edges from smaller to larger index
        const ScafVertex& v = *vr.first;
        const size_t v_idx = g[v].index;
        if(v_idx < u_idx){
          // remember to shift the vertex index by one, since boost starts with 0 and DIMACS starts with 1
          out << "e "<<v_idx + 1<<" "<<u_idx+1<<std::endl;
        }
      } // for
    }
  } // function
} // namespace
