
#include<iostream>
#include<fstream>
#include <boost/graph/graph_traits.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/graphviz.hpp>
#include <boost/program_options/variables_map.hpp>
#include <boost/program_options/parsers.hpp>

#include "utils/graph_typedefs.hpp"
#include "utils/graph_utils.hpp"
#include "utils/graph_parameters.hpp"
#include "utils/scaffolding_typedefs.hpp"
#include "utils/instance.hpp"
#include "solv/tw_dp.hpp"

#include "io/scaffold_io.hpp"
#include "io/dimacs.hpp"
#include "io/dot.hpp"

using namespace std;
using namespace boost;
namespace po = boost::program_options;

using namespace scaffold;
using namespace preprocess;



struct{
  string filename;
  bool output_graphviz;
  bool output_dimacs;
  size_t paths;
  size_t cycles;
} options;

void handle_options(const int argc, const char** argv){
  // Declare the supported options.
  po::options_description desc("Allowed options");
  desc.add_options()
      ("help", "produce help message")
      ("fn", po::value<string>(), "set the filename to read input from")
      ("gv", "output to graphviz format")
      ("di", "output to dimacs coloring format")
      ("p", po::value<size_t>(), "allowed number of paths (default 6)")
      ("c", po::value<size_t>(), "allowed number of cycles (default 3)")
  ;

  po::variables_map vm;
  po::store(po::parse_command_line(argc, argv, desc), vm);
  po::notify(vm);

  if(vm.count("help") || !vm.count("fn")) { cout << desc << "\n"; exit(EXIT_SUCCESS); }
  if(vm.count("p")) options.paths = vm["p"].as<size_t>(); else options.paths = 6;
  if(vm.count("c")) options.cycles = vm["c"].as<size_t>(); else options.cycles = 3;
  options.filename = vm["fn"].as<string>();
  options.output_graphviz = vm.count("gv");
  options.output_dimacs = vm.count("di");
}



int main(const int argc, const char** argv){
  // parse arguments
  auto opt_desc = initialize_options(argv[0]);
  const GeneralOptions options = handle_options(parse_arguments(argc, argv, opt_desc), opt_desc);

  Instance* I_ptr = io::read_instance(options);
  if(I_ptr == NULL) {
    std::cout << "couldn't read the input file"<<std::endl;
    exit(EXIT_FAILURE);
  }
  Instance& I(*I_ptr);
  Instance copy_I(I);

  DEBUG1(
    std::cout << "total graph with matching: "<<std::endl;
    write_adj_list_graph(std::cout, *I.g);
    std::cout << std::endl;
  );

  // testing area
  DEBUG1(std::cout << "begin solving process"<<std::endl);
  scaffold::solv::DP::DynProgSolution<std::greater<size_t>> sol(I.num_paths, I.num_cycles);
  scaffold::solv::DP::TW_solv(I, sol);
  if(sol.no_solutions())
    std::cout << "no solution"<<std::endl;
  else {
    std::cout << "solution matrix:"<<std::endl << sol << std::endl;

    const EdgeNameList& sol_edges = sol.get_min_entry().get_edges();

    std::cout << "got solution edges: ";
    for(const auto& e: sol_edges) std::cout<<"("<<e.first<<","<<e.second<<") ";
    std::cout<<std::endl;
    copy_I.induced_subgraph(sol_edges);

    std::cout << "optimal solution weight: "<<copy_I.weight()<<'\n';

    // finally, output the graph in the requested form
    std::cout << "writing graphviz file "<<options.out_graph_filename<<std::endl;
    std::ofstream os(options.out_graph_filename);
    io::write_scaffold(os, copy_I);
    os.close();
  }

  exit(EXIT_SUCCESS);
}




