#include<iostream>
#include<fstream>
#include<unordered_set>
#include <boost/graph/graph_traits.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/graphviz.hpp>
#include <boost/program_options/variables_map.hpp>
#include <boost/program_options/parsers.hpp>

#include "utils/graph_typedefs.hpp"
#include "utils/graph_utils.hpp"
#include "utils/graph_parameters.hpp"
#include "utils/scaffolding_typedefs.hpp"
#include "graph_io/read_dimacs_graph.hpp"
#include "graph_io/read_adj_list.hpp"
#include "graph_io/write_adj_list.hpp"
#include "graph_io/write_dimacs_graph.hpp"
#include "graph_io/scaffold_io.hpp"
#include "preprocess/pp_pendant_match.hpp"
#include "solv/degree_branching.hpp"
#include "solv/FES_branching.hpp"

#include "utils/instance.hpp"

using namespace std;
using namespace boost;
namespace po = boost::program_options;

using namespace scaffold;
using namespace preprocess;


typedef adjacency_list<hash_setS,
        listS,
        undirectedS,
    property<vertex_index_t, size_t,
      property<vertex_name_t, size_t>
    >,
    property <edge_weight_t, size_t>
  > Graph;



struct{
  string filename;
  bool output_graphviz;
  bool output_dimacs;
  size_t paths;
  size_t cycles;
} options;

void handle_options(const int argc, const char** argv){
  // Declare the supported options.
  po::options_description desc("Allowed options");
  desc.add_options()
      ("help", "produce help message")
      ("fn", po::value<string>(), "set the filename to read input from")
      ("gv", "output to graphviz format")
      ("di", "output to dimacs coloring format")
      ("p", po::value<size_t>(), "allowed number of paths (default 6)")
      ("c", po::value<size_t>(), "allowed number of cycles (default 3)")
  ;

  po::variables_map vm;
  po::store(po::parse_command_line(argc, argv, desc), vm);
  po::notify(vm);

  if(vm.count("help") || !vm.count("fn")) { cout << desc << "\n"; exit(EXIT_SUCCESS); }
  if(vm.count("p")) options.paths = vm["p"].as<size_t>(); else options.paths = 6;
  if(vm.count("c")) options.cycles = vm["c"].as<size_t>(); else options.cycles = 3;
  options.filename = vm["fn"].as<string>();
  options.output_graphviz = vm.count("gv");
  options.output_dimacs = vm.count("di");
}

int main(const int argc, const char** argv){
  // parse arguments
  handle_options(argc, argv);

  // read the input
  std::ifstream infile(options.filename);
  Instance<Scaffold_Graph> *I_ptr = read_instance(infile, options.paths, options.cycles);
  if(I_ptr == NULL) {
    std::cout << "couldn't read the input file"<<std::endl;
    exit(EXIT_FAILURE);
  } else std::cout << "read instance"<<std::endl;
  Instance<Scaffold_Graph>& I(*I_ptr);
  Instance<Scaffold_Graph> copy_I(I);

  std::cout << "total graph with matching: "<<std::endl;
  write_adj_list_graph(std::cout, *I.g);
  std::cout << std::endl;
  // testing area

  std::cout << "begin solving process"<<std::endl;
  Solution<>* sol = new Solution<>(I);
  scaffold::solv::FES_solve_recursively(I, sol);
  if(sol->no_solutions())
    std::cout << "no solution"<<std::endl;
  else {
    std::cout << "solution matrix:"<<std::endl << *sol << std::endl;

    const std::list<EdgeName>& sol_edges = sol->get_min_entry().get_edges();

    std::cout << "got solution edges: ";
    for(const auto& e : sol_edges) std::cout<<"("<<e.first<<","<<e.second<<") ";
    std::cout<<std::endl;
    for(const auto& e : sol_edges) copy_I.delete_edge_by_name(e);

    std::cout << "graph with optimal solution applied (weight "<<copy_I.weight()<<"):"<<std::endl;
    write_adj_list_graph(std::cout, *copy_I.g);
  }

  delete sol;

  // finally, output the graph in the requested form
  if(options.output_graphviz) {
    std::cout << "writing graphviz file "<<options.filename<<".dot"<<std::endl;
    std::ofstream os(options.filename+".dot");
    write_scaffold_as_dot(os, copy_I, std::list<contig_jump<Scaffold_Graph> >());
    os.close();
  }
  if(options.output_dimacs){
    std::cout << "writing DIMACS file "<<options.filename<<".dmc"<<std::endl;
    std::ofstream os(options.filename+".dmc");
    write_dimacs_graph(os, *copy_I.g);
    os.close();
  }

  exit(EXIT_SUCCESS);
}




