#include<iostream>
#include<fstream>

#include "utils/command_line.hpp"
#include "utils/instance.hpp"
#include "utils/profiling.hpp"
#include "io/scaffold_io.hpp"
#include "io/fasta.hpp"
#include "preprocess/pp_pendant_match.hpp"
#include "preprocess/jump_preprocess.hpp"
#include "preprocess/negative_length.hpp"
#include "preprocess/find_missing.hpp"
#include "preprocess/linearization.hpp"

#ifdef CPLEX

#include "utils/ilp_contig_jumps.hpp"
#include "solv/ilp_model_multi.hpp"

// those use different models
//#include "solv/ilp_model_onegraph.hpp"
//#include "solv/ilp_model_3layers.hpp"
//#include "solv/ilp_model_multi_manygraph.hpp"

#endif

#include "utils/lin_options.hpp"


using namespace io;
using namespace scaffold;

#ifdef CPLEX
using namespace scaffold::solv;
#endif

#ifdef STATISTICS
JumpStatistics jump_stats;
#endif

typedef struct {
  GeneralOptions general;
  AlignmentOptions alignment;
#ifdef CPLEX
  ILPOptions ilp;
#endif
  LinearizationOptions lin;

} Options;


SequenceMap *read_sequences(const std::string &filename)
{
  if(filename != "") {
    std::ifstream is(filename);
    SequenceMap *sequences = new SequenceMap();
    read_fasta_file(is, *sequences);
    return sequences;
  } else return nullptr;
}

SequenceMap *get_sequences(const std::string &in_seq_filename)
{
  if(in_seq_filename != "") {
    timer read_fa_timer(true);
    SequenceMap *sequences = read_sequences(in_seq_filename);
    read_fa_timer.stop();

    if(sequences == nullptr) {
      std::cout << "couldn't read the sequences from file " << in_seq_filename << std::endl;
      exit(EXIT_FAILURE);
    } else {
      std::cout << "read sequences (took " << read_fa_timer.seconds_passed() << " seconds)" << std::endl;
#ifdef STATISTICS
      SequenceStatistics in_stat(*sequences);
      in_stat.print();
#endif
      return sequences;
    }
  } else return nullptr;
}

Instance *get_instance(const Options &options, SequenceMap *sequences)
{
  timer read_timer(true);
  std::ifstream infile(options.general.in_graph_filename);
  Instance *I_ptr = read_instance(infile, options.general, sequences);
  read_timer.stop();

  if(I_ptr == nullptr) {
    std::cout << "couldn't read the input file " << options.general.in_graph_filename << std::endl;
    exit(EXIT_FAILURE);
  } else std::cout << "read instance (took " << read_timer.seconds_passed() << " seconds)" << std::endl;
  return I_ptr;
}

#ifdef CPLEX

void preprocessing(const Options &options,
                   Instance &I,
                   SequenceMap *sequences,
                   ScafVPairDependencyMap &dependencies)
{
  timer preprocess_timer(true);

  // Step 1: data cleanup
  // Step 1a: cut off edges by threshold
  if(options.general.cut_off) {
    std::cout << "cutting off edges below weight " << options.general.cut_off << "..." << std::endl;
    scaffold::cut_off_threshold(I, options.general.cut_off);
  }

  std::cout << "fixing non-contig multiplicities..." << std::endl;
  if(!options.ilp.ignore_multiplicities)
    fix_non_matching_multiplicities(I);
  else
    remove_multiplicities(I);

  // Step 1b: ignore edge-lengths that are clearly due to misalignment
  std::cout << "fixing implausible lengths..." << std::endl;
  if(sequences)
    fix_negative_non_contig_lengths(I, *sequences, options.alignment, false);
  else
    fix_implausible_lengths(I, options.alignment);

  if(options.general.find_missed_pairs) {
    std::cout << "overlapping contigs to find missed edges..." << std::endl;
    find_missing_links(I, options.alignment, *sequences);
  }
  if(!options.ilp.ignore_contig_jumps) {
    std::cout << "preparing contig jumps..." << std::endl;
#ifdef STATISTICS
    preprocess_jumps(I, options.alignment, dependencies, jump_stats, sequences);
#else
    preprocess_jumps(I, options.alignment, dependencies, sequences);
#endif
  }

  // Step 4: scale weights by multiplicities if wanted
  if(!options.ilp.multi_take_all) {
    std::cout << "scaling weights..." << std::endl;
    scale_weights(I);
  }

  preprocess_timer.stop();
  std::cout << "preprocessing took " << preprocess_timer.seconds_passed() << " seconds" << std::endl;
}

IloCplex *build_model(const Options &options,
                      ScafVPairDependencyMap &dependencies,
                      IloEnv &env,
                      IloModel &model,
                      VAR_COLLECTION &vars,
                      CallbackStatistics &cb_stats)
{
  std::cout << "building CPLEX model..." << std::endl;
  // initialize the ILP model
  timer model_timer(true);

#ifdef STATISTICS
  solv::populate_model(model, env, vars, jump_stats, options.ilp);
#else
  solv::populate_model(model, env, vars, options.ilp);
#endif
  if(!options.ilp.ignore_contig_jumps)
    vars.respect_jump_dependencies(model, env, dependencies);

  // initialize the CPLEX solver with the ILP model
  IloCplex *cplex = new IloCplex(model);

  // set the clock type as to use CPU seconds
  cplex->setParam(IloCplex::ClockType, 1);

  // add the callback to remove cycles in the layer-formulation
  cplex->use(solv::GetLazyCyclesBlock(env, vars, cb_stats));
  cplex->use(GetRelaxCallback(env, cb_stats));

  if(options.general.out_model_filename != "")
    cplex->exportModel(options.general.out_model_filename.c_str());

  model_timer.stop();
  std::cout << "CPU-seconds spent modelling: " << model_timer.seconds_passed() << std::endl;

  return cplex;
}

void solve(IloCplex &cplex)
{
  // Optimize the problem and obtain solution.
  timer opt_timer(true);
  try {
    if(!cplex.solve()) {
      std::cout << "Failed to optimize LP" << std::endl;
      exit(EXIT_FAILURE);
    }
  } catch(IloCplex::Exception& e) {
    std::cout << "caught exception: " << e.getMessage() << std::endl;
    exit(EXIT_FAILURE);
  }
  opt_timer.stop();
  std::cout << "CPU-seconds spent optimizing: " << opt_timer.seconds_passed() << std::endl;
}

void apply_solution(const IloCplex &cplex, Instance &I, const VAR_COLLECTION &vars)
{
  DEBUG1(std::cout << "solution edges: " << std::endl);

  // decide which variable map to use to look up whether an edge is in the solution or not
  std::stack<Edge<RawScaffoldGraph> > to_delete;
  for(auto e_it = I.get_edges(); e_it; ++e_it) {
    const ScafEdge &e = *e_it;
    const ScafVertex &u = I.source(e);
    const ScafVertex &v = I.target(e);
    const ScafVertexPair uv(u, v);
    const ScafVertexPair vu(v, u);

    // find how often uv & vu were being used
    auto uv_it = vars.times_used_undir.find(uv);
    if(uv_it == vars.times_used_undir.cend()) uv_it = vars.times_used_undir.find(vu);
    assert(uv_it != vars.times_used_undir.cend());
    const unsigned used = lround(cplex.getValue(uv_it->second))
                          - (vars.num_jumped(uv, cplex) + vars.num_jumped(vu, cplex));

    // remove the number of times uv & vu have been used as contig jumps

    if(used) {
      DEBUG1(
        std::cout << I.get_edge_name(uv);
        if(!I[e].is_matching_edge()) std::cout << "\t(w = " << I[e].weight << ") ";
        if(used > 1) std::cout << "x " << used;
        std::cout << std::endl;
      );
      // modify the edge multiplicity to the number of times it is actually used in the solution
      I[e].multiplicity = used;
    } else to_delete.push(e);
  }
  while(!to_delete.empty()) {
    I.delete_edge(to_delete.top());
    to_delete.pop();
  }
}

void print_cplex_stats(const Instance &I, const IloCplex &cplex, const VAR_COLLECTION &vars, const CallbackStatistics &cb_stats)
{
  unsigned opt_sol = cplex.getObjValue();
  std::cout << "Solution value  = " << opt_sol << std::endl;
  std::cout << "Solution status = " << cplex.getStatus() << std::endl;
#ifdef STATISTICS
  std::cout << "Branching nodes = " << cplex.getNnodes() << std::endl;

  std::cout << "Root relaxation value = " << cb_stats.root_relax << std::endl;
  std::cout << "Solution gap = " << 100 * (cb_stats.root_relax - opt_sol)/opt_sol << "%" << std::endl;
  std::cout << "callback stats:" << std::endl << "\tseconds used: " << cb_stats.time_spent.seconds_passed() << "\n\tcuts: "<< cb_stats.cuts_added << "\n\tcalls: "<< cb_stats.times_called << std::endl;

  std::cout << "reasons for jumps: avg " << jump_stats.num_reasons_for_jump.mean << " var " << jump_stats.num_reasons_for_jump.variance << std::endl;
  std::cout << "jump network size: avg " << jump_stats.jump_network_sizes.mean << " var " << jump_stats.jump_network_sizes.variance << std::endl;
#endif

  DEBUG5(
    std::cout << "edges used in solution: " << std::endl;
    print_true_map_vars(cplex, I, vars.times_used_undir);
  );

  DEBUG5(
    std::cout << "contig jumps: " << std::endl;
    vars.print_contig_jumps(cplex);
  );

  const PathsAndCycles pc(vars.num_paths(cplex), vars.num_cycles(cplex));
  std::cout << "solution has " << pc.p + pc.c << " objects: " << pc.p << " paths & " << pc.c << " cycles" << std::endl;
}

#endif


void linearization(ScaffoldGraph &sg, const LinearizationOptions &lin_opts, LinearizationStatistics &stats)
{
  DEBUG1(std::cout << "deconstructing the solution by destroying ambigous paths..." << std::endl);

  DEBUG4(std::cout << "removing ambigous paths..." << std::endl);

  if(lin_opts.use_brute)
    kill_ambigous_paths_brutal(sg, stats);
  else if(lin_opts.use_approx)
    kill_ambigous_paths_approx(sg, stats);
  else if(lin_opts.tw)
    kill_ambigous_paths_tw(sg, stats, lin_opts.count_vertices);
  else
    kill_ambigous_paths_ILP(sg, stats, !lin_opts.count_vertices);


  if(lin_opts.deconstruct) deconstruct_solution(sg, lin_opts);
}


void handle_output_seq(ScaffoldGraph &sg, SequenceMap *sequences, const Options &options)
{
  MultiSequenceMap output_sequences;
  if(sequences) {
    // step 2: merge contigs with negative-length non-contig edges between them
    DEBUG1(std::cout << "fixing negative non-contig lengths..." << std::endl);
    fix_negative_non_contig_lengths(sg, *sequences, options.alignment);

    // step 3: convert the instance into sequences by tracing paths
    simple_solution_to_sequences(sg, options.alignment, *sequences, output_sequences);

#ifdef STATISTICS
    // give sequence statistiques
    SequenceStatistics sstat(output_sequences);
    sstat.print();
#endif
  }
  if(options.general.out_seq_filename != "") {
    std::cout << "writing contigs in fasta format to " << options.general.out_seq_filename << std::endl;
    std::ofstream os(options.general.out_seq_filename);
    write_contigs_as_fasta(os, output_sequences);
  }
}

void handle_output_graph(const ScaffoldGraph &sg, const Options &options)
{
  // finally, output the graph in the requested form
  if(options.general.out_graph_filename != "") {
    std::cout << "writing graphviz file " << options.general.out_graph_filename << std::endl;
    std::ofstream os(options.general.out_graph_filename);
    write_scaffold(os, sg);
  }
}

void print_stats_solution_graph(const std::string &file, const ScaffoldGraph &sg)
{
  unsigned nb_ambiguous_paths = 0;
  unsigned nb_non_ambiguous_paths = 0;
  unsigned sum_delta = 0;
  unsigned min_degree = UINT_MAX;
  unsigned max_degree = 0;
  unsigned sum_weight = sg.sum_multiplicities();

  for(auto v_iter = sg.get_vertices(); v_iter; ++v_iter) {
    const ScafVertex &u = *v_iter;
    // if u is an extremity of an ambigous path, kill its non-matching edges
    const std::pair<ScafVertex, bool> result = is_ambigous_path_extremity(u, sg);
    if(result.second) {
      ++nb_ambiguous_paths;
      sum_delta += sg.degree(u);
      if(sg.degree(u) > max_degree) max_degree = sg.degree(u);
      if(sg.degree(u) < min_degree) min_degree = sg.degree(u);
    } else ++nb_non_ambiguous_paths;
  }
  nb_ambiguous_paths /= 2;
  nb_non_ambiguous_paths /= 2;
  std::cout << "Writing stats in " << file << std::endl;
  std::ofstream ofs;
  ofs.open(file, std::ofstream::out | std::ofstream::app);
  ofs << " Number of ambiguous paths: " << nb_ambiguous_paths << std::endl;
  ofs << " Number of non ambiguous paths: " << nb_non_ambiguous_paths << std::endl;
  ofs << " Total weight: " << sum_weight << std::endl;
  ofs << " Average degree of ambiguous paths extremities: " << (float) sum_delta / (2 * nb_ambiguous_paths) << std::endl;
  ofs << " Degree Max: " << max_degree << std::endl;
  ofs << " Degree Min: " << min_degree << std::endl;

  ofs.close();
}

void print_stats_linearization(const LinearizationOptions &lin_opts, LinearizationStatistics &stats)
{
  if(lin_opts.stats_file != "") {
    DEBUG1(std::cout << "Writing stats in " << lin_opts.stats_file << std::endl);
    std::ofstream ofs;
    ofs.open(lin_opts.stats_file, std::ofstream::out | std::ofstream::app);
    ofs << " Number of cuts: " << stats.nb_cuts << std::endl;
    ofs << " Weight removed: " << stats.weight_cut << std::endl;
    ofs.close();
  }
}

void apply_reduction_rules(ScaffoldGraph &g)
{
  DEBUG1(std::cout << "applying linearization reduction rules..." << std::endl);
  preprocess::contract_uniform_paths(g);
  preprocess::split_non_ambiguous_paths(g);
}


int main(const int argc, const char **argv)
{
  // initialize command-line options
  auto opt_desc = initialize_options(argv[0]);
  add_alignment_options(opt_desc);
#ifdef CPLEX
  add_ilp_options(opt_desc);
#endif
  add_linearization_options(opt_desc);
  // parse arguments
  const auto var_map = parse_arguments(argc, argv, opt_desc);

  SequenceMap *sequences = get_sequences(get_option<std::string>(var_map, "is", ""));
  const Options options =
    {
      handle_options(var_map, opt_desc),
      handle_alignment_options(var_map, sequences),
#ifdef CPLEX
      handle_ilp_options(var_map),
#endif
      handle_linearization_options(var_map)
    };

  Instance *I(get_instance(options, sequences));

  if(options.general.do_scaffolding) {
#ifdef CPLEX
    ScafVPairDependencyMap dependencies;
    preprocessing(options, *I, sequences, dependencies);

    IloEnv env;
    IloModel model(env);
    VAR_COLLECTION vars(*I);
    CallbackStatistics cb_stats;

    IloCplex *cplex_p = build_model(options, dependencies, env, model, vars, cb_stats);
    solve(*cplex_p);

    print_cplex_stats(*I, *cplex_p, vars, cb_stats);
    apply_solution(*cplex_p, *I, vars);
#endif
  }
  if(options.general.do_linearization){
    if(options.lin.solution_graph_stats_file == "") {
      ScaffoldGraph solution_copy((ScaffoldGraph) *I);
      LinearizationStatistics stats({0, 0});

      if(options.lin.reduction_rules)
        apply_reduction_rules(solution_copy);

      linearization(solution_copy, options.lin, stats);
      print_stats_linearization(options.lin, stats);

      //TODO: remove me
      preprocess::remove_deg1(*I);


      handle_output_seq(solution_copy, sequences, options);
      handle_output_graph(solution_copy, options);
    } else {
      apply_reduction_rules(*I);
      print_stats_solution_graph(options.lin.solution_graph_stats_file, *I);
    }
  }

  exit(EXIT_SUCCESS);

}




