

#ifndef FES_BRANCHING_HPP
#define FES_BRANCHING_HPP

#include <pthread.h>

#include "utils/graph_typedefs.hpp"
#include "utils/graph_utils.hpp"
#include "utils/graph_parameters.hpp"
#include "utils/instance.hpp"
#include "utils/scaffolding_utils.hpp"
#include "degree_branching.hpp"

// if a bridge splits away an FES bigger than this, we take it
#define BRIDGE_THRESHOLD 1
// apply preprocessing each x layers
#define PP_EACH_DEPTH 2

#define MIN_FES_TO_PARALLELIZE 8
#define MAX_NUM_THREADS 16


namespace scaffold{
  namespace solv{
    pthread_mutex_t thread_lock;
    size_t num_threads = 0;

    template<class Graph>
    void FES_solve_recursively(Instance<Graph>&, DynProgSolution<>*&, size_t depth = 0 );

    // pthread stuff
    template<class Graph>
    struct pthread_data_block{
      Instance<Graph>* I;
      DynProgSolution<>* S;
      size_t depth;

      pthread_data_block(Instance<Graph>* const _I, DynProgSolution<>* const _S, const size_t _depth) : I(_I), S(_S), depth(_depth) {}
    };
    // parallel version of FES_solve_recursively
    template<class Graph>
    void* FES_solve_rec_parallel(void* _data){
      pthread_data_block<Graph>* data = (pthread_data_block<Graph>*)_data;
      DEBUG1(std::cout << "thread "<<pthread_self()<<" starting..."<<std::endl);
      FES_solve_recursively(*data->I, data->S, data->depth);
      DEBUG1(std::cout << "thread "<<pthread_self()<<" finishing..."<<std::endl);
      // update number of threads
      pthread_mutex_lock(&thread_lock);
        --num_threads;
      pthread_mutex_unlock(&thread_lock);
      return NULL;
    }




    // store all branches around v in a list
    template<class FGraph, class Graph>
    void get_FES_branches_for_vertex(const Instance<Graph>& I, const FGraph& fg, const Vertex<FGraph>& v, std::vector<Branch<Graph> >& branches){
      // reserve space for #branches = degree(v)
      branches.clear();
      branches.reserve(degree(v, fg));
      // add the "delete-all" branch
      branches.emplace_back(DeleteBranch);
      // for each edge uv incident to v, create the branch that takes uv and add uv to the "delete-all" branch
      for(AdjIterRange<FGraph> r = adjacent_vertices(v, fg); r.first != r.second; ++r.first)
        if(v != I.matched_with(*r.first)) {
          const VertexPair<Graph> p(v, *r.first);
          // construct a new branch at the end of branches
          branches.emplace_back(TakeBranch, p);
          // add the edge to the "delete all"-branch at the front
          branches.front().edges.emplace_back(p);
        }// if
    }// function

    // get the vertex of maximum degree in fg, breaking ties with the degree in g
    template<class FGraph, class Graph>
    Vertex<Graph> get_best_branching_vertex(const FGraph& fg, const Graph& g){
      VertexIterRange<Graph> r = vertices(g);
      // assume we have vertices
      assert(r.first != r.second);
      VertexIter<Graph> max_v = r.first;
      size_t max_deg_fg = degree(*max_v, fg);
      size_t max_deg_g  = degree(*max_v, g);
      ++r.first;
      for(; r.first != r.second; ++r.first){
        const Vertex<Graph>& v = *r.first;
        const size_t deg_fg = degree(v, fg);
        if(deg_fg >= max_deg_fg){
          const size_t deg_g = degree(v, g);
          if((deg_fg == max_deg_fg) ? (deg_g > max_deg_g) : true){
            max_deg_fg = deg_fg;
            max_deg_g  = deg_g;
            max_v = r.first;
          }// if
        }// if
      }// for
      return *max_v;
    }// function

    // outsource the branching
    template<class Graph>
    void branch_to_get_min_sol(Instance<Graph>& I, std::vector<Branch<Graph> >& branches, DynProgSolution<>& min_sol, const size_t depth){
      const size_t bnum = branches.size();
      Instance<Graph>* instances[bnum];
      DynProgSolution<>* solutions[bnum];
      pthread_data_block<Graph>* pthread_data_blocks[bnum];
      pthread_t threads[bnum];
      // to create joinable threads
      pthread_attr_t attr;
      pthread_attr_init(&attr);
      pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
      size_t threads_created = 0;

      // Step 1: apply the branches to the graph, creating a list of graphs
      for(size_t i = 0; i < bnum; ++i){
        // Step 1a: get a copy H of I to which the branch has been applied
        solutions[i] = new DynProgSolution<>(I);
        instances[i] = copy_and_apply_branch(I, branches[i], solutions[i]);
      }// for
      // Step 2: recurse for each created instance
      for(size_t i = 0; i < bnum; ++i){
        // prepare data block
        DEBUG3(std::cout << "depth "<<depth<<":"<<std::endl<<"branching: "<< BranchAndGraph<Graph>(branches[i], *I.g)<<std::endl);

        // don't go parallel if the FES is too low.
        bool create_parallel = false;
        if(I.infos.FES() > MIN_FES_TO_PARALLELIZE){
          // see whether we have enough slots to branch in parallel
          pthread_mutex_lock(&thread_lock);
            create_parallel = (num_threads < MAX_NUM_THREADS);
            if(create_parallel) ++num_threads;
          pthread_mutex_unlock(&thread_lock);
          if(!create_parallel) std::cout << "not creating new thread since we have already "<<num_threads<<" of them"<<std::endl;
        }// if
        if(create_parallel){
          // prepare the data block
          pthread_data_blocks[threads_created] = new pthread_data_block<Graph>(instances[i], solutions[i], depth + 1);
          // create the thread
          int pt_err = pthread_create(&threads[threads_created], &attr, FES_solve_rec_parallel<Graph>, (void*)pthread_data_blocks[threads_created]);
          if(pt_err){
            std::cerr << "error "<<pt_err<<" while creating thread"<<std::endl;
            exit(-1);
          }
          ++threads_created;
        } else FES_solve_recursively(*(instances[i]), solutions[i], depth + 1);

        DEBUG1(if(depth == 0) std::cout << (i+1)*100/bnum <<"% done" << std::endl);
      }// for
      // wait for all threads to finish
      if(threads_created){
        pthread_attr_destroy(&attr);
        for(size_t i = 0; i < threads_created; ++i){
          int pt_err = pthread_join(threads[i], NULL);
          if(pt_err) {
            std::cerr << "error "<<pt_err<<" while joining threads"<<std::endl;
            exit(-1);
          }// if
          delete pthread_data_blocks[i];
        }// for
      }// if
      // Step 3: update solution
      for(size_t i = 0; i < bnum; ++i){
        if(!solutions[i]->no_solutions()) min_sol.combine_min_of_two(*(solutions[i]));
        delete instances[i];
        delete solutions[i];
      }// for
      // clean up
    }// function

    /*
    // outsource the branching
    template<class Graph>
    void branch_to_get_min_sol(const Instance<Graph>& I, std::vector<Branch<Graph> >& branches, DynProgSolution<>& min_sol, const size_t depth, size_t& searchtree_nodes){
      const size_t bnum = branches.size();
      std::vector<Instance<Graph>*> instances(bnum, NULL);
      std::vector<DynProgSolution<>*> solutions(bnum, NULL);
      // Step 1: apply the branches to the graph, creating a list of graphs
      for(size_t i = 0; i < bnum; ++i){
        // Step 1a: get a copy H of I to which the branch has been applied
        solutions[i] = new DynProgSolution<>(I);
        instances[i] = copy_and_apply_branch(I, branches[i], solutions[i]);
      }// for
      // Step 2: recurse for each created instance
//  #pragma omp parallel for if (depth < 5) num_threads(8)
      for(size_t i = 0; i < bnum; ++i){
        size_t tmp = 0;
        DEBUG3(std::cout << "depth "<<depth<<":"<<std::endl<<"branching: "<< BranchAndGraph<Graph>(branches[i], *I.g)<<std::endl);
        FES_solve_recursively(*(instances[i]), solutions[i], depth + 1, tmp);
        DEBUG4(std::cout << "depth "<<depth<<": returned with solution"<<std::endl);
        DEBUG1(if(depth == 0) std::cout << (i+1)*100/bnum <<"% done" << std::endl);
        delete instances[i];
      }
      // Step 3: update solution
      for(size_t i = 0; i < bnum; ++i){
        if(!solutions[i]->no_solutions()) min_sol.combine_min_of_two(*(solutions[i]));
        delete solutions[i];
      }// for
    }// function
    */

    // recursively transform I into a solution, TODO: add upper and lower bounds
    template<class Graph>
    void FES_solve_recursively(Instance<Graph>& I, DynProgSolution<>*& S, size_t depth = 0){
      DEBUG3(std::cout << "depth "<< depth<<": solving instance with " << num_vertices(*I.g) << " vertices and "<<num_edges(*I.g)<<" edges." << std::endl);
      DEBUG5(write_adj_list_graph(std::cout, *I.g); std::cout << std::endl);

      // Step 0: preprocess
      if(depth % PP_EACH_DEPTH == 0)
        while(preprocess::pp_matching_pendant(I, S)) {};

      // Step 1: check for connected components
      Instance<Graph> *J = I.split_off_instance();
      if(J){
        // Step 1c: if a component has been split off, then recurse for the split-off instance
        DEBUG3(std::cout << "splitting into components of size "<<num_vertices(*I.g)<<" and "<<num_vertices(*J->g)<<std::endl);
        DynProgSolution<>* solJ = new DynProgSolution<>(*J);
        FES_solve_recursively(I, S, depth);
        if(!S->no_solutions()){
          FES_solve_recursively(*J, solJ, depth);
          delete J;

          DEBUG3(std::cout << "depth "<<depth<<": rejoining solutions for components"<<std::endl);

          if(!solJ->no_solutions()){
            // if both components have solutions, unite them and find a solution subset
            S->combine_disjoint_union(*solJ);
            delete solJ;
            return;
          }
        }
        if(S->no_solutions() || solJ->no_solutions()){
          // if either component does not have a solution, return failure
          S->invalidate();
          delete solJ;
          return;
        }
      } 
      // Step 2: base cases
      if(I.max_deg_two()) {
        DEBUG3(std::cout << "it's already max-deg 2, getting all sub-solutions"<<std::endl);
        treat_deg_two(I, S);
        return;
      }
      if(I.is_acyclic()){
        DEBUG3(std::cout << "it's a tree, getting all sub-solutions"<<std::endl);
        solve_on_tree(I, S);
        return;
      }
      // Step 4: find a good branching
      // Step 4a: try the bridges
      std::vector<Branch<Graph> > branches;
      const size_t I_FES = I.infos.FES();
      typename BridgeMap<Graph>::const_iterator best_b = I.get_best_bridge();
      // if there is a bridge splitting the FES nicely...
      if( (best_b != I.infos.bridges.end()) ? ((best_b->second > BRIDGE_THRESHOLD) && (best_b->second < I_FES - BRIDGE_THRESHOLD)) : false){
        // ... then take it
        std::pair<Edge<Graph>,bool> e = edge(best_b->first.first, best_b->first.second, *I.g);
        assert(e.second);
        DEBUG4(std::cout << "branching on bridge: "<< VertexPairAndGraph<Graph>(best_b->first, *I.g)<<" splitting away an FES of "<<best_b->second<<"/"<<I_FES<<std::endl);
        branches.emplace_back(TakeBranch, e.first, *I.g);
        branches.emplace_back(DeleteBranch, e.first, *I.g);
      } else {
        // otherwise, get the maximum degree vertex v in fg, and
        // Step 4b:  create a filtered graph, edge filtered by "not in bridges" (invert the "is bridges"-filter)
        typedef VP_double_contained_filter<Graph, BridgeMap<Graph>, Matching<Graph> > BridgeFilter;
        typedef filtered_graph<Graph, BridgeFilter> FGraph;
        BridgeFilter bfilter(&I.infos.bridges, &I.matched, I.g, true);
        FGraph fg(*I.g, bfilter);

        const Vertex<Graph> v = get_best_branching_vertex(fg, *I.g);
        DEBUG4(std::cout << "branching on vertex: "<< ( *I.g)[v].name<<std::endl);
        // get all branches for v
        get_FES_branches_for_vertex(I, fg, v, branches);
      }

      DEBUG1(if(depth == 0) std::cout << "starting with solution "<<std::endl<<*S<<std::endl);
      // Step 5: for each branch, make a copy of I, apply the branching, and recurse
      DynProgSolution<> min_sol(I);
      branch_to_get_min_sol(I, branches, min_sol, depth);
      // Step 6: evaluate solution
      if(min_sol.no_solutions()) S->invalidate(); else S->destructive_append(min_sol);
      DEBUG3(std::cout << "returning from depth "<<depth<<" with "; if(S->no_solutions()) std::cout << "no solutions"<<std::endl; else std::cout << "solutions: "<<*S<<std::endl;);
    } // function

  } // namespace
} // namespace

#endif
