

#ifndef ILP_MODEL_MULTI_HPP
#define ILP_MODEL_MULTI_HPP

#include "boost/multi_array.hpp"

#include <string>
#include <vector>

#ifdef STATISTICS
#include "utils/stats.hpp"
#endif


#include "utils/scaffold_graph.hpp"
#include "utils/profiling.hpp"
#include "utils/ilp_options.hpp"
#include "utils/ilp_common.hpp"
#include "utils/ilp_contig_jump_network.hpp"

/*
// support keeping track of the subgraph number, the layer, and the corresponding vertex in the original graph
namespace boost{
  enum vertex_subgraph_nr_t { vertex_subgraph_nr = 201 };
  BOOST_INSTALL_PROPERTY(vertex, subgraph_nr);

  enum vertex_layer_t { vertex_layer = 202 };
  BOOST_INSTALL_PROPERTY(vertex, layer);
};
*/

namespace scaffold{

  // auxiliary graph for callback construction
  typedef boost::adjacency_list<
    boost::hash_setS,
    boost::listS,
    boost::bidirectionalS,
    boost::property<boost::vertex_subgraph_nr_t, size_t,
      boost::property<boost::vertex_layer_t, size_t> 
    >,
    boost::no_property
  > AuxGraphs;

  // the y variable map has an entry for each pairs of subgraphs
  template<class Graph>
  using EdgeVarMapMatrix = boost::multi_array<EdgeVarMap<Graph>, 2>;

  // the vertex-translator needs to have an entry for each (subgraph)x(layer)
  template<class Graph>
  using TranslateMatrix = boost::multi_array<TranslateBiMap<Graph, AuxGraphs>, 2>;


  template<class Graph>
  struct var_collection {
    const Instance<Graph>& I;
    const size_t max_multi;

    EdgeVarMap<Graph> x,z;
    EdgeVarMapMatrix<Graph> y;
    std::vector<vertex_var_map<Graph> > s, tp, tc;

    var_collection(const Instance<Graph>& _I):
      I(_I),
      max_multi(_I.get_max_multiplicity_no_update()),
      y(boost::extents[max_multi][max_multi]),
      s(max_multi),
      tp(max_multi),
      tc(max_multi)
    {}
  };


  // block isolated cycles as lazy constraint callback
  template<class Graph>
  class LazyCyclesBlockCallback: public IloCplex::LazyConstraintCallbackI{
    const var_collection<Graph>& vars;
    callback_statistics& stats;

  public:

    void build_aux_graph(AuxGraphs& aux, TranslateMatrix<Graph>& translate);
    void cleanup_aux_graph(AuxGraphs& aux, TranslateMatrix<Graph>& translate);
    void main();

    IloCplex::CallbackI* duplicateCallback() const{
      return (new (getEnv()) LazyCyclesBlockCallback(getEnv(), vars, stats));
    }
    
    LazyCyclesBlockCallback(IloEnv env, const var_collection<Graph>& _vars, callback_statistics& _stats):
      IloCplex::LazyConstraintCallbackI(env),
      vars(_vars),
      stats(_stats)
    {}

  }; // class

  // callback function
  template<class Graph>
  IloCplex::Callback GetLazyCyclesBlock(IloEnv env, const var_collection<Graph>& vars, callback_statistics& stats) {
    return (new (env) LazyCyclesBlockCallback<Graph>(env, vars, stats));
  }

  // build the auxiliary graph using the var collection vars
  template<class Graph>
  void LazyCyclesBlockCallback<Graph>::build_aux_graph(AuxGraphs& aux, TranslateMatrix<Graph>& translate){
    const Instance<Graph>& I = vars.I;

    // construct the auxiliary graph from the current solution
    for(size_t i = 0; i < vars.max_multi; ++i){
      for(size_t j = 0; j < vars.max_multi; ++j){
        for(const auto& y : vars.y[i][j]){
          const VertexPair<Graph>& uv = y.first;
          const Vertex<Graph>& u = uv.first;
          const Vertex<Graph>& v = uv.second;
          const IloNumVar& y_val = y.second;
          const unsigned layer = I.is_matching_edge(uv) ? 0 : 1;

          if(is_true(getValue(y_val))){
            // add vertices: first see if they're already there & if they are not there, add them to aux

            auto u_it = translate[i][layer].left.find(u);
            if(u_it == translate[i][layer].left.end()){
              const Vertex<AuxGraphs> au = boost::add_vertex(aux);
              boost::put(vertex_subgraph_nr, aux, au, i);
              boost::put(vertex_layer, aux, au, layer);
              u_it = translate[i][layer].left.insert(u_it, typename TranslateBiMap<Graph, AuxGraphs>::left_map::value_type(u, au));
            }

            auto v_it = translate[j][1-layer].left.find(v);
            if(v_it == translate[j][1-layer].left.end()){
              const Vertex<AuxGraphs> av = boost::add_vertex(aux);
              boost::put(vertex_subgraph_nr, aux, av, j);
              boost::put(vertex_layer, aux, av, 1-layer);
              v_it = translate[j][1-layer].left.insert(v_it, typename TranslateBiMap<Graph, AuxGraphs>::left_map::value_type(v, av));
            }
            
            // add edge
            const Vertex<AuxGraphs>& aux_u = u_it->second;
            const Vertex<AuxGraphs>& aux_v = v_it->second;
            boost::add_edge(aux_u, aux_v, aux);
          } // if
        }// for
      }// for
    }// for
  }// function


  // remove all paths that are incident with s (that is, incident with a vertex in pending)
  template<class Graph>
  void LazyCyclesBlockCallback<Graph>::cleanup_aux_graph(AuxGraphs& aux, TranslateMatrix<Graph>& translate){
    // construct the set of vertices in layer 1 that are reached from s
    VertexSet<AuxGraphs>* pending = new VertexSet<Graph>();
    DEBUG5(std::cout << "reachable from s: ");
    for(size_t i = 0; i < vars.max_multi; ++i)
      for(const auto& y_su : vars.s[i])
        if(is_true(getValue(y_su.second))) {
          const Vertex<Graph>& u = y_su.first;
          DEBUG5(std::cout << ( *vars.I.g)[u].name <<" ");
          // CPLEX might give us a solution with s->u->t_p paths; if, in this case, u doesn't have an edge in aux, we can't translate
          auto u_iter = translate[i][0].left.find(u);
          if(u_iter != translate[i][0].left.end()) pending->insert(u_iter->second);
        }// if
    DEBUG5(std::cout << "."<< std::endl);

    // parse & remove all paths & cycles incident with pending
    unsigned current_layer = 0;
    while(!pending->empty()){
      VertexSet<AuxGraphs>* neighbors = new VertexSet<AuxGraphs>();
      for(const Vertex<AuxGraphs>& aux_u : *pending){
        const size_t subgraph_nr = boost::get(vertex_subgraph_nr, aux, aux_u);
        const size_t layer = boost::get(vertex_layer, aux, aux_u);
        // step 1: add all neighbors of u to the list of next pending vertices
        for(OEdgeIterRange<AuxGraphs> r = boost::out_edges(aux_u, aux); r.first != r.second; ++r.first)
          neighbors->insert(boost::target(*r.first, aux));
        // step 2: remove u from the translate map
        translate[subgraph_nr][layer].right.erase(aux_u);
        // step 3: remove u from the aux graph
        boost::clear_vertex(aux_u, aux);
        boost::remove_vertex(aux_u, aux);
      }// for
      current_layer = 1 - current_layer;
      delete pending;
      pending = neighbors;
    }// while
    delete pending;
  }// function

  template<class Graph>
  void LazyCyclesBlockCallback<Graph>::main(){
    stats.time_spent.resume();
    DEBUG5(std::cout << "callback starts"<<std::endl);

    const Graph& g = *vars.I.g;
    AuxGraphs aux;
    // a translate map for each 2-layered "subgraph"
    TranslateMatrix<Graph> translate(boost::extents[vars.max_multi][2]);

    // Step 1: build the auxiliary graph & remove everything that is reachable from s
    DEBUG5(std::cout << "building auxiliary graph..."<<std::endl);
    build_aux_graph(aux, translate);
    DEBUG5(std::cout << "built aux graph with "<<boost::num_vertices(aux)<<" vertices & "<<boost::num_edges(aux)<<" edges"<<std::endl);

    DEBUG5(std::cout << "removing valid paths/cycles from auxiliary graph..."<<std::endl);
    cleanup_aux_graph(aux, translate);
    DEBUG5(std::cout << "after cleanup, "<<boost::num_vertices(aux)<<" vertices & "<<boost::num_edges(aux)<<" edges remain"<<std::endl);

    // whatever remains of aux are cycles that are not attached to s, that is, invalid cycles;
    // Step 2: forbid them by adding a constraint forcing at least one arc going out of the cycle

    unsigned num_new_cuts = 0;
#ifdef LAZY_ADD_ALL_CYCLES
    while(boost::num_edges(aux)){
#else
    if(boost::num_edges(aux)){
#endif
      IloExpr expr(getEnv()), rev_expr(getEnv());
      VertexSet<AuxGraphs> S;

      DEBUG5(std::cout << "forbidding cyclic structure: ");
      // find a set S of vertices in aux such that aux[S] is a cycle; write down the maximum multiplicity within S
      size_t max_multiplicity = 1;
      const Edge<AuxGraphs> e = *(boost::edges(aux).first);
      const Vertex<AuxGraphs> x = boost::source(e, aux);
      VertexQueue<AuxGraphs> pending;
      pending.push(x);
      S.insert(x);
      while(!pending.empty()){
        const Vertex<AuxGraphs> aux_u = pending.front(); pending.pop();
        const unsigned u_subgraph_nr = boost::get(vertex_subgraph_nr, aux, aux_u);
        const unsigned u_layer = boost::get(vertex_layer, aux, aux_u);
        const Vertex<Graph>& u = translate[u_subgraph_nr][u_layer].right.at(aux_u);
        for(OEdgeIterRange<AuxGraphs> r = boost::out_edges(aux_u, aux); r.first != r.second;){
          const Edge<AuxGraphs> aux_uv = *r.first;
          const Vertex<AuxGraphs> aux_v = boost::target(aux_uv, aux);
          const unsigned v_subgraph_nr = boost::get(vertex_subgraph_nr, aux, aux_v);
          const unsigned v_layer = boost::get(vertex_layer, aux, aux_v);

          const Vertex<Graph>& v = translate[v_subgraph_nr][v_layer].right.at(aux_v);
          const VertexPair<Graph> uv = VertexPair<Graph>(u,v);
          const VertexPair<Graph> vu = VertexPair<Graph>(v,u);

          const Edge<Graph> e = boost::edge(u, v, g).first;
          max_multiplicity = std::max(max_multiplicity, boost::get(edge_multiplicity, g, e));
          
          DEBUG5(std::cout << " " << VertexPairAndGraph<Graph>(uv, g));

          // modify the expressions for the edge uv
          expr     -= vars.y[u_layer][v_layer].at(uv);
          rev_expr -= vars.y[v_layer][u_layer].at(vu);

          // delete the offending edge from aux
          ++r.first;
          boost::remove_edge(aux_uv, aux);

          // add aux_v to the queue & mark it seen
          if(!contains(S, aux_v)) {
            pending.push(aux_v);
            S.insert(aux_v);
          }// if
        }// for
      }// while (pending is not empty)
      DEBUG5(std::cout << std::endl);
      const size_t S_size = S.size();
      // add the exiting arcs to the expressions
      for(const Vertex<AuxGraphs>& aux_u : S){
        const unsigned u_subgraph = boost::get(vertex_subgraph_nr, aux, aux_u);
        const unsigned u_layer = boost::get(vertex_layer, aux, aux_u);
        const Vertex<Graph>& u = translate[u_subgraph][u_layer].right.at(aux_u);
        // note: if the layer of u is 0, then uv is a matching edge and there are no edges from u leaving S (except to t_c)
        if(u_layer == 1){
          for(AdjIterRange<Graph> r = boost::adjacent_vertices(u, g); r.first != r.second; ++r.first){
            const Vertex<Graph>& v = *r.first;
            for(size_t v_layer = 0; v_layer < vars.max_multi; ++v_layer){
              const auto aux_v_it = translate[v_layer][0].left.find(v);
              // if the aux-version of uv is not within SxS, then modify the expressions using uv
              if( (aux_v_it == translate[v_layer][0].left.end()) || !contains(S, aux_v_it->second)){
                const VertexPair<Graph> uv = VertexPair<Graph>(u, v);
                const VertexPair<Graph> vu = VertexPair<Graph>(v, u);

                expr += (double)(S_size * max_multiplicity) * vars.y[u_layer][v_layer].at(uv);
                rev_expr += (double)(S_size * max_multiplicity) * vars.y[v_layer][u_layer].at(vu);
              }// if
            }// for
          }// for
        } else {
          // if the layer of u is 0, then add u -> t_c
          expr += (double)(S_size * max_multiplicity) * vars.tc[u_subgraph].at(u);
          rev_expr += (double)(S_size * max_multiplicity) * vars.tc[u_subgraph].at(u);

        }// if layer = 1 / else
      }// for S
      // add the expressions to the model
      add(expr >= 0);
      add(rev_expr >= 0);
      num_new_cuts += 2;
#ifdef LAZY_ADD_ALL_CYCLES
    }// while  aux has edges
#else
    } // if aux has edges
#endif
    // Step 3: collect statistical information
    DEBUG2(std::cout << "=======   callback called, added "<<num_new_cuts<<" new cuts ======"<<std::endl);
    stats.times_called++;
    stats.cuts_added += num_new_cuts;
    stats.time_spent.pause();
  } // main()

    
#warning TODO: use each edges multiplicity to restrict the domain of its variable instead of adding a constraint
  // initialize the variables for the ILP formulation
  template<class Graph>
  void populate_variables(IloEnv& env, var_collection<Graph>& vars){
    const Instance<Graph>& I = vars.I;
    const Graph& g = *I.g;
    const size_t max_multi = vars.max_multi;

    DEBUG5(std::cout << "creating z_{u,v}/x_{u,v}/y_uv"<<std::endl);
    for(EdgeIterRange<Graph> e = boost::edges(g); e.first != e.second; ++e.first){
      const Vertex<Graph>& u = source(*e.first, g);
      const Vertex<Graph>& v = target(*e.first, g);
      const std::string uname = std::to_string(boost::get(boost::vertex_name, g, u));
      const std::string vname = std::to_string(boost::get(boost::vertex_name, g, v));
      
      // Step 1: add binary variables z_e = 1   <=>  e belogs to a solution path/cycle
      const std::string zname("z_" + uname + "," + vname);
      vars.z.emplace(std::piecewise_construct, std::make_tuple(u,v), std::make_tuple(env, 0, 1, IloNumVar::Bool, zname.c_str()));
      const std::string xname("x_" + uname + "," + vname);
      vars.x.emplace(std::piecewise_construct, std::make_tuple(u,v), std::make_tuple(env, 0, max_multi, IloNumVar::Int, xname.c_str()));
      // Step 2: add variables y^k_ij
      const std::string yname1("y_" + uname + "," + vname);
      const std::string yname2("y_" + vname + "," + uname);
      for(size_t i = 0; i < max_multi; ++i)
        for(size_t j = 0; j < max_multi; ++j){
          vars.y[i][j].emplace(std::piecewise_construct, std::make_tuple(u,v), std::make_tuple(env, 0, 1, IloNumVar::Bool, yname1.c_str()));
          vars.y[i][j].emplace(std::piecewise_construct, std::make_tuple(v,u), std::make_tuple(env, 0, 1, IloNumVar::Bool, yname2.c_str()));
        }
    }
    DEBUG5(std::cout << "creating s/tp/tc interaction variables"<<std::endl);
    // Step 3: add variables y_s,i and y_i,tp and y_i,tc
    for(VertexIterRange<Graph> v = boost::vertices(g); v.first != v.second; ++v.first){
      const Vertex<Graph>& u = *v.first;
      const std::string name(std::to_string(boost::get(boost::vertex_name, g, u)));
      for(size_t i = 0; i < max_multi; ++i){
        // on layer 0, come up from s
        vars.s[i].emplace(std::piecewise_construct, std::make_tuple(u), std::make_tuple(env, 0, 1, IloNumVar::Bool, ("y_s," + name).c_str()));
        // on layer 1, go up to tp
        vars.tp[i].emplace(std::piecewise_construct, std::make_tuple(u), std::make_tuple(env, 0, 1, IloNumVar::Bool, ("y_" + name + ",tp").c_str()));
        // on layer 0, go up to tc
        vars.tc[i].emplace(std::piecewise_construct, std::make_tuple(u), std::make_tuple(env, 0, 1, IloNumVar::Bool, ("y_" + name + ",tc").c_str()));
      }// for
    }// for
  }

  // initialize the constraints for the ILP formulation
  template<class Graph>
  void populate_constraints(IloEnv& env, IloRangeArray& c, const var_collection<Graph>& vars){
    const Instance<Graph>& I = vars.I;
    const Graph& g(*I.g);

    // constraint (16): x_e = 0 --> z_e = 0
    DEBUG5(std::cout << "adding (16): x_e = 0 --> z_e = 0"<<std::endl);
    for(auto x : vars.x){
      const VertexPair<Graph> uv = x.first;

      IloExpr expr(env);
      expr += x.second;
      expr -= vars.z.at(uv);
      c.add(expr >= 0);
    }

    // constraint (17): of all the representations of ij, exactly x_ij should be chosen!
    DEBUG5(std::cout << "adding (17): of all the representations of e, exactly x_e should be chosen"<<std::endl);
    for(auto x : vars.x){
      const VertexPair<Graph> uv = x.first;
      const VertexPair<Graph> vu(uv.second, uv.first);

      IloExpr expr(env);
      for(size_t i = 0; i < vars.max_multi; ++i){
        for(size_t j = 0; j < vars.max_multi; ++j){
          expr += vars.y[i][j].at(uv);
          expr += vars.y[i][j].at(vu);
        }
      }
      expr -= x.second;
      c.add(expr == 0);
    }

    // constraint (2): all matching-edges should be chosen!
    DEBUG5(std::cout << "adding (2'): any matching-edge e should be chosen between 1 and m(e) times"<<std::endl);
    for(auto x : vars.x)
      if(I.is_matching_edge(x.first)) {
        const Edge<Graph> e = boost::edge(x.first.first, x.first.second, g).first;
        c.add(x.second >= 1);
        c.add(x.second <= (double)boost::get(boost::edge_multiplicity, g, e));
        DEBUG3(std::cout<<"allowing "<<EdgeAndGraph<Graph>(e, g)<<" with multi "<<(double)boost::get(boost::edge_multiplicity, g, e)<<std::endl);
      }
  
    // constraint (4),(6): there should be at most sigma_p paths & sigma_c cycles
    DEBUG5(std::cout << "adding (4)&(6): there should be at most sigma_p paths & sigma_c cycles"<<std::endl);
    {
      IloExpr p_expr(env), c_expr(env);
      for(size_t i = 0; i < vars.max_multi; ++i) {
        for(auto y : vars.tp[i]) p_expr += y.second;
        for(auto y : vars.tc[i]) c_expr += y.second;
      }
      c.add(p_expr <= (double)I.num_paths);
      c.add(c_expr <= (double)I.num_cycles);
    }

    // constraint (3): everything that comes into layer l should leave layer l
    DEBUG5(std::cout << "adding (18): flow conservation"<<std::endl);
    for(VertexIterRange<Graph> u = boost::vertices(g); u.first != u.second; ++u.first){
      for(size_t i = 0; i < vars.max_multi; ++i){
        for(size_t layer = 0; layer < 2; ++layer){
          IloExpr expr(env);
          for(AdjIterRange<Graph> nh = boost::adjacent_vertices(*u.first, g); nh.first != nh.second; ++nh.first){
            const VertexPair<Graph> uv(*u.first, *nh.first);
            const VertexPair<Graph> vu(*nh.first, *u.first);
            // remember: layer 0 -> 1 contains only matching edges
            if(I.is_matching_edge(uv) == (layer == 0)){
              for(size_t j = 0; j < vars.max_multi; ++j) expr += vars.y[i][j].at(uv);
              DEBUG5(std::cout << " + y[" << i << "][*]_" << VertexPairAndGraph<Graph>(uv, g));
            } else {
              for(size_t j = 0; j < vars.max_multi; ++j) expr -= vars.y[j][i].at(vu);
              DEBUG5(std::cout << " - y[*]["<<i<<"]_"<<VertexPairAndGraph<Graph>(vu, g));
            }// if
          }// for
          if(layer == 0){
            // on layer 0, p-flow can come up from s
            DEBUG5(std::cout << " - y_s["<<i<<"],"<< g[*u.first].name);
            expr -= vars.s[i].at(*u.first);
            // on layer 0, p-flow can go up to tc
            DEBUG5(std::cout << " + y_"<< g[*u.first].name<<",tc["<<i<<"]");
            expr += vars.tc[i].at(*u.first);
          } else {
            // on layer 1, p-flow can go up to tp
            DEBUG5(std::cout << " + y_"<< g[*u.first].name<<",tp["<<i<<"]");
            expr += vars.tp[i].at(*u.first);
          }// if
          c.add(expr == 0);
        }// for
      }// for
    }// for
    // constraint (7): everything that goes to tc has to come in from s
    DEBUG5(std::cout << "adding (7): cycle conservation"<<std::endl);
    for(VertexIterRange<Graph> u = vertices(g); u.first != u.second; ++u.first){
      IloExpr expr(env);
      for(size_t i = 0; i < vars.max_multi; ++i) {
        expr += vars.s[i].at(*u.first);
        expr -= vars.tc[i].at(*u.first);
      }
      c.add(expr >= 0);
    }
    // constraint (5): each subset S of vertices should contain at most |S|-1 endpoints of arcs
    //DEBUG5(std::cout << "adding (5): each subset S of vertices should contain at most |S|-1 endpoints of arcs"<<std::endl);
    // done via callback!
  }


  // compute the ILP formulation for a graph g
  template<class Graph>
  void populate_model(IloModel& model, IloEnv& env, var_collection<Graph>& vars){
    const Graph& g(*vars.I.g);
    IloRangeArray c(env);
    // construct variables, keeping note of the x_ij, y^l_ij and z^l_ij variables
    DEBUG3(std::cout << "creating variables"<<std::endl);
    populate_variables(env, vars);
    // add optimization function
    DEBUG3(std::cout << "creating optimization expression"<<std::endl);
    IloExpr opt_exp(env);
    for(EdgeIterRange<Graph> e = boost::edges(g); e.first != e.second; ++e.first){
      const Vertex<Graph>& u = boost::source(*e.first, g);
      const Vertex<Graph>& v = boost::target(*e.first, g);
      // get the appropriate x_ij variable
      auto uv = vars.z.find(std::make_pair(u,v));
      if(uv == vars.z.end()) uv = vars.z.find(std::make_pair(v,u));
      assert(uv != vars.z.end());
      // add the term weight(e)*x_e
      opt_exp += (double)boost::get(boost::weight, g, *e.first) * uv->second;
    }
    model.add(IloMaximize(env, opt_exp));
    // construct constraints
    DEBUG3(std::cout << "creating constraints"<<std::endl);
    populate_constraints(env, c, vars);
    model.add(c);
  }

}

#endif

