

#ifndef DEGREE_BRANCHING_HPP
#define DEGREE_BRANCHING_HPP

#include <list>

#include "utils/graph_typedefs.hpp"
#include "utils/scaffolding_typedefs.hpp"
#include "utils/instance.hpp"
#include "solv/trees.hpp"
#include "solv/branching.hpp"


namespace scaffold{
  namespace solv{


    // translate a branch using a translate map
    template<class Graph>
    void translate_branch(Branch<Graph>& branch, const VTranslateMap<Graph,Graph>& translate){
      for(auto e = branch.edges.begin(); e != branch.edges.end();){
        // Step 1: insert the translation of the edge b before e
        e = branch.edges.emplace(e, std::make_pair(translate.at(e->first), translate.at(e->second)));
        // Step 2: delete the original, advancing the iterator
        e = branch.edges.erase(++e);
      } // for
    } // function

    // store all branches around v in a list
    template<class Graph>
    void get_branches_for_vertex(const Graph& g, const Vertex<Graph>& v, std::list<Branch<Graph> >& branches){
      // for each edge uv incident to v, create the branch that takes uv
      for(AdjIterRange<Graph> r = adjacent_vertices(v, g); r.first != r.second; ++r.first)
        branches.push_back(Branch<Graph>(TakeBranch, std::make_pair(v, *r.first)));
    }// function

    // apply a branch b to an instance H
    template<class Graph>
    void apply_branch(Instance<Graph>& I, const Branch<Graph>& b, DynProgSolution<>* const S = NULL){
      switch(b.btype){
        case DeleteBranch:
          // the branch contains a list of edges to delete
          I.delete_edges(b.edges, true);
          break;
        case TakeBranch:
          // the branch contains a list of edges to take
          for(cosnt auto& e : b.edges) {
            DEBUG4(std::cout << "taking edge "<<VertexPairAndGraph<Graph>(e, *I.g)<<std::endl);
            // remove all around e.first and e.second except for e and the 2 matching edges incident to them
            I.clear_vertex_except(e.first, &e.second, true);
            I.clear_vertex_except(e.second, &e.first, true);
          } // for
          break;
      } // switch
    } // function

    // copy a given instance and apply the given branch to the copy
    template<class Graph>
    Instance<Graph>* copy_and_apply_branch(const Instance<Graph>& I, Branch<Graph>& b, DynProgSolution<>* const S = NULL){
      // Step 1: make a copy H of I, but don't update infos, as we will invalidate it anyways by deleting edges
      VTranslateMap<Graph,Graph> translate;
      Instance<Graph>* H = new Instance<Graph>(I, &translate, false);
      // Step 2: translate branch
      translate_branch(b, translate);
      // Step 3: apply the branch 
      apply_branch(*H, b, S);
      // Step 4: update H's infos
      H->infos.update_all();
      return H;
    }


    // recursively transform I into a solution, TODO: add upper and lower bounds
    template<class Graph>
    void deg_solve_recursively(Instance<Graph>& I, DynProgSolution<>*& S, size_t depth = 0){
      // Step 0: check for connected components
      Instance<Graph> *J = I.split_off_component();
      if(J){
        // Step 1c: if a component has been split off, then recurse for the split-off instance
        DEBUG3(std::cout << "splitting into components of size "<<num_vertices(*I.g)<<" and "<<num_vertices(*J->g)<<" (at "<<J<<")"<<std::endl);
        DynProgSolution<>* solJ = new DynProgSolution<>(*J);
        deg_solve_recursively(I, S, depth + 1);
        deg_solve_recursively(*J, solJ, depth + 1);
        delete J;

        DEBUG3(std::cout << "depth "<<depth<<": rejoining solutions for components"<<std::endl);

        if(!S->no_solutions() && !solJ->no_solutions()){
          // if both components have solutions, unite them and find a solution subset
          S->combine_disjoint_union(*solJ);
          delete solJ;
          return;
        } else {
          // if either component does not have a solution, return failure
          S->invalidate();
          delete solJ;
          return;
        }
      } 
      // Step 1: base cases
      DEBUG3(std::cout << "depth "<< depth<<": solving instance with " << num_vertices(*I.g) << " vertices:" << std::endl);
      DEBUG5(write_adj_list_graph(std::cout, *I.g); std::cout << std::endl);
      if(I.max_deg_two()) {
        DEBUG3(std::cout << "it's already max-deg 2, getting all sub-solutions"<<std::endl);
        get_all_solutions(I, S);
        return;
      }
      if(I.is_acyclic()){
        DEBUG3(std::cout << "it's a tree, getting all sub-solutions"<<std::endl);
        solve_on_tree(I, S);
        return;
      }
      // Step 2: preprocess
      // Step 3: find a good branching
      // Step 3a: get the maximum degree vertex v in I.g
      const VertexAndDegree<Graph> vd = get_max_degree(*I.g);
      const Vertex<Graph>& v = vd.first;
      DEBUG4(std::cout << "branching on vertex: "<< ( *I.g)[v].name<<std::endl);
      // Step 2b: get all branches for v
      std::list<Branch<Graph> > branches;
      get_branches_for_vertex(*I.g, v, branches);

      // Step 3: for each branch, make a copy of I, apply the branching, and recurse
      DynProgSolution<> max_sol(I);
      for(Branch<Graph> b : branches){
        DEBUG3(std::cout << "depth "<<depth<<":"<<std::endl<<" branching: "<< BranchAndGraph<Graph>(b, *I.g)<<std::endl);
        // Step 3a: get a copy H of I to which the branch has been applied
        DynProgSolution<>* solH = new DynProgSolution<>(I);
        Instance<Graph>* H = copy_and_apply_branch(I, b, solH);
        // Step 3b: recurse
        deg_solve_recursively(*H, solH, depth + 1);
        DEBUG4(std::cout << "depth "<<depth<<": returned with solution"<<std::endl);
        delete H;
        // update solution
        if(!solH->no_solutions()) max_sol.combine_min_of_two(*solH);
        delete solH;
      }// for
      if(max_sol.no_solutions()) S->invalidate(); else S->destructive_append(max_sol);
      DEBUG3(std::cout << "returning from depth "<<depth<<" with "; if(S->no_solutions()) std::cout << "no solutions"<<std::endl; else std::cout << "solutions: "<<*S<<std::endl;);
    } // function

    /*
    // create a list of graphs yielded by branching over high degree vertices, or an empty list if all degrees are 0
    template<class Graph>
    std::list<Graph*> deg_branch(const Graph& g){
      // Step 1: find the highest degree vertex
      const VertexAndDegree<Graph> vd = get_max_degree(g);      
      // Step 2: if the maximum degree is greater than 0, then branch on this vertex; otherwise, return an empty graphlist
      if(vd.second > 2)
        return deg_branch_on_vertex(g, vd.first);
      else
        return std::list<Graph*>();
    }// function
    */

  } // namespace
} // namespace

#endif
