
#ifndef DP_SOLUTION_HPP
#define DP_SOLUTION_HPP

#include <vector>
#include <list>
#include <iostream>

#include "utils/graph_typedefs.hpp"
#include "utils/min_matrix.hpp"
#include "utils/adv_iters.hpp"

namespace scaffold{ namespace solv{ namespace DP{


  // a solution is a matrix of weights: S[x,y] = minimum weight of a set X of edges such that G - X is a collection of x paths & y cycles
  // each entry S[x,y] also has a list of pairs of vertex names attached to it, which indicates which edges to delete to get to the solution
  // on top, there is a global modifier for weights and edge lists that applies to all entries
  template<class Compare = std::less<size_t> >
  class DynProgSolution : public PCMinMatrix<Compare> {
    using Parent = PCMinMatrix<Compare>;
  protected:
    // global weight & edges
    MatrixEntry global;
  public:
    // inherit constructors of PCMinMatrix
    using Parent::Parent;
    using Parent::entries;
    using Parent::max;
    using Parent::best_index;
    using Parent::smallest_index;
    using Parent::put;
    using typename Parent::entries_t;

    // construct
    DynProgSolution(const Parent& S): Parent(S) {}

    // return whether there are no solutions stored in the object
    bool no_solutions() const { return entries->empty(); }
    // invalidate all solutions stored in the object
    void invalidate() { entries->clear(); }

    // combine two solutions to a solution to the disjoint union of the graphs
    // NOTE: max_paths and max_cycles remain unchanged. Therefor, this operation is NOT SYMMETRIC
    void combine_disjoint_union(const DynProgSolution& S)
    {
      // combine all entries of S with all our entries and store the minima in a new matrix
      entries_t* old_entries = entries;
      entries = new entries_t();
      for(const auto& E : *old_entries){
        const PathsAndCycles pc = E.first;
        for(const auto& SE : *S.entries) put(pc + SE.first, *E.second, *SE.second);
      }// for
      // replace our matrix with the result
      delete old_entries;
      // finally, merge global modifiers
      global += S.global;
    } // function

    // combine two concurrent solutions by taking the min of the two for each entry.
    // NOTE: this requires copying the global solution into each of the local solutions since the globals are likely not shared between the two solutions
    void combine_min_of_two(const DynProgSolution& S){
      DEBUG4(std::cout << "combining solutions"<<std::endl);
      // Step 1: merge global modifier into entries if necessary
      if(global.weight || !global.has_no_edges()){
        for(auto& E : *entries){
          const PathsAndCycles pc = E.first;
          const typename entries_t::const_iterator SE = S.entries->find(pc);
          // if S has an entry for pc and its better than our own, then replace our own by it, otherwise add global modifier to E
          if((SE == S.entries->end()) || !put(pc, *SE->second, S.global)) *E.second += global;
        }// for
        // Step 2: clear global modifier
        global.clear();
      }
      // Step 2: go through S and update our entries with theirs if necessary
      for(const auto& SE : *(S.entries)) put(SE.first, *SE.second, S.global);
    }// function


    // append the global and all local solutions of S onto the global solution of us, clearing S' entries in the process
    void destructive_append(DynProgSolution& S){
      assert(entries->size() == 0);
      // Step 1: kill our (empty) entries
      delete entries;
      // Step 2: add global modifier of S
      global += S.global;
      // Step 3: make the local entries of S our own & update best_index
      entries = S.entries;
      best_index = S.best_index;
      // Step 4: give S new entries
      S.entries = new entries_t();
    }

    // add edge to solution
    void add_edge_global(const VertexName& u, const VertexName& v, const size_t weight){
      DEBUG4(std::cout << "globally adding ("<<u<<","<<v<<") to solution"<<std::endl);
      global.append_edge(EdgeName(u,v), weight);
    }


    // << operator for solutions outputting a table of weights
    friend std::ostream& operator<<(std::ostream& os, const DynProgSolution<Compare>& S){
      return os << "globally +"<<S.global.weight<<std::endl << (Parent)(S);
    } // function


  };

  // given an instance that is connected and deg-2, add it as path solution or add the cycle and the cheapest subpath as solutions
  template<class Compare = std::less<unsigned> >
  void treat_deg_two(Instance& I, DynProgSolution<Compare>* const S, const Compare& is_better = Compare())
  {
    // get the number of paths and cycles in I.g
    const RawScaffoldGraph& g = I.get_graph();
    const unsigned n = boost::num_vertices(g);
    const unsigned m = boost::num_edges(g);
    const bool is_cycle = (n == m);
    DEBUG4(std::cout << "instance has "<<n<<" vertices, "<<m<<" edges. Requested: "<<I.num_paths<<" & "<<I.num_cycles<<std::endl);
    // if it's a cycle, add the cycle solution & find the cheapest subpath to also add
    if(is_cycle){
      if(I.num_cycles) S->put_all(PathsAndCycles(0,1), g);
      if(I.num_paths){
        // delete an edge of the cycle to turn it into a path
        const std::pair<ScafVertexPair, unsigned> nb = I.get_infos().bridges.get_cheapest_nonbridge();
        const EdgeName nb_name(g[nb.first.first].name, g[nb.first.second].name);

        if(is_better(0,1)){
          // if we are looking for a MINIMUM solution, then put the cheapest nonbridge & its weight at (1,0)
          S->put(PathsAndCycles(1,0), nb_name, nb.second);
        } else {
          // if we are looking for a MAXIMUM solution, then put everything but the non-bridge
          MatrixEntry* ME = new MatrixEntry(0);
          for(auto e = I.get_edges(); e; ++e)
            if(!is_equal(nb.first, *e, g))
              ME->append_edge(I.get_edge_name(*e), nb.second);
          if(!S->put(PathsAndCycles(1,0), ME)) delete ME;
        }
      }// if
    } if(I.num_paths) S->put_all(PathsAndCycles(1,0), g);
  }// function


}}} // namespace


#endif
