

#ifndef DP_TABLE_HPP
#define DP_TABLE_HPP


#include "utils/utils.hpp"
#include "utils/min_matrix.hpp"
#include "utils/graph_typedefs.hpp"
#include "utils/permissive_matching.hpp"
#include "utils/scaffold_graph.hpp"
#include "solv/dp_solution.hpp"

namespace scaffold{ namespace solv{ namespace DP {

  using DegMap = std::bitset<MAX_TW>;

  DegMap permute_degmap(const DegMap& d, const BytePermutation& perm)
  {
    DegMap result;
    for(byte i = 0; i < MAX_TW; ++i)
      if(d.test(i)) result.set(perm[i]);
    return result;
  }

  // check whether (d1,M1) and (d2,M2) are compatible
  inline bool compatible(const DegMap& d1, const DegMap& M1, const DegMap& d2, const DegMap& M2){
    return ((d1 ^ M1) & (d2 ^ M2)).none();
  }


  // the real table consists of 3 levels:
  // level 1: map p,c to the solution using a PCMinMatrix
  typedef PCMinMatrix<std::greater<size_t> > pc_to_sol;

  // level 2: map P to a pc_to_sol matrix
  class P_to_pc_map;
  
  // level 3: map d to a P_to_pc_map
  class d_to_P_map;

  // modify a permissive matching according to the insertion of edge uv
  // return the number of paths and cycles that were closed by adding uv
  PathsAndCycles modify_P_with_uv(PermissiveMatching& P, const byte u_idx, const byte du, const byte v_idx, const byte dv)
  {
    PathsAndCycles result;
    // Case 2: d[u] = d[v] = 1:
    // there is only the edge uv, so add uv to P
    if((du == 1) && (dv == 1)){
      // add uv to P and add the weight of uv to the weight of this entry; if this is impossible, delete this entry
      assert(P[u_idx] == u_idx);
      assert(P[v_idx] == v_idx);
      P[u_idx] = v_idx;
      P[v_idx] = u_idx;
    }
    // Case 3: d[u] = 2, d[v] = 1:
    // there is a path x-u-v with x != v
    if((du == 2) && (dv == 1)){
      const byte x_idx = P[u_idx];
      if(x_idx != u_idx){
        assert(x_idx != v_idx);
        // if x is deg-1, then extend xu to xv in P
        P[v_idx] = x_idx;
        P[x_idx] = v_idx;
        // and clear P[u]
        P[u_idx] = u_idx;
      }
    }
    // Case 1: d[u] = 2, d[v] = 2
    // there is a path p ending in u and a path q ending in v that we have to combine by inserting uv
    if((du == 2) && (dv == 2)){
      // get the vertices x and y that are paired with u and v, respectively, by P
      const byte x_idx = P[u_idx];
      const byte y_idx = P[v_idx];

      // TODO: optimize cases to reduce "if" checks
      
      // Case 1a: x = v & y = u, then adding uv closes a cycle
      if(x_idx == v_idx){
        assert(y_idx == u_idx);
        result.c++;
      }

      // Case 1c1: u is unpaired & v is paired
      if((x_idx == u_idx) && (y_idx != v_idx)) P[y_idx] = y_idx;

      // Case 1c2: v is unpaired & u is paired
      if((x_idx != u_idx) && (y_idx == v_idx)) P[x_idx] = x_idx;
     
      // Case 1d: x and y are different vertices with d[x] = d[y] = 1
      if((x_idx != u_idx) && (x_idx != v_idx) && (y_idx != u_idx) && (y_idx != v_idx)){
        P[x_idx] = y_idx;
        P[y_idx] = x_idx;
      }

      // Case 1b: u and v are unpaired
      if((x_idx == u_idx) && (y_idx == v_idx)){
        result.p++;
      } else {
        P[u_idx] = u_idx;
        P[v_idx] = v_idx;
      }
    }// if
    DEBUG4(std::cout << "new P = "<<P<<"     and we closed "<<result<<" paths and cycles"<<std::endl);
    return result;
  }// function

  // level 2: map P to a pc_to_sol matrix
  class P_to_pc_map : public unordered_map<PermissiveMatching, pc_to_sol*>
  {
    using Parent = unordered_map<PermissiveMatching, pc_to_sol*>;
  public:
    using Parent::iterator;
    using Parent::const_iterator;
    using Parent::erase;
    using Parent::empty;
    using Parent::emplace;
    using Parent::begin;
    using Parent::end;
    using Parent::find;

    // forget a vertex in all permissive matchings of the map
    P_to_pc_map* forget_vertex(const byte v_idx)
    {
      P_to_pc_map* const new_map = new P_to_pc_map();
      
      for(iterator old_i = begin(); old_i != end(); old_i = erase(old_i)){
        const PermissiveMatching& P = old_i->first;
        DEBUG5(std::cout << "constructing a new map @"<<new_map<<" by forgetting "<<(unsigned)v_idx<<" from P = "<<P<<std::endl);
        // look where v is mapped in P
        const byte u_idx = P[v_idx];
        if(v_idx != u_idx){
          // if v is mapped to some u != v, then create an entry where uv is replaced by uu
          PermissiveMatching new_P(P);
          new_P[u_idx] = u_idx;
          new_P[v_idx] = v_idx;

          new_map->merge(new_P, old_i->second);
        } else {
          DEBUG5(std::cout << "forgetting "<<(unsigned)v_idx<<" closes a path in matrix @"<<old_i->second<<std::endl);
          // if v is mapped to itself, then increase p in all pc entries
          old_i->second->add_to_each_key(PathsAndCycles(1,0));
          new_map->merge(P, old_i->second);
        }// if
      }// for
      return new_map;
    }// function

    // create a new P_to_pc_map that is the result of introducing an edge into us
    P_to_pc_map* intro_edge(const byte u_idx, const byte v_idx, const byte du, const byte dv, const EdgeName& en, const size_t weight)
    {
      P_to_pc_map* const new_map = new P_to_pc_map();

      for(iterator old_i = begin(); old_i != end(); old_i = erase(old_i)){
        const PermissiveMatching& old_P = old_i->first;
        pc_to_sol* sol_matrix = old_i->second;

        // get P and modify it according to the degrees of u and v
        PermissiveMatching P(old_P);
        const PathsAndCycles closed_pc = modify_P_with_uv(P, u_idx, du, v_idx, dv);
        DEBUG5(std::cout << "modified P:"<<std::endl<<old_P<<std::endl<<P<<" closing "<<closed_pc<<" pc's"<<std::endl);
        // add the new edge to the matrix
        sol_matrix->add_to_each_key(closed_pc);
        sol_matrix->add_to_all(en, weight);

        // merge the result into the new map
        new_map->merge(P, sol_matrix);
      }// for
      return new_map;
    }// function

    // create a new P_to_pc_map that is the join of us and T
    P_to_pc_map* join(const P_to_pc_map& T, const DegMap& new_deg2) const
    {
      P_to_pc_map* new_map = new P_to_pc_map();
      // go through our and T's permissive matchings and try to join them up
      for(const_iterator old_i = begin(); old_i != end(); ++old_i)
        for(auto& T_Ppc: T){
          const pc_to_sol* const my_matrix = old_i->second;
          // merge the permissive matchings (should always be possible), getting the number of closed paths and cycles
          PermissiveMatching P(old_i->first);
          const PermissiveMatching& T_P(T_Ppc.first);
          const pc_to_sol* const T_matrix = T_Ppc.second;

          DEBUG5(std::cout << "merging:"<<std::endl<<old_i->first<<" & "<<std::endl<<T_P<<" to "<<std::endl);
          const PathsAndCycles closed_pc(P.merge_with(T_P, new_deg2));
          DEBUG5(std::cout<<P<<std::endl);

          
          // prepare solution matrix to hold the combination of matrices
          pc_to_sol* pc = new pc_to_sol(my_matrix->max);
          // go through our matrix and T's matrix and mix them with the paths & cycles created in the join of the matchings
          for(auto my_pc = my_matrix->get_entries(); my_pc.first != my_pc.second; ++my_pc.first)
            for(auto T_pc = T_matrix->get_entries(); T_pc.first != T_pc.second; ++T_pc.first){
              DEBUG5(std::cout << "joining:"<< my_pc.first->first<<" & " <<  T_pc.first->first<<" --> "<< my_pc.first->first + T_pc.first->first + closed_pc<<std::endl);
              MatrixEntry* ME = new MatrixEntry(*my_pc.first->second + *T_pc.first->second);
              if(!pc->put(my_pc.first->first + T_pc.first->first + closed_pc,ME)) delete ME;
            }// for
          // merge the created entry into the container
          new_map->merge(P, pc);
        }// for
      return new_map;
    }// function

    // merge a new entry into a P_to_pc_maps
    void merge(const PermissiveMatching& P, pc_to_sol* const pc)
    {
      if(pc->size()){
        DEBUG5(std::cout << "merging "<<std::endl<<*pc<<" into "<<P<<std::endl);
        iterator i = find(P);
        if(i != end()){
          DEBUG6(std::cout << "merging the matrix "<<std::endl<<*pc<<" with the present entry"<<std::endl<<*i->second<<" for P = "<<P<<std::endl);
          // if there is already an entry for P in m, then merge the matrices
          pc_to_sol* present_pc = i->second;
          present_pc->destructive_merge(*pc);
          delete pc;
        } else {
          DEBUG6(std::cout << "adding the matrix @"<<pc<<std::flush<<" (size "<<pc->size()<<") to P = "<<P<<std::endl);
          emplace(P, pc);
        }
      } else {
        DEBUG5(std::cout << "not merging now-empty matrix "<<std::endl);
        delete pc;
      }
    }// function

    // merge a P_to_pc_map m, destroying it
    void destructive_merge(P_to_pc_map& m)
    {
      DEBUG6(std::cout << "destructively merging "<<m.size()<<" matrices"<<std::endl);
      for(iterator i = m.begin(); i != m.end(); i = m.erase(i)) {
        merge(i->first, i->second);
      }
    }

    // copy a P_to_pc_map to another
    void make_copy(P_to_pc_map& target)
    {
      for(iterator i = begin(); i != end(); ++i){
        DEBUG5(std::cout << "copying matrix with P = "<<i->first<<" @"<<i->second<<" (size "<<i->second->size()<<")"<<std::endl);
        // create a new matrix as copy of the old one
        pc_to_sol* pc = new pc_to_sol(*i->second);
        // add it with the same permissive matching as in source
        target.emplace(i->first, pc);
      }
    }

    // apply the BytePermutation to each P, destroying the map
    P_to_pc_map* apply_permutation(const BytePermutation& perm)
    {
      P_to_pc_map* new_map = new P_to_pc_map();

      for(iterator i = begin(); i != end(); i = erase(i)){
        PermissiveMatching P(i->first);
        P.permute(perm);
        new_map->emplace(P, i->second);
      }
      return new_map;
    }


    friend std::ostream& operator<<(std::ostream& os, const P_to_pc_map& M)
    {
      for(auto i  = M.begin(); i != M.end(); ++i) os <<"\t\t\t\t"<<i->first<<":"<<std::endl<< *i->second << std::endl;
      return os;
    }

  };


  // level 3: map d to a P_to_pc_map
  class d_to_P_map : public unordered_map<DegMap, P_to_pc_map*, std::hash<DegMap> >
  {
  public:
    using unordered_map<DegMap, P_to_pc_map*, std::hash<DegMap> >::iterator;
    using unordered_map<DegMap, P_to_pc_map*, std::hash<DegMap> >::erase;
    using unordered_map<DegMap, P_to_pc_map*, std::hash<DegMap> >::emplace;
    using unordered_map<DegMap, P_to_pc_map*, std::hash<DegMap> >::begin;
    using unordered_map<DegMap, P_to_pc_map*, std::hash<DegMap> >::end;
    using unordered_map<DegMap, P_to_pc_map*, std::hash<DegMap> >::find;
    // forget a vertex in all permissive matchings of the map
    d_to_P_map* forget_vertex(const byte v_idx)
    {
      d_to_P_map* const new_map = new d_to_P_map();

      for(iterator old_i = begin(); old_i != end(); old_i = erase(old_i)){
        const DegMap& d = old_i->first;
        // stuff happens only if degree of v is one
        if(d.test(v_idx)){
          DegMap new_d = d;
          new_d.reset(v_idx);

          DEBUG4(std::cout << "forgetting "<<(unsigned)v_idx<<" from entry with d = "<<d<<" (new d = "<<new_d<<")"<<std::endl);

          P_to_pc_map* old_Ppc_map = old_i->second;
          P_to_pc_map* new_Ppc_map = old_Ppc_map->forget_vertex(v_idx);
          // exchange the old map for the new map
          delete old_Ppc_map;
          old_i->second = NULL;

          // merge the new d entry into the new table
          new_map->merge(new_d, new_Ppc_map);
        } else new_map->merge(d, old_i->second);
      }// for
      return new_map;
    }// function

    // add an edge to all permissive matchings and the matrices of the map
    d_to_P_map* intro_edge(byte u_idx, byte v_idx, const bool mu, const bool mv, const bool mu_old, const bool mv_old, const bool is_matching_edge, const EdgeName& en, const size_t weight)
    {
      d_to_P_map* const new_map = new d_to_P_map();

      // Step 1: if uv is not a matching edge, then copy the old entries into the new table, corresponding to not taking the edge uv
      if(!is_matching_edge){
        DEBUG4(std::cout << "making a normal copy of the table to represent not-taking the edge ("<<en.first<<","<<en.second<<")"<<std::endl);
        // TODO: don't make a copy if we know that the old entry will be thrown away anyways
        make_copy(*new_map);
      }

      // Step 2: add the edge to each entry and merge the results into the new table
      for(iterator old_i = begin(); old_i != end(); old_i = erase(old_i)){
        const DegMap& d = old_i->first;

        // to take the edge, make sure that: if it's not a matching edge, then the non-matching degree of u and v are 0 
        if(is_matching_edge || ((mu_old == d.test(u_idx)) && (mv_old == d.test(v_idx)))){
          DEBUG4(std::cout << "taking "<<(is_matching_edge?"M-":"  ")<<"edge ("<<en.first<<","<<en.second<<") with indices ("<<(unsigned)u_idx<<","<<(unsigned)v_idx<<") & old M-values ("<<mu_old<<","<<mv_old<<") for d = "<<d<<std::endl);
          DegMap new_d = d;
          // Step 2a: increase both d[u] and d[v]
          new_d.flip(u_idx);
          new_d.flip(v_idx);

          byte du = mu + (new_d.test(u_idx) != mu);
          byte dv = mv + (new_d.test(v_idx) != mv);
          // assert d[u] >= d[v]
          if(du < dv) {
            std::swap(u_idx, v_idx);
            std::swap(du, dv);
          }

          // Step 2b: introduce the edge into the Permissive Matchings and Matrices
          P_to_pc_map* old_Ppc = old_i->second;
          P_to_pc_map* new_Ppc = old_Ppc->intro_edge(u_idx, v_idx, du, dv, en, weight);

          delete old_Ppc;
          old_i->second = NULL;
          // Step 2c: merge the new entry into the table
          new_map->merge(new_d, new_Ppc);
        }// if
      }// for
      return new_map;
    }// function

    d_to_P_map* join(const d_to_P_map& T, const DegMap& old_M, const DegMap& T_M)
    {
      d_to_P_map* result = new d_to_P_map();
      // go through our table...
      for(iterator old_i = begin(); old_i != end(); old_i = erase(old_i)){
        // save our degrees
        const DegMap& my_d = old_i->first;
        P_to_pc_map* const old_Ppc = old_i->second;
        // go through T's table
        for(auto& T_dPpc: T){
          // save T's degrees
          const DegMap& T_d = T_dPpc.first;
          // check if the degrees are compatible
          if(compatible(my_d, old_M, T_d, T_M)){
            const P_to_pc_map& T_Ppc = *T_dPpc.second;
            DEBUG5(std::cout << "--> merging d's:"<<std::endl<<my_d<<" ("<<old_Ppc->size()<<" P's) & "<< std::endl<<T_d<<" ("<<T_Ppc.size() <<" P's)"<<std::endl);
            // save the set of vertices who became degree-2 in the join
            const DegMap& new_deg2 = my_d & T_d;
            // produce the new degrees as the sum of the old ones
            DegMap new_d = my_d ^ T_d;
            // prepare the container to hold the merge of the P_to_pc_maps corresponding to my_d and T_d
            P_to_pc_map* new_Ppc = old_Ppc->join(T_Ppc, new_deg2);
            // merge the combined P_to_pc_maps into the new table
            result->merge( new_d, new_Ppc);
          }// if
        }// for (T's table)
        // once we combined old_Ppc with all P entries of T, deep-delete it
        for(auto i = old_Ppc->begin(); i != old_Ppc->end(); ++i) delete i->second;
        delete old_Ppc;
        old_i->second = NULL;
      }// for (our table)
      return result;
    }

    // merge a new entry into a d_to_P_map
    void merge(const DegMap& d, P_to_pc_map* const Ppc)
    {
      if(!Ppc->empty()){
        iterator i = find(d);
        if(i != end()){
          DEBUG6(std::cout << "merging the map @"<<Ppc<<" with the present entry (@"<<i->second<<") for d = "<<d<<std::endl);
          // if there is already an entry for d in m, then merge the maps
          P_to_pc_map* present_Ppc = i->second;
          present_Ppc->destructive_merge(*Ppc);
          delete Ppc;
        } else emplace(d, Ppc);
      } else {
        DEBUG5(std::cout << "not merging empty P_to_pc map into d = "<<d<<std::endl);
        delete Ppc;
      }
    }// function

    // deep-copy a d_to_P_map to another
    void make_copy(d_to_P_map& target)
    {
      for(iterator i = begin(); i != end(); ++i){
        DEBUG5(std::cout << "deep-copying entry with d = "<<i->first<<" & map @"<<i->second<<std::endl);
        // create a new P_to_pc_map as copy of the old one
        P_to_pc_map*  Ppc = new P_to_pc_map();
        i->second->make_copy(*Ppc);
        // add it with the same DegMap as in source
        target.emplace(i->first, Ppc);
      }
    }

    // apply a BytePermutation to each d and each P, destroying the current map
    d_to_P_map* apply_permutation(const BytePermutation& perm)
    {
      d_to_P_map* new_map = new d_to_P_map();

      for(iterator i = begin(); i != end(); i = erase(i)){
        DegMap d = permute_degmap(i->first, perm);
        P_to_pc_map* new_Ppc = i->second->apply_permutation(perm);

        // note: we can use emplace here, because no conflicts can arise from applying a permutation
        new_map->emplace(d, new_Ppc);
      }// for
      return new_map;
    }


    friend std::ostream& operator<<(std::ostream& os, const d_to_P_map& M)
    {
      for(auto i = M.begin(); i != M.end(); ++i) os <<"\t\t"<<i->first<<":"<<std::endl<< *i->second << std::endl;
      return os;
    }
  
  };









  class DP_Table
  {
    const Instance& I;
  public:
    // index of vertices used; TODO: use an array instead of the unordered_set_of<byte>
    typedef boost::bimap< boost::bimaps::unordered_set_of<ScafVertex >, boost::bimaps::unordered_set_of<byte> > VertexIndex;
    VertexIndex* index;

    // a bitset of vertices that are incident to edges of M that have been introduced in this subtree
    DegMap M_adj;

    typedef d_to_P_map DP_Entries;
    DP_Entries* table;

    // contstructing the index from the vertex set
    void construct_index(const ScafVertexSet& VSet)
    {
      for(auto& v : VSet) intro_vertex(v);
    }

  public:
    // constructors
    DP_Table(const Instance& _I, const ScafVertexSet& VSet):
      I(_I),
      index(new VertexIndex()), table(new DP_Entries())
    {
      construct_index(VSet);
    }
    // given a TOL Vertexset, convert it to a set of ScafVertex's
    DP_Table(const Instance& _I, const TOL::GraphVertexSet& VSet, VTranslate* vtranslate):
      I(_I),
      index(new VertexIndex()), table(new DP_Entries())
    {
      ScafVertexSet good_vset;
      for(auto& v: VSet) good_vset.emplace(vtranslate->right.at(v));
      construct_index(good_vset);
    }

    // destructor
    ~DP_Table(){ delete table; delete index; }


    // clear the table, freeing all memory helt
    void clear()
    {
      for(auto i = table->begin(); i != table->end(); ++i){
        for(auto j = i->second->begin(); j != i->second->end(); ++j) delete j->second;
        i->second->clear();
        delete i->second;
      }
      table->clear();
    }

    // return whether this is a fresh table, without content
    bool empty() const
    {
      return table->empty();
    }

    void get_VSet(std::list<ScafVertex >& l) const
    {
      for(auto& i: index->left) l.emplace_back(i.first);
    }


    // ==========================
    // ===== DP OPERATIONS ======
    // ==========================
    // always take care of: index, free_indices, g, M_adj and the table

    // introduce a vertex
    void intro_vertex(const ScafVertex& v)
    {
      DEBUG3(std::cout << "introducing new vertex "<< I[v].name<<std::endl);
      // index & free_indices
      // try to insert the vertex at name % MAX_TW; if vertices don't have names use ((size_t)v % MAX_TW)
      byte v_idx = I[v].index % MAX_TW;
      // if there is already such an index, find the next free index
      while(contains(index->right, v_idx)) v_idx = (v_idx + 1) % MAX_TW;
      index->left.insert(std::make_pair(v, v_idx));
      // g
      //g.introduce_vertex(v);
      // M_adj & table do not change since no edges are incident with v yet
    }
    void intro_vertices(const ScafVertexSet& V)
    {
      for(auto& v : V) intro_vertex(v);
    }


    // forget a vertex
    void forget_vertex_from_table(const byte& v_idx)
    {
      DP_Entries* const old_table = table;
      table = old_table->forget_vertex(v_idx);
      delete old_table;
      DEBUG5(std::cout << "table after forgetting "<<(unsigned)v_idx<<":"<<std::endl<<*table<<std::endl);
    }// function
    void forget_vertex(const ScafVertex& v)
    {
      // index
      const auto i = index->left.find(v);
      assert(i != index->left.end());
      const byte v_idx = i->second;
      DEBUG5(std::cout<< "forgetting vertex["<<(unsigned)v_idx<<"] = "<< std::flush<< I[index->right.at(v_idx)].name<<std::endl);
      index->left.erase(i);

      // g
      //g.forget_vertex(v);

      forget_vertex_from_table(v_idx);
      // M_adj
      M_adj.reset(v_idx);
    }
    void forget_vertices(const ScafVertexSet& V)
    {
      for(auto& v : V) forget_vertex(v);
    }


    // introduce a set of edges & set M_adj, remember d = degree % 2
    void intro_edge_to_table(byte u_idx, byte v_idx, const bool is_matching_edge, const EdgeName& en, const size_t weight)
    {
      bool mu, mv;
      bool mu_old = M_adj.test(u_idx);
      bool mv_old = M_adj.test(v_idx);

      if(is_matching_edge) {
        assert(!mu_old && !mv_old);
        M_adj.set(u_idx);
        M_adj.set(v_idx);
        mu = mv = true;
      } else {
        mu = mu_old;
        mv = mv_old;
      }
      DEBUG4(std::cout << "new M = "<<M_adj<<std::endl);

      // setup a new table
      DP_Entries* old_table = table;
      table = old_table->intro_edge(u_idx, v_idx, mu, mv, mu_old, mv_old, is_matching_edge, en, weight);
      delete old_table;
    }// function

    void intro_edge(const ScafEdge& e)
    {
      DEBUG4(std::cout << "introducing edge "<< I.get_edge_name(e) <<std::endl);
      // index & free_indices don't change
           
      const bool is_in_M = I.is_matching_edge(e);
      const ScafVertex u = I.source(e);
      const ScafVertex v = I.target(e);
      const auto iu = index->left.find(u);
      const auto iv = index->left.find(v);
      assert(iu != index->left.end());
      assert(iv != index->left.end());
      const byte u_idx = iu->second;
      const byte v_idx = iv->second;

      // M_adj & table
      intro_edge_to_table(u_idx, v_idx, is_in_M, EdgeName(I[u].name, I[v].name) , I[e].weight);
    }
    void intro_edges(const std::list<ScafEdge >& E)
    {
      for(auto& e : E) intro_edge(e);
    }


    // make our indices compatible with the indices of T, that is, reorder our indices such that all our vertices have the same index here as in T
    void unify_indices(const DP_Table& T)
    {
      BytePermutation perm;
      // Step 1: compute the index permutation
      for(const auto& i : index->left){
        const ScafVertex& v = i.first;
        const byte v_idx = i.second;

        perm[v_idx] = T.index->left.at(v);
      }
      if(!perm.is_identity()){
        DEBUG3(std::cout << "remapping indices with permutation "<<perm<<std::endl);
        // Step 2: apply the permutation to the index
        VertexIndex* new_index = new VertexIndex;
        for(const auto& i : index->left)
          new_index->insert(typename VertexIndex::value_type(i.first, perm[i.second]));  
        delete index;
        index = new_index;
        // Step 3: remap M
        M_adj = permute_degmap(M_adj, perm);
        // Step 4: apply the permutation to the table
        DP_Entries* new_table = table->apply_permutation(perm);
        table = new_table;
      } else DEBUG3(std::cout << "rejoice, the indices are identical, saving the remap!"<<std::endl);
    }


    // join table with child table
    void join(DP_Table& T)
    {
      // TODO: test whether an index array for index-translation is faster than unifying the indices
      // TODO: try to insert a vertex each time at the same index, so that tables are inheritently compatible!
      // make indices compatible
      if(index->size() < T.index->size()) unify_indices(T); else T.unify_indices(*this);
      // unite the subgraphs
      //g.join(T.g);
      // save the old table to work with
      DP_Entries* old_table = table;
      // if we don't have table, then just steal the table of T and its M_adj
      if(table->empty()){
        DEBUG4(std::cout << "--> stealing size-"<<T.table->size()<<" table from child"<<std::endl);
        delete table;
        table = T.table;
        T.table = new DP_Entries();

        M_adj = T.M_adj;
        DEBUG4(std::cout << "--> new M = "<<M_adj<<std::endl);
      } else {
        DegMap old_M = M_adj;
        // wipe the table clean & merge matchings

        // update the matching
        M_adj |= T.M_adj;
        DEBUG4(std::cout << "--> new M = "<<M_adj<<std::endl);

        table = old_table->join(*T.table, old_M, T.M_adj);
        delete old_table;
      }// if
    }// function

    // initialize this table as a leaf-table
    void init_as_leaf()
    {
      DEBUG5(std::cout << "--> initializing leaf"<<std::endl);
      typename d_to_P_map::iterator dPpc_i = table->emplace(DegMap(), new P_to_pc_map()).first;
      typename P_to_pc_map::iterator Ppc_i = dPpc_i->second->emplace(PermissiveMatching(), new pc_to_sol(I.num_paths, I.num_cycles)).first;
      pc_to_sol& leaf_matrix = *Ppc_i->second;
      leaf_matrix.put_empty(PathsAndCycles(0,0));
    }

    // return the size in MatrixEntries
    size_t size() const
    {
      size_t result = 0;
      for(auto& i : *table)
        for(auto& j : *i.second)
          result += j.second->size();
      return result;
    }


    // read a solution from the table
    void read_solution(DynProgSolution<std::greater<size_t> >& S){
      for(auto& dPpc: *table)
        for(auto& Ppc: *dPpc.second)
          for(auto pc = Ppc.second->get_entries(); pc.first != pc.second; ++pc.first)
            S.put(pc.first->first, pc.first->second);
    }

    friend std::ostream& operator<<(std::ostream& os, const DP_Table& T)
    {
      for(typename DP_Entries::iterator i = T.table->begin(); i != T.table->end(); ++i) os <<i->first<<":"<<std::endl<< *i->second << std::endl;
      return os;
    }
  }; // class DP_Table

}}}// namespace 


#endif

