

#pragma once

#include <string>
#include <vector>

#ifdef STATISTICS
#include "utils/stats.hpp"
#endif

#include "utils/scaffold_graph.hpp"
#include "utils/profiling.hpp"
#include "utils/ilp_options.hpp"
#include "utils/ilp_common.hpp"
#include "utils/ilp_contig_jump_network.hpp"

#define VAR_COLLECTION var_collection_multi

namespace scaffold{ namespace solv {



  // block isolated cycles as lazy constraint callback
  class LazyCyclesBlockCallback: public IloCplex::LazyConstraintCallbackI{

    struct VertexProp {
      unsigned layer;
      const ScafVertex& orig;

      VertexProp(const ScafVertex& _orig, const unsigned _layer = 0): layer(_layer), orig(_orig) {}
    };
    typedef AuxiliaryGraph<VertexProp, no_property> AuxGraph;
    typedef std::vector<VTranslateMap<RawScaffoldGraph, AuxGraph> > TranslateMatrix;

    const VAR_COLLECTION& vars;
    CallbackStatistics& stats;

  public:

    // build the auxiliary graph using the var collection vars 
    void build_auxgraph(AuxGraph& aux, TranslateMatrix& translate_layer)
    {
      const Instance& I = vars.I;

      // construct the auxiliary graph from the current solution
      for(const auto& y : vars.times_used) if(is_true(getValue(y.second))){
        const ScafVertexPair& uv = y.first;
        const ScafVertex& u = uv.first;
        const ScafVertex& v = uv.second;
        const unsigned layer = I.is_matching_edge({u,v}) ? 0 : 1;


        // add vertices: first see if they're already there & if they are not there, add them to aux
        auto u_it = translate_layer.at(layer).find(u);
        if(u_it == translate_layer.at(layer).end())
          u_it = translate_layer.at(layer).insert(u_it, {u, boost::add_vertex(VertexProp(u, layer), aux)});

        auto v_it = translate_layer.at(1-layer).find(v);
        if(v_it == translate_layer.at(1-layer).end())
          v_it = translate_layer.at(1-layer).insert(v_it, {v, boost::add_vertex(VertexProp(v, 1-layer), aux)});
        
        // add edge
        const Vertex<AuxGraph>& aux_u = u_it->second;
        const Vertex<AuxGraph>& aux_v = v_it->second;
        boost::add_edge(aux_u, aux_v, aux);
      }// for
    }// function

    // return the startpoints of al paths & cycles
    void compute_startpoints(const TranslateMatrix& translate_layer, VertexSet<AuxGraph>& startpoints)
    {
      DEBUG5(std::cout << "reachable from s: ");
      for(const auto& y_su : vars.from_source) if(is_true(getValue(y_su.second))) {
        const ScafVertex& u = y_su.first;
        DEBUG5(std::cout << vars.I[u].name <<" ");
        // note that CPLEX might give us a solution that has some s->u->t_c paths; if, in this case, u does not have an edge in aux, we can't translate
        auto u_iter = translate_layer.at(0).find(u);
        if(u_iter != translate_layer.at(0).end()) startpoints.insert(u_iter->second);
      }// if
      DEBUG5(std::cout << "."<< std::endl);
    }

    unsigned forbid_cycles_in_y()
    {
      const Instance& I = vars.I;
      AuxGraph aux;
      TranslateMatrix translate(2);
      VertexSet<AuxGraph> startpoints;

      // Step 1: build the auxiliary graph & remove everything that is reachable from s
      DEBUG5(std::cout << "building auxiliary graph..."<<std::endl);
      build_auxgraph(aux, translate);
      DEBUG5(std::cout << "built aux graph with "<<boost::num_vertices(aux)<<" vertices & "<<boost::num_edges(aux)<<" edges"<<std::endl);

      DEBUG5(std::cout << "removing valid paths/cycles from auxiliary graph..."<<std::endl);
      // parse & remove all paths & cycles incident with pending
      compute_startpoints(translate, startpoints);
      remove_reachable(aux, startpoints, dir_fwd);
      DEBUG5(std::cout << "after cleanup, "<<boost::num_vertices(aux)<<" vertices & "<<boost::num_edges(aux)<<" edges remain"<<std::endl);

      // whatever remains of aux are cycles that are not attached to s, that is, invalid cycles;
      // Step 2: forbid them by adding a constraint forcing at least one arc going out of the cycle

      unsigned num_new_cuts = 0;
      while(boost::num_edges(aux) != 0){
        IloExpr expr(getEnv()), rev_expr(getEnv());
        VertexSet<AuxGraph> S;

        DEBUG5(std::cout << "forbidding cyclic structure: ");
        // find a set S of vertices in aux such that aux[S] is a cycle; write down the maximum multiplicity within S
        MultiType max_multiplicity = 1;
        const Edge<AuxGraph> e = *(boost::edges(aux).first);
        const Vertex<AuxGraph> x = boost::source(e, aux);
        VertexQueue<AuxGraph> pending;
        pending.push(x);
        S.insert(x);
        while(! pending.empty()){
          const Vertex<AuxGraph> aux_u = pending.front(); pending.pop();
          const VertexProp& u_prop = aux[aux_u];
          const ScafVertex& u = u_prop.orig;
          for(OEdgeIterRange<AuxGraph> r = boost::out_edges(aux_u, aux); r.first != r.second;){
            const Edge<AuxGraph> aux_uv = *r.first;
            const Vertex<AuxGraph> aux_v = boost::target(aux_uv, aux);
            const ScafVertex& v = aux[aux_v].orig;
            const ScafVertexPair uv(u,v);
            const ScafVertexPair vu(v,u);
            const ScafEdge e = I.find_edge(u, v).first;
            max_multiplicity = std::max(max_multiplicity, I[e].multiplicity);
            
            DEBUG5(std::cout << " " << vars.I.get_edge_name(uv));

            // modify the expressions for the edge uv
            expr -= vars.times_used.at(uv);
            rev_expr -= vars.times_used.at(vu);

            // delete the offending edge from aux
            ++r.first;
            boost::remove_edge(aux_uv, aux);

            // add aux_v to the queue & mark it seen
            if(!contains(S, aux_v)) {
              pending.push(aux_v);
              S.insert(aux_v);
            }// if
          }// for
        }// while (pending is not empty)
        DEBUG5(std::cout << "." << std::endl);
        const size_t S_size = S.size();
        // add the exiting arcs to the expressions
        for(const Vertex<AuxGraph>& aux_u : S){
          const VertexProp& u_prop = aux[aux_u];
          const ScafVertex& u = u_prop.orig;
          // note: if layer_of_u is 0, then uv is a matching edge and there are no edges from u leaving S (except to t_c)
          if(u_prop.layer == 1){
            for(auto e = I.get_incident(u); e; ++e){
              const ScafVertex& v = I.target(*e);
              const auto aux_v_it = translate.at(0).find(v);
              // if the aux-version of uv is not within SxS, then modify the expressions using uv
              if( (aux_v_it == translate.at(0).end()) || !contains(S, aux_v_it->second)){
                const ScafVertexPair uv(u, v);
                const ScafVertexPair vu(v, u);

                expr += (double)(S_size * max_multiplicity) * vars.times_used.at(uv);
                rev_expr += (double)(S_size * max_multiplicity) * vars.times_used.at(vu);
              }// if
            }// for
          } else {
            if(I.num_cycles > 0){
              // if the layer of u is 0, then add u -> t_c
              expr += (double)(S_size * max_multiplicity) * vars.to_cycle_sink.at(u);
              rev_expr += (double)(S_size * max_multiplicity) * vars.to_cycle_sink.at(u);
            }
          }// if layer = 1 / else
        }// for S
        // add the expressions to the model
        add(expr >= 0);
        add(rev_expr >= 0);
        num_new_cuts += 2;
      }// while  aux has edges
      return num_new_cuts;
    }


    void main()
    {
      stats.time_spent.resume();
      DEBUG5(std::cout << "callback starts"<<std::endl);
      // note: forbidding cycles in w takes a bit longer, so don't run the cycle-interdiction on w unless y is acyclic
      unsigned new_cuts = 0; //forbid_cycles_in_y();

      // Step 3: collect statistical information
      DEBUG2(std::cout << "=======   callback called, added "<<new_cuts<<" new cuts ======"<<std::endl);
      stats.times_called++;
      stats.cuts_added += new_cuts;
      stats.time_spent.pause();
    } // main()


    IloCplex::CallbackI* duplicateCallback() const{
      return (new (getEnv()) LazyCyclesBlockCallback(getEnv(), vars, stats));
    }
    
    LazyCyclesBlockCallback(IloEnv env, const VAR_COLLECTION& _vars, CallbackStatistics& _stats):
      IloCplex::LazyConstraintCallbackI(env),
      vars(_vars),
      stats(_stats)
    {}

  }; // class

  // callback function
  IloCplex::Callback GetLazyCyclesBlock(IloEnv env, const VAR_COLLECTION& vars, CallbackStatistics& stats) {
    return (new (env) LazyCyclesBlockCallback(env, vars, stats));
  }





  // initialize the variables for the ILP formulation
  void populate_variables(IloModel& model,
                          const IloEnv& env,
                          VAR_COLLECTION& vars,
                          const ILPOptions& opts,
#ifdef STATISTICS
                          JumpStatistics& jump_stats,
#endif
                          const IloNum max_mult = IloInfinity)
  {
    const Instance& I = vars.I;
    const RawScaffoldGraph& g = I.get_graph();
    //IloRangeArray c(env);

    DEBUG4(std::cout << "creating z_{u,v}/x_{u,v}/y_uv"<<std::endl);
    for(auto e_it = I.get_edges(); e_it; ++e_it){
      const ScafEdge e = *e_it;
      const ScafVertex& u = I.source(e);
      const ScafVertex& v = I.target(e);
      const bool is_matching = I[e].is_matching_edge();
      // note: matching edges have to be taken at least once and have a restricted upper bound
      const IloNum uv_upper = I[e].multiplicity;
      const IloNum uv_lower = is_matching ? 1 : 0;
      const std::string uname = I[u].name;
      const std::string vname = I[v].name;
      
      if(opts.multi_take_all){
        // Step 1: add binary variables z_e = 1   <=>  e belogs to a solution path/cycle
        const std::string zname("z_" + uname + "," + vname);
        vars.used_at_all.emplace(PWC, std::make_tuple(u,v), std::make_tuple(env, 0, 1, IloNumVar::Bool, zname.c_str()));
      }
      // Step 1a: add variables x_e = # of occurances of e
      const std::string xname("x_" + uname + "," + vname);
      //DEBUG5(std::cout << xname << ": [" << uv_lower << "," << uv_upper << "]"<<std::endl);
      vars.times_used_undir.emplace(PWC, std::make_tuple(u,v), std::make_tuple(env, uv_lower, uv_upper, IloNumVar::Int, xname.c_str()));
      // Step 2: add variables y^k_ij
      const std::string yname1("y_" + uname + "," + vname);
      const std::string yname2("y_" + vname + "," + uname);
      //DEBUG5(std::cout << yname1 << ": [0," << uv_upper << "]"<<std::endl);
      //DEBUG5(std::cout << yname2 << ": [0," << uv_upper << "]"<<std::endl);
      vars.times_used.emplace(PWC, std::make_tuple(u,v), std::make_tuple(env, 0, uv_upper, IloNumVar::Int, yname1.c_str()));
      vars.times_used.emplace(PWC, std::make_tuple(v,u), std::make_tuple(env, 0, uv_upper, IloNumVar::Int, yname2.c_str()));
    }

    DEBUG4(std::cout << "creating s/tp/tc interaction variables"<<std::endl);
    // Step 3: add variables y_s,i and y_i,tp and y_i,tc
    for(ScafVIterRange v = boost::vertices(g); v.first != v.second; ++v.first){
      const ScafVertex& u = *v.first;
      const std::string name(g[u].name);
      // on layer 0, come up from s
      //DEBUG5(std::cout << "y_s," << name << ": [0," << max_mult << "]"<<std::endl);
      vars.from_source[u] = IloNumVar(env, 0, max_mult, IloNumVar::Int, ("y_s," + name).c_str());
      // on layer 1, go up to tp
      if(I.num_paths > 0){
        //DEBUG5(std::cout << "y_" << name << ",tp" << ": [0," << max_mult << "]"<<std::endl);
        vars.to_path_sink[u] = IloNumVar(env, 0, max_mult, IloNumVar::Int, ("y_" + name + ",tp").c_str());
      }
      // on layer 0, go up to tc
      if(I.num_cycles > 0){
        //DEBUG5(std::cout <<"y_"<< name <<",tc"<< ": [0," << max_mult << "]"<<std::endl);
        vars.to_cycle_sink[u] = IloNumVar(env, 0, max_mult, IloNumVar::Int, ("y_" + name + ",tc").c_str());
      }
    }

    if(!opts.ignore_contig_jumps){
      DEBUG4(std::cout << "creating contig-jump variables & expressions"<<std::endl);
      STAT(VertexPairToInt<RawScaffoldGraph> network_sizes);
      // Step 3: construct a contig jump network for uv
      for(auto e_it = I.get_non_matching_edges(); e_it.is_valid(); ++e_it){
        const ScafEdge& uv_edge = *e_it;
        const int static_length = I[uv_edge].length;
        if(static_length != NO_LENGTH){
          const ScafVertex& u = I.source(uv_edge);
          const ScafVertex& v = I.target(uv_edge);
          const UncertainLength uv_length = I.get_uncertain_length(uv_edge);
          const unsigned uv_multi = I[uv_edge].multiplicity;
          for(const ScafVertexPair uv: {ScafVertexPair(u,v), ScafVertexPair(v,u)}){
            // NOTE: the ILPContigJumpNetwork is only constructed to be integrated in the model
            // after integration it does not serve any more purpose and is discarded
            ILPContigJumpNetwork jump_network(I, uv, opts.jump_threshold_probability, uv_length);
            if(!jump_network.empty()){
              const std::string name = "jump_" + std::to_string(I.get_edge_name(uv));
              const IloNumVar& main_jump_var = vars.contig_jumps.emplace(PWC,
                  std::make_tuple(uv), std::make_tuple(env, 0, uv_multi, IloNumVar::Int, name.c_str())).first->second;
              jump_network.integrate_into_model(env, model, uv_multi, main_jump_var, vars.times_used);
            }
            STAT(DEBUG3(if(jump_network.num_edges() > 10) std::cout << "adding size-"<<jump_network.num_edges()<<" network"<<std::endl;))
            STAT(network_sizes[uv] = jump_network.num_edges();)
          }
        }
      }
      STAT(jump_stats.jump_network_sizes.from_pairs(network_sizes);)
    }// if we consider contig jumps
  }

  // initialize the constraints for the ILP formulation
  void populate_constraints(IloModel& model, const IloEnv& env, const VAR_COLLECTION& vars, const ILPOptions& opts)
  {
    const Instance& I = vars.I;
    const RawScaffoldGraph& g = I.get_graph();
    const ScafMatching& matching = I.get_matching();
    IloRangeArray c(env);

    if(opts.multi_take_all){
      // constraint (8): x_e = 0 --> z_e = 0
      DEBUG4(std::cout << "adding (8): x_e = 0 --> z_e = 0"<<std::endl);
      for(auto x : vars.times_used_undir){
        const ScafVertexPair uv = x.first;

        IloExpr expr(env);
        expr += x.second;
        expr -= vars.used_at_all.at(uv);
        c.add(expr >= 0);
      }
    }

    // constraint (1): of all the representations of ij, exactly x_ij should be chosen!
    DEBUG4(std::cout << "adding (1): of all the representations of e, exactly x_e should be chosen"<<std::endl);
    for(auto x : vars.times_used_undir){
      const ScafVertexPair uv = x.first;
      const ScafVertexPair vu(uv.second, uv.first);

      IloExpr expr(env);
      expr -= x.second;
      expr += vars.times_used.at(uv);
      expr += vars.times_used.at(vu);
      // add all the contig jump representations of uv and vu
      if(!opts.ignore_contig_jumps)
        if(!I.is_matching_edge(uv)){
          const auto uv_iter = vars.contig_jumps.find(uv);
          if(uv_iter != vars.contig_jumps.cend()) expr += uv_iter->second;
          const auto vu_iter = vars.contig_jumps.find(vu);
          if(vu_iter != vars.contig_jumps.cend()) expr += vu_iter->second;
        }
      c.add(expr == 0);
    }

    /* NOTE: this is now implemented as domain restrictions on the x-variables
    // constraint (2): all matching-edges should be chosen!
    DEBUG5(std::cout << "adding (2'): any matching-edge e should be chosen between 1 and m(e) times"<<std::endl);
    for(auto x : vars.times_used_undir)
      if(I.is_matching_edge(x.first)) {
        const ScafEdge e = boost::edge(x.first.first, x.first.second, g).first;
        c.add(x.second >= 1);
        c.add(x.second <= (double)boost::get(boost::edge_multiplicity, g, e));
        DEBUG3(std::cout<<"allowing "<<EdgeAndGraph<ScaffoldGraph>(e, g)<<" with multi "<<(double)boost::get(boost::edge_multiplicity, g, e)<<std::endl);
      }
      */
  
    // constraint (4),(6): there should be at most sigma_p paths & sigma_c cycles
    DEBUG4(std::cout << "adding (4)&(6): there should be at most sigma_p paths & sigma_c cycles & sigma objects in total"<<std::endl);
    {
      IloExpr p_expr(env), c_expr(env), o_expr(env);
      for(auto y : vars.to_path_sink) p_expr += y.second;
      c.add(p_expr <= (double)I.num_paths);
      
      for(auto y : vars.to_cycle_sink) c_expr += y.second;
      c.add(c_expr <= (double)I.num_cycles);

      for(auto y : vars.from_source)  o_expr += y.second;
      c.add(o_expr <= (double)I.num_objects);
    }

    // constraint (3): everything that comes into layer l should leave layer l
    DEBUG4(std::cout << "adding (3): flow conservation"<<std::endl);
    for(unsigned layer = 0; layer < 2; ++layer){
      DEBUG5(std::cout << "==== layer "<<layer<<" ===="<<std::endl);
      for(ScafVIterRange range = boost::vertices(g); range.first != range.second; ++range.first){
        const ScafVertex& u = *range.first;
        //DEBUG5(std::cout << "vertex "<<I[u].name<<": ");
        IloExpr expr(env);
        for(ScafAdjIterRange nh = adjacent_vertices(u, g); nh.first != nh.second; ++nh.first){
          const ScafVertex& v = *nh.first;
          const ScafVertexPair uv(u, v);
          const ScafVertexPair vu(v, u);
          // remember: layer 0 -> 1 contains only matching edges
          if((matching.at(u) == v) == (layer == 0)){
            // flow goes to next layer along uv
            expr += vars.times_used.at(uv);
            //DEBUG5(std::cout << " + y_" << I[u].name <<"->"<<I[v].name);
          } else {
            // flow comes from the other layer alog vu
            expr -= vars.times_used.at(vu);
            //DEBUG5(std::cout << " - y_"<< I[v].name <<"->"<<I[u].name);
          }// if
        }// for
        if(layer == 0){
          // on layer 0, p-flow can come up from s
          //DEBUG5(std::cout << " - y_s,"<< I[u].name);
          expr -= vars.from_source.at(u);
          if(I.num_cycles > 0){
            // on layer 0, p-flow can go up to tc
            //DEBUG5(std::cout << " + y_"<< I[u].name<<",tc");
            expr += vars.to_cycle_sink.at(u);
          }
        } else {
          if(I.num_paths > 0){
            // on layer 1, p-flow can go up to tp
            //DEBUG5(std::cout << " + y_"<< I[u].name<<",tp");
            expr += vars.to_path_sink.at(u);
          }
        }// if
        c.add(expr == 0);
        //DEBUG5(std::cout << " = 0" << std::endl);
      }// for
    }// for

    if(I.num_cycles > 0){
      // constraint (7): everything that goes to tc has to come in from s
      DEBUG4(std::cout << "adding (7): cycle conservation"<<std::endl);
      for(ScafVIterRange u = boost::vertices(g); u.first != u.second; ++u.first){
        IloExpr expr(env);
        expr += vars.from_source.at(*u.first);
        expr -= vars.to_cycle_sink.at(*u.first);
        c.add(expr >= 0);
      }
    }

    // constraint (5): each subset S of vertices should contain at most |S|-1 endpoints of arcs
    //DEBUG5(std::cout << "adding (5): each subset S of vertices should contain at most |S|-1 endpoints of arcs"<<std::endl);
    // done via callback!
    model.add(c);
  }

  void create_optimization_target(IloModel& model, const IloEnv& env, VAR_COLLECTION& vars, const ILPOptions& opts)
  {
    const Instance& I = vars.I;
    const RawScaffoldGraph& g = I.get_graph();

    IloExpr opt_exp(env);
    for(auto e_it = I.get_edges(); e_it; ++e_it){
      const ScafEdge& e = *e_it;
      const ScafEdgeProperty& uv_info = g[e];
      const ScafVertex& u = I.source(e);
      const ScafVertex& v = I.target(e);
      if(!uv_info.is_matching_edge()){
        // get the appropriate x_ij variable
        // note, if the solution weight does not scale with the multiplicity, we have boolean variables z indicating whether x > 0
        const EdgeVarMap& x = opts.multi_take_all ? vars.used_at_all : vars.times_used_undir;

        auto uv = x.find(ScafVertexPair(u,v));
        if(uv == x.cend()) uv = x.find(ScafVertexPair(v,u));
        assert(uv != x.cend());
        // add the term weight(e)*x_e
        opt_exp += (double)(I[e].weight) * uv->second;
      }
    }
    model.add(IloMaximize(env, opt_exp));

  }

  // compute the ILP formulation for a graph g
  void populate_model(IloModel& model,
                      const IloEnv& env,
                      VAR_COLLECTION& vars,
#ifdef STATISTICS
                      JumpStatistics& jump_stats,
#endif
                      const ILPOptions& opts)
  {
    // construct variables, keeping note of the x_ij, y^l_ij and z^l_ij variables
    DEBUG3(std::cout << "creating variables"<<std::endl);
#ifdef STATISTICS
    populate_variables(model, env, vars, opts, jump_stats);
#else
    populate_variables(model, env, vars, opts);
#endif
    // construct constraints
    DEBUG3(std::cout << "creating constraints"<<std::endl);
    populate_constraints(model, env, vars, opts);
    // add optimization function
    DEBUG3(std::cout << "creating optimization expression"<<std::endl);
    create_optimization_target(model, env, vars, opts);
  }


}} // namespace

