

#pragma once

#include <string>
#include <vector>

#ifdef STATISTICS
#include "utils/stats.hpp"
#endif

#include "utils/scaffold_graph.hpp"
#include "utils/profiling.hpp"
#include "utils/ilp_options.hpp"
#include "utils/ilp_common.hpp"
#include "utils/ilp_contig_jump_network.hpp"

#define VAR_COLLECTION var_collection_manygraph

namespace scaffold{ namespace solv {
 
  struct VAR_COLLECTION{
    const Instance& I;

    // integer variables x_e indicating how many times an edge e is used in the solution
    EdgeVarMap times_used_undir;
    // binary variables z_e indicating whether x_e > 0
    EdgeVarMap used_at_all;
    
    // binary variables y_u[i]v[j] indicating if flow goes from graph i to graph j via edge uv
    std::vector2d<EdgeVarMap> times_used;
    // binary variables jmp_u[i]v[j] indicating whether y_u[i]v[j] may be collected for free (violating flow preservation)
    std::vector2d<EdgeVarMap> contig_jumps;
    // binary variables y_su[i] indicating whether a solution flow uses s->u[i]
    std::vector<VertexVarMap> from_source;
    // binary variables y_u[i]tp indicating whether a solution flow uses u[i]->tp
    std::vector<VertexVarMap> to_path_sink;
    // binary variables y_u[i]tc indicating whether a solution flow uses u[i]->tc
    std::vector<VertexVarMap> to_cycle_sink;

    VAR_COLLECTION(const Instance& _I): I(_I)
    {
      MultiType max_multi = 0;
      // get the maximum multiplicity from I
      for(const auto& uv: I.get_matching())
        max_multi = std::max(max_multi, I[uv].multiplicity);

      times_used.resize(max_multi, max_multi);
      contig_jumps.resize(max_multi, max_multi);
      from_source.resize(max_multi);
      to_path_sink.resize(max_multi);
      to_cycle_sink.resize(max_multi);
    }
    MultiType max_multi() const { return times_used.cols(); }

    // return the number of times that an arc uv has been jumped
    unsigned num_jumped(const ScafVertexPair& uv, const IloCplex& cplex) const
    {
      const MultiType multi = max_multi();
      unsigned result = 0;
      for(MultiType i = 0; i < multi; ++i)
        for(MultiType j = 0; j < multi; ++j){
          const auto jump_uv = contig_jumps[{i,j}].find(uv);
          if(jump_uv != contig_jumps[{i,j}].cend())
            result += is_true(cplex.getValue(jump_uv->second));
        }
      return result;
    }

    // return number of paths in a solution
    unsigned num_paths(const IloCplex& cplex) const
    {
      const MultiType multi = max_multi();
      unsigned result = 0;
      for(MultiType i = 0; i < multi; ++i)
        for(const auto &u : to_path_sink[i]) result += is_true(cplex.getValue(u.second));
      return result;
    }

    // return number of paths in a solution
    unsigned num_cycles(const IloCplex& cplex) const
    {
      const MultiType multi = max_multi();
      unsigned result = 0;
      for(MultiType i = 0; i < multi; ++i)
        for(const auto &u : to_cycle_sink[i]) result += is_true(cplex.getValue(u.second));
      return result;
    }

    // teach a model to respect jump dependecies
    void respect_jump_dependencies(IloModel& model, const IloEnv& env, const ScafVPairDependencyMap& dependencies)
    {
      const MultiType multi = max_multi();
      // the dependency map maps each VertexPair p to a list of VertexPairs on which they depend, that is,
      // p cannot be taken more times than the sum of its dependencies
      IloRangeArray c(env);
      DEBUG5(std::cout << "adding contig-jump-shortcut constraints: if we take an added shortcut, we should also take its dependency"<<std::endl);
      for(const auto& dlist: dependencies){
        const ScafVertexPair& uv = dlist.first;
        const auto& uv_dep_list = dlist.second;
        // build the constraint
        IloExpr expr(env);
        IloExpr rev_expr(env);
        for(const auto& uv_dep_edge: uv_dep_list){
          const auto vu_dep_edge(reverse(uv_dep_edge));
          for(MultiType i = 0; i < multi; ++i)
            for(MultiType j = 0; j < multi; ++j){
              expr += times_used[{i,j}].at(uv_dep_edge);
              rev_expr += times_used[{j,i}].at(vu_dep_edge);
            }
        }
        const auto vu = reverse(uv);
        for(MultiType i = 0; i < multi; ++i)
          for(MultiType j = 0; j < multi; ++j){
            expr -= times_used[{i,j}].at(uv);
            rev_expr -= times_used[{j,i}].at(vu);
          }
        c.add(expr >= 0);
        c.add(rev_expr >= 0);
      }
      model.add(c);
    }

    // print contig jumps from a ContigJumpMap
    void print_contig_jumps(const IloCplex& cplex,
                            std::ostream& os = std::cout) const
    {
      const MultiType multi = max_multi();
      boost::unordered_map<ScafVertexPair, MultiType> true_count;
      for(MultiType i = 0; i < multi; ++i)
        for(MultiType j = 0; j < multi; ++j)
          for(const auto& e_and_var: contig_jumps[{i,j}])
            if(is_true(cplex.getValue(e_and_var.second)))
              true_count[e_and_var.first]++;
      os << true_count;
    }


  };


  // block isolated cycles as lazy constraint callback
  class LazyCyclesBlockCallback: public IloCplex::LazyConstraintCallbackI{

    struct VertexProp {
      // index of the graph containing this vertex
      unsigned graph_idx;
      // link to the original vertex in the graph corresponding to this vertex
      const ScafVertex& orig;

      VertexProp(const ScafVertex& _orig, const unsigned _graph_idx = 0): graph_idx(_graph_idx), orig(_orig) {}
    };
    typedef AuxiliaryGraph<VertexProp, no_property> AuxGraph;
    typedef std::vector2d<VTranslateMap<RawScaffoldGraph, AuxGraph> > TranslateMatrix;

    const VAR_COLLECTION& vars;
    CallbackStatistics& stats;

  public:

    void build_auxgraph(AuxGraph& aux, TranslateMatrix& translate_matrix);
    void compute_startpoints(const TranslateMatrix& translate_matrix, VertexSet<AuxGraph>& startpoints);
    unsigned forbid_cycles_in_y();
    void main();

    IloCplex::CallbackI* duplicateCallback() const{
      return (new (getEnv()) LazyCyclesBlockCallback(getEnv(), vars, stats));
    }
    
    LazyCyclesBlockCallback(IloEnv env, const VAR_COLLECTION& _vars, CallbackStatistics& _stats):
      IloCplex::LazyConstraintCallbackI(env),
      vars(_vars),
      stats(_stats)
    {}

  }; // class

  // callback function
  IloCplex::Callback GetLazyCyclesBlock(IloEnv env, const VAR_COLLECTION& vars, CallbackStatistics& stats) {
    return (new (env) LazyCyclesBlockCallback(env, vars, stats));
  }

  // build the auxiliary graph using the var collection vars 
  void LazyCyclesBlockCallback::build_auxgraph(AuxGraph& aux, TranslateMatrix& translate_matrix)
  {
    const Instance& I = vars.I;
    const MultiType multi = vars.max_multi();

    // construct the auxiliary graph from the current solution
    for(MultiType i = 0; i < multi; ++i){
      for(MultiType j = 0; j < multi; ++j){
        for(const auto& y : vars.times_used[{i,j}]) if(is_true(getValue(y.second))){
          const ScafVertexPair& uv = y.first;
          const ScafVertex& u = uv.first;
          const ScafVertex& v = uv.second;
          const unsigned layer = I.is_matching_edge(uv) ? 0 : 1;

          DEBUG5(std::cout << "used "<< (!layer ? "matching " : "") << "edge "<< I[u].name << "-->"<<I[v].name<<" between graphs "<<i<<" & "<<j<<std::endl);

          // add vertices: first see if they're already there & if they are not there, add them to aux
          VTranslateMap<RawScaffoldGraph, AuxGraph>& i_translate = translate_matrix[{layer, i}];
          auto u_it = i_translate.find(u);
          if(u_it == i_translate.end())
            u_it = i_translate.insert(u_it, {u, boost::add_vertex(VertexProp(u, i), aux)});

          VTranslateMap<RawScaffoldGraph, AuxGraph>& j_translate = translate_matrix[{1-layer, j}];
          auto v_it = j_translate.find(v);
          if(v_it == j_translate.end())
            v_it = j_translate.insert(v_it, {v, boost::add_vertex(VertexProp(v, j), aux)});

          // add edge
          const Vertex<AuxGraph>& aux_u = u_it->second;
          const Vertex<AuxGraph>& aux_v = v_it->second;
          boost::add_edge(aux_u, aux_v, aux);
          DEBUG5(std::cout << "added "<<I[u].name<<"["<<boost::out_degree(aux_u, aux)<<"]-->"<<I[v].name<< (!layer?"(*)":"")<<"\tnow "<<boost::num_vertices(aux)<<" vertices & "<<boost::num_edges(aux)<<" edges remain"<<std::endl);
        }// for
      }
    }
  }// function

  // return the startpoints of all paths & cycles
  void LazyCyclesBlockCallback::compute_startpoints(const TranslateMatrix& translate_matrix, VertexSet<AuxGraph>& startpoints)
  {
    DEBUG5(std::cout << "reachable from s: ");
    const MultiType multi = vars.max_multi();
    DEBUG5(const Instance& I = vars.I);

    for(MultiType i = 0; i < multi; ++i)
      for(const auto& y_su: vars.from_source.at(i))
        if(is_true(getValue(y_su.second))) {
          const ScafVertex& u = y_su.first;
          DEBUG5(std::cout << I[u].name <<" ");
          // note that CPLEX might give us a solution that has some s->u->t_c paths; if, in this case, u does not have an edge in aux, we can't translate
          const auto& i_translate = translate_matrix.at({0, i});
          const auto u_iter = i_translate.find(u);
          if(u_iter != i_translate.end())
            startpoints.insert(u_iter->second);
        }// if
    DEBUG5(std::cout << "."<< std::endl);
  }

  unsigned LazyCyclesBlockCallback::forbid_cycles_in_y()
  {
    const MultiType multi = vars.max_multi();
    AuxGraph aux;
    TranslateMatrix translate({2, multi});
    VertexSet<AuxGraph> startpoints;

    // Step 1: build the auxiliary graph & remove everything that is reachable from s
    DEBUG5(std::cout << "building auxiliary graph..."<<std::endl);
    build_auxgraph(aux, translate);
    DEBUG5(std::cout << "built aux graph with "<<boost::num_vertices(aux)<<" vertices & "<<boost::num_edges(aux)<<" edges"<<std::endl);

    DEBUG5(std::cout << "removing valid paths/cycles from auxiliary graph..."<<std::endl);
    // parse & remove all paths & cycles incident with startpoints
    compute_startpoints(translate, startpoints);
    remove_reachable(aux, startpoints, dir_fwd);
    DEBUG5(std::cout << "after cleanup, "<<boost::num_vertices(aux)<<" vertices & "<<boost::num_edges(aux)<<" edges remain"<<std::endl);

    // whatever remains of aux are cycles that are not attached to s, that is, invalid cycles;
    // Step 2: forbid them by adding a constraint forcing at least one arc going out of the cycle
    DEBUG5(std::cout<< "remaining arcs in auxilliary graph: "<<num_edges(aux)<<std::endl);
    unsigned num_new_cuts = 0;
    if(boost::num_edges(aux)){

#ifdef LAZY_ADD_ALL_CYCLES
    while(boost::num_edges(aux)){
#endif
      Edge<AuxGraph> uv =  *(boost::edges(aux).first);
      DEBUG1(Vertex<AuxGraph> stop = boost::source(uv, aux));
      DEBUG1(const Vertex<AuxGraph> start = stop);
      IloExpr expr(getEnv()), rev_expr(getEnv());
      unsigned count = 0;
      unsigned v_deg = 0;
      DEBUG3(std::cout << "forbidding cycle: ");
      do{
        const Vertex<AuxGraph>& u = boost::source(uv, aux);
        const Vertex<AuxGraph>& v = boost::target(uv, aux);
        const ScafVertex& old_u = aux[u].orig;
        const ScafVertex& old_v = aux[v].orig;
        const ScafVertexPair old_uv(old_u, old_v);
        const ScafVertexPair old_vu(old_v, old_u);

        DEBUG3(std::cout << I.get_edge_name(old_uv)<<" ");
        boost::remove_edge(uv, aux);
        v_deg = boost::out_degree(v, aux);
        assert(v_deg <= 1);
        // if this outgoing arc is not to our initial arc, then add the next vertex to pending & remove the arc
        if(v_deg == 1) uv = *(out_edges(v, aux).first);
        
        // forbid the cycle and its reverse in y and z
        expr += vars.times_used.at({aux[u].graph_idx, aux[v].graph_idx}).at(old_uv);
        expr += vars.times_used.at({aux[v].graph_idx, aux[u].graph_idx}).at(old_vu);
        expr -= vars.to_cycle_sink.at(aux[v].graph_idx).at(old_v);

        ++count;
        DEBUG1(stop = v);
      } while(v_deg != 0);
      DEBUG1(assert(start == stop));
      DEBUG5(std::cout << " by allowing only <"<<count<<" of its edges (unless it's connected to t_c)");
      DEBUG3(std::cout << std::endl);
      add(expr < count);
      num_new_cuts++;
#ifdef LAZY_ADD_ALL_CYCLES
    }// while(num_edges(aux))
#endif
    }// if(num_edges(aux))
    return num_new_cuts;
  }


  void LazyCyclesBlockCallback::main()
  {
    stats.time_spent.resume();
    DEBUG5(std::cout << "callback starts"<<std::endl);

    // note: forbidding cycles in w takes a bit longer, so don't run the cycle-interdiction on w unless y is acyclic
    unsigned new_cuts = forbid_cycles_in_y();

    // Step 3: collect statistical information
    DEBUG2(std::cout << "=======   callback called, added "<<new_cuts<<" new cuts ======"<<std::endl);
    stats.times_called++;
    stats.cuts_added += new_cuts;
    stats.time_spent.pause();
  } // main()



  // initialize the variables for the ILP formulation
  void populate_variables(IloModel& model,
                          const IloEnv& env,
                          VAR_COLLECTION& vars,
                          const ILPOptions& opts,
#ifdef STATISTICS
                          JumpStatistics& jump_stats,
#endif
                          const IloNum max_mult = IloInfinity)
  {
    const Instance& I = vars.I;
    const MultiType multi = vars.max_multi();
    //IloRangeArray c(env);

    DEBUG4(std::cout << "creating z_{u,v}/x_{u,v}/y_uv"<<std::endl);
    for(auto e_it = I.get_edges(); e_it; ++e_it){
      const ScafEdge& e = *e_it;
      const ScafVertex& u = I.source(e);
      const ScafVertex& v = I.target(e);
      const bool is_matching = I[e].is_matching_edge();
      // note: matching edges have to be taken at least once and have a restricted upper bound
      const IloNum uv_upper = I[e].multiplicity;
      const IloNum uv_lower = is_matching ? 1 : 0;
      const std::string& uname = I[u].name;
      const std::string& vname = I[v].name;
      
      if(opts.multi_take_all){
        // Step 1: add binary variables z_e = 1   <=>  e belogs to a solution path/cycle
        const std::string zname("z_" + uname + "," + vname);
        vars.used_at_all.emplace(PWC, std::make_tuple(u,v), std::make_tuple(env, 0, 1, IloNumVar::Bool, zname.c_str()));
      }
      // Step 1a: add variables x_e = # of occurances of e
      const std::string xname("x_" + uname + "," + vname);
      //DEBUG5(std::cout << xname << ": [" << uv_lower << "," << uv_upper << "]"<<std::endl);
      vars.times_used_undir.emplace(PWC, std::make_tuple(u,v), std::make_tuple(env, uv_lower, uv_upper, IloNumVar::Int, xname.c_str()));
      // Step 2: add variables y
      for(MultiType i = 0; i < multi; ++i)
        for(MultiType j = 0; j < multi; ++j){
          const std::string yname1("y_" + uname + "[" + to_string(i) + "]," + vname + "[" + to_string(j) + "]");
          const std::string yname2("y_" + vname + "[" + to_string(j) + "]," + uname + "[" + to_string(i) + "]");
          vars.times_used[{j,i}].emplace(PWC, std::make_tuple(v,u), std::make_tuple(env, 0, 1, IloNumVar::Bool, yname2.c_str()));
          vars.times_used[{i,j}].emplace(PWC, std::make_tuple(u,v), std::make_tuple(env, 0, 1, IloNumVar::Bool, yname1.c_str()));
        }
    }

    DEBUG4(std::cout << "creating s/tp/tc interaction variables"<<std::endl);
    // Step 3: add variables y_s,i and y_i,tp and y_i,tc
    for(auto u_it = I.get_vertices(); u_it; ++u_it){
      const ScafVertex& u = *u_it;
      const std::string& name = I[u].name;

      for(MultiType i = 0; i < multi; ++i){
        // on layer 0, come up from s
        //DEBUG5(std::cout << "y_s," << name << ": [0," << max_mult << "]"<<std::endl);
        vars.from_source[i].emplace(PWC, std::make_tuple(u), std::make_tuple(env, 0, max_mult, IloNumVar::Int, ("y_s," + name).c_str()));
        // on layer 1, go up to tp
        if(I.num_paths > 0){
          //DEBUG5(std::cout << "y_" << name << ",tp" << ": [0," << max_mult << "]"<<std::endl);
          vars.to_path_sink[i].emplace(PWC, std::make_tuple(u), std::make_tuple(env, 0, max_mult, IloNumVar::Int, ("y_" + name + ",tp").c_str()));
        }
        // on layer 0, go up to tc
        if(I.num_cycles > 0){
          //DEBUG5(std::cout <<"y_"<< name <<",tc"<< ": [0," << max_mult << "]"<<std::endl);
          vars.to_cycle_sink[i].emplace(PWC, std::make_tuple(u), std::make_tuple(env, 0, max_mult, IloNumVar::Int, ("y_" + name + ",tc").c_str()));
        }
      }
    }

    if(!opts.ignore_contig_jumps){
      DEBUG4(std::cout << "creating contig-jump variables & expressions"<<std::endl);
      STAT(VertexPairToInt<RawScaffoldGraph> network_sizes);
      // Step 3: construct a contig jump network for uv
      for(auto e_it = I.get_non_matching_edges(); e_it.is_valid(); ++e_it){
        const ScafEdge& uv_edge = *e_it;
        const auto& uv_prop = I[uv_edge];
        if(uv_prop.length != NO_LENGTH){
          const ScafVertex& u = I.source(uv_edge);
          const ScafVertex& v = I.target(uv_edge);
          const UncertainLength uv_length = I.get_uncertain_length(uv_edge);
          const unsigned uv_multi = 1;
          for(const ScafVertexPair uv: {ScafVertexPair(u,v), ScafVertexPair(v,u)}){
            // NOTE: the ILPContigJumpNetwork is only constructed to be integrated in the model
            // after integration it does not serve any more purpose and is discarded
            ILPContigJumpNetwork jump_network(I, uv, opts.jump_threshold_probability, uv_length);
            if(!jump_network.empty()){
              const std::string& uname = I[uv.first].name;
              const std::string& vname = I[uv.second].name;
              for(MultiType i = 0; i < multi; ++i)
                for(MultiType j = 0; j < multi; ++j){
                  const std::string name("jmp_" + uname + "[" + to_string(i) + "]," + vname + "[" + to_string(j) + "]");

                  const IloNumVar& main_jump_var = vars.contig_jumps[{i,j}].emplace(PWC,
                      std::forward_as_tuple(uv), std::forward_as_tuple(env, 0, uv_multi, IloNumVar::Int, name.c_str())).first->second;
                  jump_network.integrate_into_model(env, model, uv_multi, main_jump_var, vars.times_used[{i,j}]);
                }
            }
            STAT(DEBUG3(if(jump_network.num_edges() > 10) std::cout << "adding size-"<<jump_network.num_edges()<<" network"<<std::endl))
            STAT(network_sizes[uv] = jump_network.num_edges();)
          }
        }
      }
      STAT(jump_stats.jump_network_sizes.from_pairs(network_sizes);)
    }// if we consider contig jumps
  }

  // initialize the constraints for the ILP formulation
  void populate_constraints(IloModel& model, const IloEnv& env, const VAR_COLLECTION& vars, const ILPOptions& opts)
  {
    const Instance& I = vars.I;
    const RawScaffoldGraph& g = I.get_graph();
    const ScafMatching& matching = I.get_matching();
    const MultiType multi = vars.max_multi();
    IloRangeArray c(env);

    if(opts.multi_take_all){
      // constraint (8): x_e = 0 --> z_e = 0
      DEBUG4(std::cout << "adding (8): x_e = 0 --> z_e = 0"<<std::endl);
      for(const auto& x : vars.times_used_undir){
        const ScafVertexPair& uv = x.first;

        IloExpr expr(env);
        expr += x.second;
        expr -= vars.used_at_all.at(uv);
        c.add(expr >= 0);
      }
    }

    // constraint (1): of all the representations of ij, exactly x_ij should be chosen!
    DEBUG4(std::cout << "adding (1): of all the representations of e, exactly x_e should be chosen"<<std::endl);
    for(const auto& x : vars.times_used_undir){
      const ScafVertexPair& uv = x.first;
      const ScafVertexPair vu(uv.second, uv.first);

      IloExpr expr(env);
      expr -= x.second;
      for(MultiType i = 0; i < multi; ++i)
        for(MultiType j = 0; j < multi; ++j){
          expr += vars.times_used[{i,j}].at(uv);
          expr += vars.times_used[{i,j}].at(vu);
        }
      // add all the contig jump representations of uv and vu
      if(!opts.ignore_contig_jumps)
        if(!I.is_matching_edge(uv))
          for(MultiType i = 0; i < multi; ++i)
            for(MultiType j = 0; j < multi; ++j){
              const auto uv_iter = vars.contig_jumps[{i,j}].find(uv);
              if(uv_iter != vars.contig_jumps[{i,j}].cend()) expr += uv_iter->second;
              const auto vu_iter = vars.contig_jumps[{j,i}].find(vu);
              if(vu_iter != vars.contig_jumps[{j,i}].cend()) expr += vu_iter->second;
            }
      c.add(expr == 0);
    }

    /* NOTE: this is now implemented as domain restrictions on the x-variables
    // constraint (2): all matching-edges should be chosen!
    DEBUG5(std::cout << "adding (2'): any matching-edge e should be chosen between 1 and m(e) times"<<std::endl);
    for(auto x : vars.times_used_undir)
      if(I.is_matching_edge(x.first)) {
        const ScafEdge e = boost::edge(x.first.first, x.first.second, g).first;
        c.add(x.second >= 1);
        c.add(x.second <= (double)boost::get(boost::edge_multiplicity, g, e));
        DEBUG3(std::cout<<"allowing "<<EdgeAndGraph<ScaffoldGraph>(e, g)<<" with multi "<<(double)boost::get(boost::edge_multiplicity, g, e)<<std::endl);
      }
      */
  
    // constraint (4),(6): there should be at most sigma_p paths & sigma_c cycles
    DEBUG4(std::cout << "adding (4)&(6): there should be at most sigma_p paths & sigma_c cycles & sigma objects in total"<<std::endl);
    {
      IloExpr p_expr(env), c_expr(env), o_expr(env);
      for(MultiType i = 0; i < multi; ++i)
        for(auto y : vars.to_path_sink[i]) p_expr += y.second;
      c.add(p_expr <= (double)I.num_paths);
      
      for(MultiType i = 0; i < multi; ++i)
        for(auto y : vars.to_cycle_sink[i]) c_expr += y.second;
      c.add(c_expr <= (double)I.num_cycles);

      for(MultiType i = 0; i < multi; ++i)
        for(auto y : vars.from_source[i])  o_expr += y.second;
      c.add(o_expr <= (double)I.num_objects);
    }

    // constraint (3): everything that comes into layer l should leave layer l
    DEBUG4(std::cout << "adding (3): flow conservation"<<std::endl);
    for(ScafVIterRange range = boost::vertices(g); range.first != range.second; ++range.first){
      const ScafVertex& u = *range.first;
      //DEBUG5(std::cout << "vertex "<<I[u].name<<": ");
      for(MultiType i = 0; i < multi; ++i){
        for(short layer = 0; layer < 2; ++layer){
          IloExpr expr(env);
          for(MultiType j = 0; j < multi; ++j){
            for(ScafAdjIterRange nh = adjacent_vertices(u, g); nh.first != nh.second; ++nh.first){
              const ScafVertex& v = *nh.first;
              const ScafVertexPair uv(u, v);
              const ScafVertexPair vu(v, u);
              // remember: layer 0 -> 1 contains only matching edges
              if((matching.at(u) == v) == (layer == 0)){
                // flow goes to next layer along uv
                expr += vars.times_used[{i,j}].at(uv);
                //DEBUG5(std::cout << " + y_" << I[u].name <<"->"<<I[v].name);
              } else {
                // flow comes from the other layer alog vu
                expr -= vars.times_used[{j,i}].at(vu);
                //DEBUG5(std::cout << " - y_"<< I[v].name <<"->"<<I[u].name);
              }// if
            }// for
          }// for each copy j
          if(layer == 0){
            // on layer 0, p-flow can come up from s
            //DEBUG5(std::cout << " - y_s,"<< I[u].name);
            expr -= vars.from_source[i].at(u);
            if(I.num_cycles > 0){
              // on layer 0, p-flow can go up to tc
              //DEBUG5(std::cout << " + y_"<< I[u].name<<",tc");
              expr += vars.to_cycle_sink[i].at(u);
            }
          } else {
            if(I.num_paths > 0){
              // on layer 1, p-flow can go up to tp
              //DEBUG5(std::cout << " + y_"<< I[u].name<<",tp");
              expr += vars.to_path_sink[i].at(u);
            }
          }// if
          //DEBUG5(std::cout << " = 0" << std::endl);
          c.add(expr == 0);
        }// for each layer in 0,1
      }// for each copy i
    }// for each vertex u

    if(I.num_cycles > 0){
      // constraint (7): everything that goes to tc has to come in from s
      DEBUG4(std::cout << "adding (7): cycle conservation"<<std::endl);
      for(ScafVIterRange u = boost::vertices(g); u.first != u.second; ++u.first){
        IloExpr expr(env);
        for(MultiType i = 0; i < multi; ++i){
          expr += vars.from_source[i].at(*u.first);
          expr -= vars.to_cycle_sink[i].at(*u.first);
        }
        c.add(expr >= 0);
      }
    }

    // constraint (5): each subset S of vertices should contain at most |S|-1 endpoints of arcs
    //DEBUG5(std::cout << "adding (5): each subset S of vertices should contain at most |S|-1 endpoints of arcs"<<std::endl);
    // done via callback!
    model.add(c);
  }

  void create_optimization_target(IloModel& model, const IloEnv& env, VAR_COLLECTION& vars, const ILPOptions& opts)
  {
    const Instance& I = vars.I;
    const RawScaffoldGraph& g = I.get_graph();

    IloExpr opt_exp(env);
    for(ScafEdgeIterRange e = boost::edges(g); e.first != e.second; ++e.first){
      const ScafEdgeProperty& uv_info = g[*e.first];
      const ScafVertex& u = boost::source(*e.first, g);
      const ScafVertex& v = boost::target(*e.first, g);
      if(!uv_info.is_matching_edge()){
        // get the appropriate x_ij variable
        // note, if the solution weight does not scale with the multiplicity, we have boolean variables z indicating whether x > 0
        const EdgeVarMap& x = opts.multi_take_all ? vars.used_at_all : vars.times_used_undir;

        auto uv = x.find(ScafVertexPair(u,v));
        if(uv == x.cend()) uv = x.find(ScafVertexPair(v,u));
        assert(uv != x.cend());
        // add the term weight(e)*x_e
        opt_exp += (double)(I[*e.first].weight) * uv->second;
      }
    }
    model.add(IloMaximize(env, opt_exp));

  }

  // compute the ILP formulation for a graph g
  void populate_model(IloModel& model,
                      const IloEnv& env,
                      VAR_COLLECTION& vars,
#ifdef STATISTICS
                      JumpStatistics& jump_stats,
#endif
                      const ILPOptions& opts)
  {
    // construct variables, keeping note of the x_ij, y^l_ij and z^l_ij variables
    DEBUG3(std::cout << "creating variables"<<std::endl);
#ifdef STATISTICS
    populate_variables(model, env, vars, opts, jump_stats);
#else
    populate_variables(model, env, vars, opts);
#endif
    // construct constraints
    DEBUG3(std::cout << "creating constraints"<<std::endl);
    populate_constraints(model, env, vars, opts);
    // add optimization function
    DEBUG3(std::cout << "creating optimization expression"<<std::endl);
    create_optimization_target(model, env, vars, opts);
  }


}} // namespace

