

#pragma once

#include <string>
#include <vector>

#ifdef STATISTICS
#include "utils/stats.hpp"
#endif

#include "utils/scaffold_graph.hpp"
#include "utils/profiling.hpp"
#include "utils/ilp_options.hpp"
#include "utils/ilp_common.hpp"
#include "utils/ilp_contig_jump_network.hpp"

#define VAR_COLLECTION var_collection


namespace scaffold{
  namespace solv{

  using AuxGraph = AuxiliaryGraph<>;

  // block isolated cycles as lazy constraint callback
  class LazyCyclesBlockCallback: public IloCplex::LazyConstraintCallbackI{
    const VAR_COLLECTION& vars;
    CallbackStatistics& stats;
  public:

    IloCplex::CallbackI* duplicateCallback() const{
      return (new (getEnv()) LazyCyclesBlockCallback(getEnv(), vars, stats));
    }
    
    LazyCyclesBlockCallback(IloEnv env, const VAR_COLLECTION& _vars, CallbackStatistics& _stats):
      IloCplex::LazyConstraintCallbackI(env),
      vars(_vars),
      stats(_stats)
    {}
    
    void output_current_aux_graph(const AuxGraph& aux, const TranslateBiMap<RawScaffoldGraph,AuxGraph>& translate) const;
    void main();

    // adds an arc to the auxiliary graph, given an IloNumVar and a VertexPair
    void add_arc_from_variable(
        const IloNumVar& var,
        const ScafVertexPair& uv,
        const TranslateBiMap<RawScaffoldGraph,AuxGraph>& translate,
        AuxGraph& aux) const
    {
      if(is_true(getValue(var))){
        const ScafVertex u = uv.first;
        const ScafVertex v = uv.second;
        const Vertex<AuxGraph>& new_u = translate.left.at(u);
        const Vertex<AuxGraph>& new_v = translate.left.at(v);
        add_edge(new_u, new_v, aux);
        DEBUG5(std::cout << "adding  "<< vars.I.get_edge_name(uv)<<std::endl); 
      } else DEBUG5(std::cout << "not adding "<< vars.I.get_edge_name(uv) << " since "
                              << getValue(var)<<" translates to "<<is_true(getValue(var))<<std::endl);
    }

  }; // class

  // callback function
  IloCplex::Callback GetLazyCyclesBlock(IloEnv env, const VAR_COLLECTION& vars, CallbackStatistics& stats) {
    return (new (env) LazyCyclesBlockCallback(env, vars, stats));
  }


  void LazyCyclesBlockCallback::output_current_aux_graph(const AuxGraph& aux,
                                                         const TranslateBiMap<RawScaffoldGraph,AuxGraph>& translate) const
  {
    const Instance& I = vars.I;
    for(VertexIterRange<AuxGraph> i = vertices(aux); i.first != i.second; ++i.first)
      if(out_degree(*(i.first), aux) != in_degree(*(i.first), aux)){
        const ScafVertex u = translate.right.at(*(i.first));

        std::cout << "found vertex "<< I[u].name <<" with outdeg = "<< out_degree(*(i.first), aux)
                                                 <<" & indeg = "<< in_degree(*(i.first), aux) << std::endl;
        std::cout << "we have: ";
        if( is_true(getValue(vars.from_source.at(u)))) std::cout<<"s ";
        if( is_true(getValue(vars.to_path_sink.at(u)))) std::cout<<"tp ";
        if( is_true(getValue(vars.to_cycle_sink.at(u)))) std::cout<<"tc ";
        for(VertexIterRange<AuxGraph> j = vertices(aux); j.first != j.second; ++j.first){
          const ScafVertex v = translate.right.at(*(j.first));
          ScafVertexPair uv(u, v);
          ScafVertexPair vu(v, u);
          
          if(contains(vars.times_used, uv) && is_true(getValue(vars.times_used.at(uv)))) std::cout << "y_"<<I.get_edge_name(uv)<<", ";
          if(contains(vars.times_used, vu) && is_true(getValue(vars.times_used.at(vu)))) std::cout << "y_"<<I.get_edge_name(vu)<<", ";
        }// for
        std::cout << std::endl;
       
        Vertex<AuxGraph> ux = *(i.first);
        while(in_degree(ux, aux) > 0){
          const auto in_edge = in_edges(ux, aux).first;
          ux = source(*in_edge, aux);
          std::cout <<  I[translate.right.at(ux)].name<< " has indeg "<<in_degree(ux, aux)<<std::endl;
        }
        while(out_degree(ux, aux) > 0){
          const auto out_edge = out_edges(ux, aux).first;
          ux = target(*out_edge, aux);
          std::cout <<  I[translate.right.at(ux)].name<< " has outdeg "<<out_degree(ux, aux)<<std::endl;
        }

        exit(1);
    }// if & for
  }

  void LazyCyclesBlockCallback::main(){
    stats.time_spent.resume();
    AuxGraph aux;

    const Instance& I = vars.I;
    // Step 1: construct auxiliary graph
    TranslateBiMap<RawScaffoldGraph, AuxGraph> translate;
    typedef typename TranslateBiMap<RawScaffoldGraph, AuxGraph>::value_type TranslatePair;

    DEBUG5(
      std::cout << "solution edges: "<<std::endl;
      for(auto e = I.get_edges(); e; ++e){
        const ScafVertex& u = I.source(*e);
        const ScafVertex& v = I.target(*e);
        // get the appropriate x_ij variable
        const ScafVertexPair uv(u, v);
        const ScafVertexPair vu(v, u);
        std::cout<< I.get_edge_name(uv)<<" by ";
        if(contains(vars.times_used, uv) && is_true(getValue(vars.times_used.at(uv)))) std::cout << "y_"<<I.get_edge_name(uv)<<", ";
        if(contains(vars.times_used, vu) && is_true(getValue(vars.times_used.at(vu)))) std::cout << "y_"<<I.get_edge_name(vu)<<", ";
        std::cout << std::endl;
      }
    );

    // create vertices
    for(auto u = I.get_vertices(); u; ++u){
      const ScafVertex& orig_u = *u;
      
      DEBUG5(std::cout << "adding  "<< I[orig_u].name<<std::endl);
      const Vertex<AuxGraph>& new_u = boost::add_vertex(aux);
      translate.insert(TranslatePair(orig_u, new_u));
    }
    // create arcs, merging y-arcs and z-arcs into one auxilliary digraph
    DEBUG5(std::cout << "adding y arcs"<<std::endl);
    for( const auto& i : vars.times_used ) add_arc_from_variable(i.second, i.first, translate, aux);

    // Step 2: remove all arcs reachable from s
    std::stack<Vertex<AuxGraph> > pending;
    for( const auto& i : vars.from_source )
      if(is_true(getValue(i.second))) pending.push(translate.left.at(i.first));

    DEBUG5(
      std::cout<<"coming from y_s: ";
      for( const auto& i : vars.from_source )
        if(is_true(getValue(i.second))) std::cout <<  I[i.first].name<<" ";
      std::cout<<std::endl;
      std::cout<<"going to tp: ";
      for( const auto& i : vars.to_path_sink )
        if(is_true(getValue(i.second))) std::cout <<  I[i.first].name<<" ";
      std::cout<< std::endl<<"going to tc: ";
      for( const auto& i : vars.to_cycle_sink )
        if(is_true(getValue(i.second))) std::cout <<  I[i.first].name<<" ";
      std::cout<<std::endl;
      );

    DEBUG2(unsigned num_p = pending.size();)

    while(!pending.empty()){
      const Vertex<AuxGraph> u = pending.top(); pending.pop();
      // note that u has an incoming arc y_s -> u, so, by flow conservation, it has only 1 outgoing arc
      const unsigned char deg = boost::out_degree(u, aux);
      assert(deg <= 1);
      // if this outgoing arc is not to y_t, then add the next vertex to pending & remove the arc
      if(deg == 1) {      
        const Edge<AuxGraph>& e = *(boost::out_edges(u, aux).first);
        const Vertex<AuxGraph>& v = boost::target(e, aux);
        DEBUG5(std::cout <<  I[translate.right.at(v)].name << " is on a path by edge from "<<
                             I[translate.right.at(boost::source(e, aux))].name<<std::endl);
        pending.push(v);
        remove_edge(e, aux);
      }
    }

    // Step 3: find a cycle in the remaining part of the graph
    DEBUG5(std::cout<< "remaining arcs in auxilliary graph: "<<num_edges(aux)<<std::endl;);
    unsigned num_new_cuts = 0;
    if(boost::num_edges(aux)){
      DEBUG5(output_current_aux_graph(aux, translate));

#ifdef LAZY_ADD_ALL_CYCLES
    while(boost::num_edges(aux)){
#endif
      Edge<AuxGraph> uv =  *(edges(aux).first);
      DEBUG1(Vertex<AuxGraph> stop = boost::source(uv, aux));
      DEBUG1(const Vertex<AuxGraph> start = stop);
      IloExpr expr(getEnv()), rev_expr(getEnv());
      unsigned count = 0;
      unsigned v_deg = 0;
      DEBUG3(std::cout << "forbidding cycle: ");
      do{
        const Vertex<AuxGraph>& u = boost::source(uv, aux);
        const Vertex<AuxGraph>& v = boost::target(uv, aux);
        const ScafVertex& old_u = translate.right.at(u);
        const ScafVertex& old_v = translate.right.at(v);
        const ScafVertexPair old_uv(old_u, old_v);
        const ScafVertexPair old_vu(old_v, old_u);

        DEBUG3(std::cout << I.get_edge_name(old_uv)<<" ");
        remove_edge(uv, aux);
        v_deg = boost::out_degree(v, aux);
        assert(v_deg <= 1);
        // if this outgoing arc is not to our initial arc, then add the next vertex to pending & remove the arc
        if(v_deg == 1) uv = *(out_edges(v, aux).first);
        
        // forbid the cycle and its reverse in y and z
        expr += vars.times_used.at(old_uv);
        expr -= vars.to_cycle_sink.at(old_v);
        // NOTE: instead of adding two constraints for both directions of the cycle, we add their sum since it's a stronger constraint
//        rev_expr += vars.times_used.at(old_vu);
//        rev_expr -= vars.to_cycle_sink.at(old_u);
        expr += vars.times_used.at(old_vu);

        ++count;
        DEBUG1(stop = v);
      } while(v_deg != 0);
      DEBUG1(assert(start == stop));
      DEBUG5(std::cout << " by allowing only <"<<count<<" of its edges (unless it's connected to t_c)");
      DEBUG3(std::cout << std::endl);
      add(expr < count);
//      add(rev_expr < count);
//      num_new_cuts += 2;
      num_new_cuts++;
#ifdef LAZY_ADD_ALL_CYCLES
    }// while(num_edges(aux))
#endif
    }// if(num_edges(aux))
    // Step 4: collect statistical information
    DEBUG2(std::cout << "=======   callback called, found "<<num_p<<" paths & added "<<num_new_cuts<<" new cuts ======"<<std::endl);
    stats.times_called++;
    stats.cuts_added += num_new_cuts;
    stats.time_spent.pause();
  } // main()



  // initialize the variables for the ILP formulation
  void populate_variables(IloModel& model,
                          IloEnv& env,
                          VAR_COLLECTION& vars,
                          const ILPOptions& opts
#ifdef STATISTICS
                          ,JumpStatistics& jump_stats
#endif
                          )
  {
    const Instance& I = vars.I;
    DEBUG5(std::cout << "creating x_{u,v}/y_uv"<<std::endl);
    for(auto e = I.get_edges(); e; ++e){
      const ScafVertex& u = I.source(*e);
      const ScafVertex& v = I.target(*e);
      const std::string uname = I[u].name;
      const std::string vname = I[v].name;
      const ScafVertexPair uv(u, v);
      const ScafVertexPair vu(v, u);
      
      // Step 1: add variables x_e = 1   <=>  e belogs to a solution path/cycle
      const std::string xname("x_" + uname + "," + vname);
      // Step 2: add variables y_ij = 1  <=> i->j is used in a solution path/cycle
      const std::string yname1("y_" + uname + "," + vname);
      const std::string yname2("y_" + vname + "," + uname);

      DEBUG5(std::cout << "creating "<<xname<< " & "<<yname1 << " & "<< yname2<<std::endl);
      vars.times_used_undir[uv] = IloNumVar(env, 0, 1, IloNumVar::Bool, xname.c_str());
      vars.times_used[uv] = IloNumVar(env, 0, 1, IloNumVar::Bool, yname1.c_str());
      vars.times_used[vu] = IloNumVar(env, 0, 1, IloNumVar::Bool, yname2.c_str());
    }

    DEBUG5(std::cout << "creating s/tp/tc interaction variables"<<std::endl);
    // Step 3: add variables y_s,i and y_i,tp and y_i,tc
    for(auto v = I.get_vertices(); v; ++v){
      const std::string name = I[*v].name;
      // on layer 0, come from s
      vars.from_source[*v] = IloNumVar(env, 0, 1, IloNumVar::Bool, ("y_s," + name).c_str());
      // on layer 1, go to tp
      vars.to_path_sink[*v] = IloNumVar(env, 0, 1, IloNumVar::Bool, ("y_" + name + ",tp").c_str());
      // on layer 0, go to tc
      vars.to_cycle_sink[*v] = IloNumVar(env, 0, 1, IloNumVar::Bool, ("y_" + name + ",tc").c_str());
    }

    if(!opts.ignore_contig_jumps){
      // constraint: contig jumps
      DEBUG4(std::cout << "creating contig-jump variables & expressions"<<std::endl);
      STAT(VertexPairToInt<RawScaffoldGraph> network_sizes);
      // Step 3: construct a contig jump network for uv
      for(auto e_it = I.get_non_matching_edges(); e_it; ++e_it){
        const ScafEdge& uv_edge = *e_it;
        const int static_length = I[uv_edge].length;
        if(static_length != NO_LENGTH){
          const ScafVertex& u = I.source(uv_edge);
          const ScafVertex& v = I.target(uv_edge);
          const UncertainLength uv_length = I.get_uncertain_length(uv_edge);
          const unsigned uv_multi = 1; // note that we are in the multiplicity-1 case!
          //const unsigned uv_multi = I[uv_edge].multiplicity;
          for(const ScafVertexPair uv: {ScafVertexPair(u,v), ScafVertexPair(v,u)}){
            // NOTE: the ILPContigJumpNetwork is only constructed to be integrated in the model
            // after integration it does not serve any more purpose and is discarded
            ILPContigJumpNetwork jump_network(I, uv, opts.jump_threshold_probability, uv_length);
            if(!jump_network.empty()){
              const std::string name = "jump_" + std::to_string(I.get_edge_name(uv));
              const IloNumVar& main_jump_var = vars.contig_jumps.emplace(PWC,
                  std::make_tuple(uv), std::make_tuple(env, 0, uv_multi, IloNumVar::Int, name.c_str())).first->second;
              jump_network.integrate_into_model(env, model, uv_multi, main_jump_var, vars.times_used);
            }
            STAT(if(jump_network.num_edges() > 10) std::cout << "adding size-"<<jump_network.num_edges()<<" network"<<std::endl;)
            STAT(network_sizes[uv] = jump_network.num_edges();)
          }
        }
      }
      STAT(jump_stats.jump_network_sizes.from_pairs(network_sizes);)
    }// if

  }

  template<typename Container>
  void add_sum_at_most_expression(IloEnv& env, IloRangeArray& c, const Container& list, const double max_value)
  {
    IloExpr expr(env);
    for(const auto& y : list) expr += y.second;
    c.add(expr <= max_value);
  }

  // initialize the constraints for the ILP formulation
  void populate_constraints(IloEnv& env, IloRangeArray& c, const VAR_COLLECTION& vars, const ILPOptions& opts){
    const Instance& I = vars.I;

    // constraint (1): of all the representations of ij, exactly x_ij should be chosen!
    DEBUG4(std::cout << "adding (1): of all the representations of e, exactly x_e should be chosen"<<std::endl);
    for(auto x : vars.times_used_undir){
      const ScafVertexPair& uv = x.first;
      const ScafVertexPair vu(uv.second, uv.first);

      DEBUG5(std::cout << "adding y_" << I[uv.first].name << "," << I[uv.second].name << std::endl);
      IloExpr expr(env);
      expr += vars.times_used.at(uv);
      DEBUG5(std::cout << "adding y_" << I[uv.second].name << "," << I[uv.first].name << std::endl);
      expr += vars.times_used.at(vu);

      // use the w variables (indicating that there is a contig jump over uv or vu) to
      // allow x = 1 even if neither uv nor vu have been chosen by y
      if(!opts.ignore_contig_jumps){
        if(contains(vars.contig_jumps, uv)) expr += vars.contig_jumps.at(uv);
        if(contains(vars.contig_jumps, vu)) expr += vars.contig_jumps.at(vu);
      }

      expr -= x.second;
      c.add(expr == 0);
    }

    // constraint (2): all matching-edges should be chosen!
    DEBUG4(std::cout << "adding (2): all matching-edges should be chosen"<<std::endl);
    for(auto x : vars.times_used_undir)
      if(I.is_matching_edge(x.first))
        c.add(x.second == 1);

    // constraint (4),(6): there should be at most sigma_p paths & sigma_c cycles
    DEBUG4(std::cout << "adding (4)&(6): there should be at most " << I.num_paths << " paths, "
                                                                   << I.num_cycles << " cycles, totaling at most "
                                                                   << I.num_objects << " objects"<<std::endl);
    add_sum_at_most_expression(env, c, vars.to_path_sink, I.num_paths);
    add_sum_at_most_expression(env, c, vars.to_cycle_sink, I.num_cycles);
    add_sum_at_most_expression(env, c, vars.from_source, I.num_objects);

    // constraint (3): flow conservation
    DEBUG4(std::cout << "adding (3): flow conservation"<<std::endl);
    for(auto u = I.get_vertices(); u; ++u){
      for(unsigned layer = 0; layer < 2; ++layer){
        IloExpr expr(env);
        for(auto e = I.get_incident(*u); e; ++e){
          const ScafVertexPair uv(I.source(*e), I.target(*e));
          // remember: layer 0 -> 1 contains only matching edges
          if(I[*e].is_matching_edge() == (layer == 0)){
            // flow goes to next layer along uv
            expr += vars.times_used.at(uv);
            DEBUG5(std::cout << " + y_" << I.get_edge_name(uv));
          } else {
            const ScafVertexPair vu(uv.second, uv.first);
            // flow comes from the other layer alog vu
            expr -= vars.times_used.at(vu);
            DEBUG5(std::cout << " - y_"<<I.get_edge_name(vu));
          }// if
        }// for
        if(layer == 0){
          // on layer 0, flow can come from s
          DEBUG5(std::cout << " - y_s,"<< I[*u].name);
          expr -= vars.from_source.at(*u);
          // on layer 0, flow can go to tc
          DEBUG5(std::cout << " + y_"<< I[*u].name<<",tc");
          expr += vars.to_cycle_sink.at(*u);
        } else {
          // on layer 1, flow can go to tp
          DEBUG5(std::cout << " + y_"<< I[*u].name<<",tp");
          expr += vars.to_path_sink.at(*u);
        }// if
        c.add(expr == 0);
      }// for
    }// for

    // constraint (7): if we have flow from u to tc, then we must have flow from s to u
    DEBUG4(std::cout << "adding (7): cycle conservation"<<std::endl);
    for(auto u = I.get_vertices(); u; ++u){
      IloExpr expr(env);
      expr += vars.from_source.at(*u);
      expr -= vars.to_cycle_sink.at(*u);
      c.add(expr >= 0);
    }

    // constraint (5): each subset S of vertices should contain at most |S|-1 endpoints of arcs
    //DEBUG5(std::cout << "adding (5): each subset S of vertices should contain at most |S|-1 endpoints of arcs"<<std::endl);
    // done via callback!

  }// function


  // compute the ILP formulation for a graph g
  void populate_model(IloModel& model, 
                      IloEnv& env,
                      VAR_COLLECTION& vars,
#ifdef STATISTICS
                      JumpStatistics jump_stats,
#endif
                      const ILPOptions& opts){
    const Instance& I = vars.I;
    IloRangeArray c(env);
    // construct variables, keeping note of the x_ij, y^l_ij and z^l_ij variables
    DEBUG3(std::cout << "creating variables"<<std::endl);
#ifdef STATISTICS
    populate_variables(model, env, vars, opts, jump_stats);
#else
    populate_variables(model, env, vars, opts);
#endif
    // add optimization function
    DEBUG3(std::cout << "creating optimization expression"<<std::endl);
    IloExpr opt_exp(env);
    for(auto e = I.get_edges(); e; ++e){
      const ScafVertex& u = I.source(*e);
      const ScafVertex& v = I.target(*e);
      // get the appropriate x_ij variable
      auto uv = vars.times_used_undir.find(ScafVertexPair(u,v));
      if(uv == vars.times_used_undir.end()) uv = vars.times_used_undir.find(ScafVertexPair(v,u));
      assert(uv != vars.times_used_undir.end());
      // add the term weight(e)*x_e
      opt_exp += (double)I[*e].weight * uv->second;
    }
    model.add(IloMaximize(env, opt_exp));
    // construct constraints
    DEBUG3(std::cout << "creating constraints"<<std::endl);
    populate_constraints(env, c, vars, opts);
    model.add(c);
  }

}}
