
#ifndef TREES_HPP
#define TREES_HPP

#include <queue>
#include "utils/instance.hpp"
#include "dp_solution.hpp"

using namespace boost;

namespace scaffold{ namespace solv { namespace DP {



  ScafEdge* get_edge_to_first_unvisited_neighbor(const ScafVertex& u, const unordered_set<ScafVertex >& visited, const Instance& I)
  {
    for(auto uv = I.get_incident(u); uv; ++uv)
      if(!contains(visited, I.target(*uv)))
        return new ScafEdge(*uv);
    return nullptr;
  }



  template<class Compare = std::less<size_t> >
  struct TreeDPinfos {
    // the table is indexed by \alpha and p
    typedef PathsAndCycles IndexType;
    typedef PCMinMatrix<Compare> entries_t;
    // the DP values for 0/1 edges incident with v in S
    entries_t *table;
    // a reminder of whether the matching edge has already been seen
    size_t childs_done;
    const size_t max_paths;
    bool seen_matching;
    Compare is_better;

    TreeDPinfos(const size_t _max_paths): table(new entries_t(1, _max_paths)), childs_done(0), max_paths(_max_paths), seen_matching(false) {
      table->put_empty(IndexType(0,0));
    }
    ~TreeDPinfos(){ delete table; }

    // combine our entry with the one of our child as described in the paper
    void combine_with_child(const TreeDPinfos& child, const bool is_matching_edge, const EdgeName& uv, const size_t uv_weight){
      entries_t *result = new entries_t(table->max);
      for(auto E_i = table->get_entries(); E_i.first != E_i.second; ++E_i.first){
        const auto& E = *E_i.first;
        const IndexType ap = E.first;
        for(auto cE_i = child.table->get_entries(); cE_i.first != cE_i.second; ++cE_i.first){
          const auto& cE = *cE_i.first;
          const IndexType cap = cE.first;
          DEBUG5(std::cout << "combining index ("<<(int)ap.c<<","<<ap.p<<") with ("<<(int)cap.c<<","<<cap.p<<") and the matching flag is "<<is_matching_edge<<" weight: "<<uv_weight<<std::endl);
          // if the subtree of u has 0 paths, then it cannot have an edge incident to u; likewise for the subtree s_v[j]
          assert(ap.c <= ap.p);
          assert(cap.c <= cap.p);

          if(is_matching_edge){
            // uv in M
            seen_matching = true;
            const size_t sum_p = 1 + (ap.p - ap.c) + (cap.p - cap.c);
            if(sum_p <= max_paths) result->put(IndexType(sum_p, ap.c),  *E.second, *cE.second);
          } else {
            // uv is not in M
            // first, we don't delete uv: then, ap.c and cap.c have to be 0
            if((ap.c == 0) && (cap.c == 0)){
              // combine [..] + [0,...]
              const size_t sum_p = ap.p + cap.p - (seen_matching ? 1 : 0);
              if(sum_p <= max_paths) {
                MatrixEntry* ME = new MatrixEntry(*E.second);
                *ME += *cE.second;
                // if we're looking for a MAXIMUM solution, then add the non-deleted edge
                if(is_better(1,0)){
                  ME->append_edge(uv);
                  ME->weight += uv_weight;
                }
                if(!result->put(IndexType(sum_p, 1), ME)) delete ME;
              }
            }
            // second, we delete uv
            const size_t sum_p = ap.p + cap.p;
            if(sum_p <= max_paths){
              MatrixEntry* ME = new MatrixEntry(*E.second);
              *ME += *cE.second;
              // if we're looking for a MINIMUM solution, then add the deleted edge
              if(is_better(0,1)){
                ME->append_edge(uv);
                ME->weight += uv_weight;
              }
              if(!result->put(IndexType(sum_p, ap.c), ME)) delete ME;
            }// if
          }// else
        } // for
      }// for
      delete table;
      table = result;
    }// function

    // << operator for solutions outputting a table of weights
    friend std::ostream& operator<<(std::ostream& os, const TreeDPinfos& infos){
      // write table
      for(size_t p = 0; p <= infos.max_paths; ++p){
        os << p<<":\t";
        try{ os << infos.table->at(IndexType(p,0)).weight << "\t"; } catch(...) { os << "\t"; };
        try{ os << infos.table->at(IndexType(p,1)).weight << "\t"; } catch(...) { os << "\t"; };
        os << std::endl;
      } // for
      return os;
    } // function

  };


  // get a queue of leaves in I and init their values in the DPmap
  template<class Compare = std::less<size_t> >
  void init_leaves(const Instance& I, std::queue<ScafVertex>& leaves, unordered_map<ScafVertex, TreeDPinfos<Compare> >& DPmap){
    DEBUG4(std::cout << "initializing leaves"<<std::endl);
    for(auto r = I.get_vertices(); r; ++r)
      if(I.degree(*r) == 1){
        // construct the entry for *r.first initializing the leaf data in the TreeDPinfos
        DPmap.emplace(std::piecewise_construct, std::make_tuple(*r), std::make_tuple(I.num_paths));
        leaves.push(*r);
      }//if
    DEBUG4(std::cout << "initialized "<<leaves.size()<<" leaves"<<std::endl);
  }//function

  // solve an instance of scaffolding on a tree
  template<class Compare = std::less<size_t> >
  void solve_on_tree(Instance& I, DynProgSolution<Compare>* const S){
    assert(I.get_infos().is_acyclic());

    unordered_map<ScafVertex, TreeDPinfos<Compare> > DPmap;
    unordered_set<ScafVertex > visited;

    // Step 1: get the leaves
    std::queue<ScafVertex > leaves;
    init_leaves(I, leaves, DPmap);

    // Step 2: for each leaf, update the infos upwards
    ScafVertex u;
    typename unordered_map<ScafVertex, TreeDPinfos<Compare> >::const_iterator u_infos;
    while(!leaves.empty()){
      DEBUG4(std::cout << "queue has "<<leaves.size()<<" items, next one is "<<( *I.g)[leaves.front()].name<<std::endl);
      u = leaves.front(); leaves.pop();
      // get the DP infos
      u_infos = DPmap.find(u);
      assert(u_infos != DPmap.end());
      DEBUG4(std::cout << "with DP entry: "<<std::endl<<u_infos->second<<std::endl);
      // u should have a single neighbor that is unvisited
      ScafEdge* to_parent = get_edge_to_first_unvisited_neighbor(u, visited, I);
      // if u does not have a parent, then u is the root, so break here
      if(!to_parent) break;
      const ScafVertex parent = I.target(*to_parent);
      // get the parents DP infos, constructing them if necessary
      // update the parent's table
      const bool is_matching_edge = I.is_matching_edge({u, parent});
      const EdgeName uv_name(I[u].name, I[parent].name);
      const size_t uv_weight = I[*to_parent].weight;
      // if the parent does not yet have a DP entry, create an empty one
      TreeDPinfos<Compare>& p_infos = DPmap.emplace(parent, I.num_paths).first->second;

      p_infos.combine_with_child(u_infos->second, is_matching_edge, uv_name, uv_weight);
      p_infos.childs_done++;
      // if all children of the parent have been considered, then add the parent to the queue
      if(p_infos.childs_done == I.degree(parent) - 1) leaves.push(parent);
      // finally, forget the infos of u & mark u visited
      visited.insert(u);
      DPmap.erase(u_infos);
      delete to_parent;
    }// while
    // now, u is the root, so construct a solution from it
    for(auto E = u_infos->second.table->get_entries(); E.first != E.second; ++E.first)
      S->put(PathsAndCycles(E.first->first.p,0), E.first->second);
  }// function




}}}// namespace


#endif

