
#ifndef TW_DP_HPP
#define TW_DP_HPP

#include <stack>
#include <vector>
#include <queue>
#include <boost/graph/subgraph.hpp>

#include "TOL/GreedyFillIn.h"
#include "TOL/TreeDecomposition.h"

#include "utils/utils.hpp"
#include "utils/graph_typedefs.hpp"
#include "utils/solution.hpp"
#include "utils/instance.hpp"
#include "utils/convert.hpp"
#include "utils/td_subgraph.hpp"
#include "utils/scaffolding_utils.hpp"


#include "preprocess/pp_pendant_match.hpp"
#include "solv/dp_table.hpp"
#include "solv/trees.hpp"

using namespace boost;

namespace scaffold{ namespace solv { namespace DP{

  using scaffold::operator<<;

  // get a TOL tree decomposition of I
  TOL::TreeDecomposition* get_tree_decomp(const Instance& I, VTranslate* vtranslate = NULL)
  {
    // step 1: prepare the TOL graph & the elimination ordering
    TOL::Graph TOLg(0, I.num_vertices());
    convert(I, TOLg, vtranslate);
    // Step 2: run the greedy fill-in
    TOL::EliminationOrdering eo(TOLg);
    TOL::GreedyFillIn gfi(eo);
    gfi.run(TOLg);
    // Step 3: return the result
    return new TOL::TreeDecomposition(TOLg, eo);
  }




  // process an edge of the tree decomposition
  void DP_treat_edge(DP_Table& child_table, DP_Table& parent_table, ScafEdgeSet& already_introduced, const Instance& I)
  {
    DEBUG2(std::cout << "treating edge from child with bag ";
           for(auto& i: child_table.index->left) std::cout <<  I[i.first].name<<" ";
           std::cout << " to parent with bag ";
           for(auto& i: parent_table.index->left) std::cout <<  I[i.first].name<<" ";
           std::cout << std::endl);
    // Step 0: get all vertices that will be forgotten (i.e. all vertices in child that are not in parent)
    ScafVertexSet to_forget;
    for(auto& vi: child_table.index->left)
      if(!contains(parent_table.index->left, vi.first))
        to_forget.emplace(vi.first);

    // Step 1: introduce all edges incident to vertices that will be forgotten
    for(const ScafVertex& v: to_forget){
      DEBUG3(std::cout << "introducing edges incident with "<< g[v].name<<":"<<std::endl);
      for(auto r = I.get_incident(v); r; ++r)
        // introduce the edge only if it is not already introduced and if the target is also in the bag
        if(!contains(already_introduced, *r) && (contains(child_table.index->left, I.target(*r)))){
          child_table.intro_edge(*r);
          already_introduced.insert(*r);
          already_introduced.insert(I.reverse_edge(*r).first);
        }
    }
        
    // Step 2: forget all vertices in child that are not in parent
    child_table.forget_vertices(to_forget);

    // Step 3: introduce all vertices in parent that are not in child
    ScafVertexSet to_introduce;
    for(auto& vi: parent_table.index->left){
      const ScafVertex v = vi.first;
      const auto& child_vset = child_table.index->left;
      if(!contains(child_vset, v)) to_introduce.emplace(v);
    }
    child_table.intro_vertices(to_introduce);

    // Step 4: merge with the parent
    DEBUG3(std::cout << "joining tables..."<<std::endl);
    parent_table.join(child_table);

    DEBUG5(std::cout << "final table of parent:"<<std::endl<<parent_table<<std::endl);

    DEBUG2(
      size_t size = parent_table.size();
      if(size > 1000000) std::cout << "table size: "<<parent_table.size()<<" entries"<<std::endl;
    )
  }


  // an instance comparator based on number of vertices, by default, smaller instances are considered "<"
  template<class Compare = std::less<size_t> >
  struct Compare_Instance_Size
  {
    const Compare is_better;
    bool operator()(const Instance* I1, const Instance* I2)
    {
      return is_better(I1->num_vertices(), I2->num_vertices());
    }
  };

  // solv an instance of scaffolding using the dynamic programming on tree decompositions
  void TW_solv(Instance& I, DynProgSolution<std::greater<size_t> >& S){
    // Step 0: apply preprocessing
    DEBUG3(std::cout << "running preprocessing..."<< std::endl);
    preprocess::pp_matching_pendant(I);
    DEBUG3(std::cout << "after preprocessing: "<<boost::num_vertices(*I.g)<<" vertices & "<<boost::num_edges(*I.g)<<" edges"<<std::endl);
    // Step 1: run for each component separately
    std::priority_queue<Instance*, std::vector<Instance*>, Compare_Instance_Size<std::greater<size_t> > > components;
    while(true){
      Instance *J = I.split_off_component();
      if(J) components.push(J); else break;
    }
    DEBUG3(std::cout << "split off "<<components.size()<< " connected components for independent solving"<< std::endl);
    if(!components.empty()){
      // add I as final compoenent
      components.push(&I);
      DEBUG2(std::cout << "instance has "<<components.size()<<" components"<<std::endl);
      // prepare S for merging the component solutions
      S.put_empty(PathsAndCycles(0,0));
      // keep track of how many paths any cycles got used in previous components
      PathsAndCycles used_up(0,0);
      // treat components in ascending order of size
      while(!components.empty()){
        Instance* J = components.top(); components.pop();
        J->num_paths -= used_up.p;
        J->num_cycles -= used_up.c;

        DEBUG3(std::cout << "next component has "<<J->num_vertices()<<" vertices & we're looking for "<<J->num_paths<<" paths & "<<J->num_cycles<<" cycles"<<std::endl);
      
        DynProgSolution<std::greater<size_t> > solJ(J->num_paths, J->num_cycles);
        TW_solv(*J, solJ);

        if(J != &I) delete J;

        if(!solJ.no_solutions()){
          DEBUG4(std::cout << " joining solution" << solJ<<std::endl);
          S.combine_disjoint_union(solJ);
          // update the number of paths and cycles used
          used_up = S.smallest_index();
          DEBUG4(std::cout << " S now has "<<S.size()<<" entries, using up "<<used_up<<" paths & cycles "<<std::endl);
        } else {
          // if either component does not have a solution, return failure
          S.invalidate();
          return;
        }
      }
      return;
    }
    // solve base cases
    if(I.get_infos().is_max_deg_two()) {
      DEBUG3(std::cout << "it's max-deg 2"<<std::endl);
      treat_deg_two(I, &S);
      return;
    }
    if(I.get_infos().is_acyclic()){
      DEBUG3(std::cout << "it's a tree"<<std::endl);
      solve_on_tree(I, &S);
      return;
    }


    // Step 2: get a tree decomposition
    VTranslate* vtranslate = new VTranslate();
    const TOL::TreeDecomposition *TD = get_tree_decomp(I, vtranslate);
    const TOL::Tree t(TD->tree());
    DEBUG2(std::cout << "using tree decomposition of width "<<TD->width()<<std::endl);
    // Step 3: setup data structures
    // Step 3b: stack of child iterators, together with the table pointer for this node
    std::stack<std::pair<std::pair<TOL::Tree::const_child_iterator, TOL::Tree::const_child_iterator>, DP_Table*> > child_stack;
    // Step 3c: note who has been visited
    unordered_set<TOL::Tree::Node> visited;
    // Step 3d: note set of introduced edges
    ScafEdgeSet introduced_edges(0, I.get_graph());
    
    // make sure the root has children (if not, the graph is a clique)
    const TOL::Tree::Node r = t.root();
    assert(!t.is_leaf(r));

    // Step 5: go through the queue in-order to minimize the # of tables we have to keep in memory at the same time
    child_stack.emplace(std::make_pair(t.all_children(r), new DP_Table(I, TD->subset(r), vtranslate)));
    visited.insert(r);
    do{
      const TOL::Tree::Node& u = *(child_stack.top().first.first);

      // if we haven't seen u before, then push the children on the stack and dive into the subtree
      if(!contains(visited, u)){
        child_stack.emplace(make_pair(t.all_children(u), new DP_Table(I, TD->subset(u), vtranslate)));
        // if u is a leaf then initialize its DP table as leaf table
        if(t.is_leaf(u)) child_stack.top().second->init_as_leaf();

        visited.insert(u);
        DEBUG1(
          std::cout << "adding to stack: "<<u<<" with VSet:\t";
          std::list<ScafVertex> v_lst;
          child_stack.top().second->get_VSet(v_lst);
          for(auto u: v_lst) std::cout << I.get_name(u) << ' ';
          std::cout << "\t\t children: ";
          for(auto ra = child_stack.top().first.first; ra != child_stack.top().first.second; ++ra) std::cout << *ra<< ' ';
          std::cout << std::endl;
        );

      } else ++child_stack.top().first.first; // otherwise, we came back from a dive into the subtree, so prepare to dive to the next child

      // if all children have been considered, merge our table with the parent table
      std::pair<TOL::Tree::const_child_iterator, TOL::Tree::const_child_iterator>& childs = child_stack.top().first;
      if(childs.first == childs.second){
        // if there is only the root entry on the stack, then break the while-loop
        if(child_stack.size() > 1){
          DP_Table* child_table = child_stack.top().second;
          child_stack.pop();
          DP_Table* parent_table = child_stack.top().second;
          // combine the DP tables
          DP_treat_edge(*child_table, *parent_table, introduced_edges, I);
          // afterwards, delete the child's table since we won't need it anymore
          delete child_table;
          // NOTE: if the table we produced is, in fact, empty, then there is no solution!
          if(parent_table->empty()){
            while(!child_stack.empty()){
              DP_Table* current_table = child_stack.top().second;
              current_table->clear();
              delete current_table;
              child_stack.pop();
            }
            S.invalidate();
            return;
          }
        } else break;
      }// if
    } while(true);
    // finally read the result from the only remaining table on the stack (should be the root table)
    // to this end, merge with the empty bag
    DP_Table root_table(I, ScafVertexSet());
    DP_Table* last_table = child_stack.top().second;
    DP_treat_edge(*last_table, root_table, introduced_edges, I);
    delete last_table;
    root_table.read_solution(S);
  }// function

}}}// namespace


#endif

