
/** \file ambigous_paths.hpp
 * tools to remove ambigous paths in a solution graph with multiplicities
 *
 * A solution graph with multiplicities is not necessarily degree-2, but may contain
 * multiforcations. In this case, it may not be obvious how to split the graph into
 * paths and cycles. Indeed, it may not even be unambigous. It has been shown that
 * such a graph can be unambigously split into paths and cycles if and only if it does
 * not contain "ambigous paths". An ambigous path is an alternating path between two
 * vertices u, v such that both u and v are incident to edges of different multiplicity.
 *
 * Ambigous paths can be removed by deleting edges, but minimizing the number of
 * deleted edges to make a solution graph unambigous is basically VERTEX COVER (and,
 * thus, NP-hard). Here, we use a simple heuristic for the problem that just selects 
 * one of the two necessary vertices for each edge & removes all covered edges.
 */

#pragma once


#include "utils/profiling.hpp"
#include "utils/stats.hpp"
#include "utils/scaffolding_typedefs.hpp"
#include "utils/predicates.hpp"

#ifdef HTD
#include "treeWidthLin/tw_dp_lin.hpp"
#endif

#ifdef CPLEX
#include "utils/ilp_common.hpp"
#endif

namespace scaffold {


  //! check locally whether u may be an endpoint of an ambigous path in sg
  /** returns the multiplicity of the ambiguous path or 0 if no ambiguous path starts in u */
  unsigned maybe_ambigous_path_extremity_local(const ScafVertex &u, const ScaffoldGraph &sg)
  {
    const unsigned deg = sg.degree(u);
    if(deg == 1) return 0;

    const ScafEdge &uv = sg.incident_matching_edge(u);
    const unsigned path_mult = sg[uv].multiplicity;
    if(deg > 2) return path_mult;

    const ScafEdge &uw = *(sg.get_incident_non_matching(u));
    return (path_mult != sg[uw].multiplicity) ? path_mult : 0;
  }

  //! return (v, true) iff u is an endpoint of an uniform path & the other endpoint is v
  //! if must_be_ambiguous return (v,true) iff  uv is ambiguous
  std::pair<ScafVertex, bool> is_ambigous_path_extremity(ScafVertex u, const ScaffoldGraph &sg)
  {
    const unsigned path_mult = maybe_ambigous_path_extremity_local(u, sg);
    if(path_mult) {
      while(true) {
        const ScafEdge &uv = sg.incident_matching_edge(u);
        const ScafVertex &v = sg.target(uv);
        DEBUG4(std::cout << "Matching edge: " << sg[u].name << "->" << sg[v].name << std::endl);
        if(sg[uv].multiplicity != path_mult) return {v, false};
        // do the easy checks:
        // v has >1 non-matching edge? We have an ambigous path!
        if(sg.degree(v) > 2) return {v, true};
        // v has no non-matching edges? The path is not ambigous!
        if(sg.degree(v) == 1) return {v, false};

        // get the unique non-matching edge incident to v
        const ScafEdge &vw = *(sg.get_incident_non_matching(v));
        // check its multiplicity
        if(sg[vw].multiplicity != path_mult) return {v, true};
        // advance along the possible ambigous path
        u = sg.target(vw);
      }
    } else return {u, false};
  }

  //! find and destroy ambigous paths in sg by deleting edges
  /** this repeatedly finds an ambigous path and deletes edges at one of its endpoints to make it unambigous */
  void kill_ambigous_paths_brutal(ScaffoldGraph &sg, LinearizationStatistics &stats, const bool count_edges = true)
  {
    for(auto u_iter = sg.get_vertices(); u_iter; ++u_iter) {
      const ScafVertex &u = *u_iter;
      // if u is an extremity of an ambigous path, kill its non-matching edges
      const std::pair<ScafVertex, bool> result = is_ambigous_path_extremity(u, sg);
      if(result.second) {
        DEBUG4(std::cout << "found ambigous path " << sg[u].name << "-->" << sg[result.first].name << std::endl);
        // delete all non-matching edges incident with u
        stats.nb_cuts += (count_edges ? sg.degree(u) - 1 : 1);
        stats.weight_cut += sg.sum_multiplicity_non_matching(u);
        sg.delete_non_matching(u);
      }
    }
  }


  //! Apply the factor 2 approximation algorithm to destroy ambiguous paths
  /** this repeatedly finds the endpoint of an ambiguous path with the minimum sum of the weight of the incident edge
   * and deletes edges of it endpoint*/
  void kill_ambigous_paths_approx(ScaffoldGraph &sg, LinearizationStatistics &stats)
  {
    /* We maintain a list of extremities of remaining ambiguous paths.
     * The list is sorted in descending order according to sum_multiplicities_u*/
    std::list<Extremity_Ambiguous_Path> list_ambiguous_paths;

    /* extremities_map contains the positions of the extremities in the list to ensure a constant time access.*/
    std::map<const ScafVertex, std::list<Extremity_Ambiguous_Path>::iterator> extremities_map;

    for(auto v_iter = sg.get_vertices(); v_iter; ++v_iter) {
      const ScafVertex &u = *v_iter;
      const std::pair<ScafVertex, bool> result = is_ambigous_path_extremity(u, sg);
      const ScafVertex &v = result.first;
      if(result.second) {
        unsigned sum = sg.sum_multiplicity_non_matching(u);
        /* list_ambiguous_paths is sorted,
         * we insert a new ambiguous path such that list_ambiguous_paths remains sorted*/
        auto it = list_ambiguous_paths.insert(std::lower_bound(list_ambiguous_paths.begin(),
                                                               list_ambiguous_paths.end(),
                                                               sum,
                                                               [](const Extremity_Ambiguous_Path &a, int sum) {
                                                                 return std::get<0>(a) <= sum;
                                                               }),
                                              std::move(Extremity_Ambiguous_Path(sum, u, v)));
        extremities_map.emplace(u, it);
      }
    }
    std::cout << std::endl;

    while(!list_ambiguous_paths.empty()) {
      Extremity_Ambiguous_Path a = list_ambiguous_paths.front();

      ++stats.nb_cuts;
      stats.weight_cut += std::get<0>(a);

      list_ambiguous_paths.pop_front();
      list_ambiguous_paths.erase(extremities_map[std::get<2>(a)]);
      extremities_map.erase(std::get<2>(a));
      sg.delete_non_matching(std::get<1>(a), list_ambiguous_paths, extremities_map);
      extremities_map.erase(std::get<1>(a));

    }
  }

  /**
   * \brief Apply the tree decomposition method to linearize the graph
   */
  void kill_ambigous_paths_tw(ScaffoldGraph &sg, LinearizationStatistics &stats, bool count_edges)
  {
#ifdef HTD
    DP_LIN::ContractedGraph g(sg);
    auto *a = g.getNiceTreeDecomposition();
    
    DEBUG1(std::cout << "Solution: {";)
      auto sol = DP_LIN::getCollectionSignatureRoot(sg, g, a, a->root(),count_edges).getSolution();
    for (auto v : sol) {
      stats.nb_cuts += (count_edges ? sg.degree(v) - 1 : 1);
      stats.weight_cut += sg.sum_multiplicity_non_matching(v);

      DEBUG1(std::cout << sg[v].name << " ";)
      sg.delete_non_matching(v);
    }
    DEBUG1(std::cout << "}" << std::endl;)
#else
    std::cout << "You must compile the executable with -DDP_LIN=ON to access this feature" << std::endl;
    exit(EXIT_FAILURE);
#endif

  }


#ifdef CPLEX

  struct VarPair {
    IloNumVar remove;
    IloNumVar remove_all_neighbors;

    // construct
    VarPair(const IloEnv &env, const std::string &vertex_name) :
      remove(env, 0, 1, IloNumVar::Bool, (vertex_name + "r").c_str()),
      remove_all_neighbors(env, 0, 1, IloNumVar::Bool, (vertex_name + "rn").c_str())
    {
    }
  };

  typedef unordered_map<const ScafVertex, const VarPair> VertexVarPairMap;
  typedef boost::unordered_map<UnorderedVertexPair<RawScaffoldGraph>, IloNumVar> UEdgeVarMap;

  //! add variables of edges around a given vertex u and their dependency to u.remove
  void add_incident_edge_variables(const ScaffoldGraph &sg,
                                   const ScafVertex &u,
                                   const VarPair &u_vars,
                                   UEdgeVarMap &edge_vars,
                                   const IloEnv &env,
                                   IloRangeArray &constraints,
                                   IloExpr &opt_expr,
                                   const bool use_weight)
  {
    for(auto e_iter = sg.get_incident_non_matching(u); e_iter; ++e_iter) {
      const ScafEdge &e = *e_iter;
      const EdgeName e_name = sg.get_edge_name(e);

      const auto res = edge_vars.emplace(std::piecewise_construct,
                                         std::make_tuple(u, sg.target(e)),
                                         std::make_tuple(env, 0, 1, IloNumVar::Bool,
                                                         (e_name.first + "->" + e_name.second + "(r)").c_str()));

      const IloNumVar &e_var = res.first->second;
      constraints.add(e_var - u_vars.remove >= 0);
      if(res.second)
        opt_expr += (use_weight ? sg[e].multiplicity : 1) * e_var;
    }
  }

#endif

  //! use an ILP to find the least lossy resolution of ambigous paths
  /** if count_edges is set, then minimize the deleted edges instead of the vertex at which to delete
   * otherwise, minimize the number of vertices that are cleared */
  void kill_ambigous_paths_ILP(ScaffoldGraph &sg, LinearizationStatistics &stats, const bool count_edges = true)
  {
#ifdef CPLEX
    // initialize ILP model
    IloEnv env;
    IloModel model(env);
    IloRangeArray constraints(env);
    VertexVarPairMap vars;
    UEdgeVarMap edge_vars;

    timer opt_timer;
    opt_timer.start();
    IloExpr opt_expr(env);
    unsigned count = 0;
    for(auto v_iter = sg.get_vertices(); v_iter; ++v_iter) {
      const ScafVertex &u = *v_iter;
      // note: if u--v is an ambigous path, then this is the only ambigous path incient to u & v
      const std::pair<ScafVertex, bool> result = is_ambigous_path_extremity(u, sg);

      const ScafVertex &v = result.first;

      if((result.second) && (sg[u].index < sg[v].index)) {
        ++count;
        DEBUG1(
          std::cout << "found ambigous path #" << count << ": " << sg[u].name << " (NH:";
          for(auto e_it = sg.get_incident(u); e_it; ++e_it) std::cout << " " << sg[sg.target(*e_it)].name;
          std::cout << ")\t--> " << sg[result.first].name << " (NH:";
          for(auto e_it = sg.get_incident(result.first); e_it; ++e_it) std::cout << " " << sg[sg.target(*e_it)].name;
          std::cout << ")" << std::endl;
        );
        const VarPair &u_vars = vars.DEEP_EMPLACE_TWO(u, env, sg[u].name).first->second;
        const VarPair &v_vars = vars.DEEP_EMPLACE_TWO(v, env, sg[v].name).first->second;
        // in order to destroy this ambigous path, we have 4 possibilities:
        // 1. remove all incident non-matching edges of u
        // 2. remove all incident non-matching edges of v
        // 3. remove all incident non-matching edges of all vertices in N(u)
        // 3. remove all incident non-matching edges of all vertices in N(v)
        IloExpr uv(env);
        uv += u_vars.remove;
        uv += u_vars.remove_all_neighbors;
        uv += v_vars.remove;
        uv += v_vars.remove_all_neighbors;
        constraints.add(uv >= 1);

        if(count_edges) {
          add_incident_edge_variables(sg, u, u_vars, edge_vars, env, constraints, opt_expr, true);
          add_incident_edge_variables(sg, v, v_vars, edge_vars, env, constraints, opt_expr, true);
        } else
          for(const VarPair &x_vars: {u_vars, v_vars}) opt_expr += x_vars.remove;
      }
    }

    // add consistency constraints: remove_all_neighbors(u) = 1 <==> remove(v) = 1 for all neighbors v of u
    DEBUG3(std::cout << "adding linearization consistency constraints" << std::endl);
    for(const auto &i: vars) {
      const ScafVertex &u = i.first;
      const VarPair &u_vars = i.second;
      assert(sg.degree(u) > 1);

      IloExpr exp(env);
      bool non_ambiguous_neighbor = false;
      for(auto e_iter = sg.get_incident_non_matching(u); e_iter; ++e_iter) {
        const auto v_var_iter = vars.find(sg.target(*e_iter));
        if(v_var_iter == vars.end()) {
          non_ambiguous_neighbor = true;
          break;
        } else exp += v_var_iter->second.remove;
      }
      // if any one of the neighbors of u is not an end of an ambigous path, then it's better to remove all incident to u instead
      if(!non_ambiguous_neighbor)
        constraints.add(exp - (int) (sg.degree(u) - 1) * u_vars.remove_all_neighbors >= 0);
      else
        constraints.add(u_vars.remove_all_neighbors == 0);
    }

    DEBUG3(std::cout << "adding constraints and target to the model..." << std::endl);
    model.add(constraints);
    model.add(IloMinimize(env, opt_expr));
    IloCplex cplex(model);

    if(!cplex.solve()) {
      env.error() << "Failed to linearize" << std::endl;
      exit(EXIT_FAILURE);
    } else {
      opt_timer.stop();
      env.out() << "CPU-seconds spent linearizing: " << opt_timer.seconds_passed() << std::endl;
      DEBUG2(if(count_edges) env.out() << "will remove " << cplex.getObjValue() << " edges: " << std::endl;
             else env.out() << "will clear " << cplex.getObjValue() << " vertices: " << std::endl);

    }

    // implement the solution in sg
    for(const auto &i: vars) {
      if(is_true(cplex.getValue(i.second.remove))) {
        DEBUG2(if(count_edges) for(auto e = sg.get_incident_non_matching(i.first); e; ++e) std::cout << sg.get_edge_name(*e) << " ";
               else std::cout << sg[i.first].name << " ");
        ++stats.nb_cuts;
        stats.weight_cut += sg.sum_multiplicity_non_matching(i.first);
        sg.delete_non_matching(i.first);
      }
    }
    std::cout << std::endl;
    DEBUG2(std::cout << std::endl;);
#else
    std::cout << "You must compile the executable with -DCPLEX=ON to access this feature"<<std::endl;
#endif

  }

}//namespace

