

/** \file command_line.hpp
 * command-line interface
 * handling parameters and options
 */

#pragma once

#include <string>

#include <boost/program_options/variables_map.hpp>
#include <boost/program_options/parsers.hpp>

#include "utils/utils.hpp"
#include "utils/exceptions.hpp"


namespace po = boost::program_options;

//! options that configure the scaffolder
struct GeneralOptions{
    //! filename of the input graph
    std::string in_graph_filename;

    //! filename of the input sequence file in fasta format
    std::string in_seq_filename;

    //! filename of a file to export the cplex model to
    std::string out_model_filename;

    //! filename of a file to export the solution graph to
    std::string out_graph_filename;

    //! filename of a file to export solution sequences to
    std::string out_seq_filename;

    //! cut-off threshold: remove all links between contigs whose weight is strictly below this threshold
    unsigned cut_off;

    //! maximum number of objects (paths or cycles) that a solution graph should consist of
    unsigned objects;
    //! maximum number of paths that a solution graph should consist of
    unsigned paths;
    //! maximum number of cycles that a solution graph should consist of
    unsigned cycles;

    //! linearize the solution graph after scaffolding
    bool do_linearization;

    //! scaffold the solution graph before linearizing
    bool do_scaffolding;

    //! pairwise overlap all contigs to find non-contig edges that have been missed
    bool find_missed_pairs;

    //! return true iff the arguments make sense
    /** f.ex. we need a fasta file with contig sequences if we want to output the sequences of the resulting contigs */
    void sanity_check() const
    {
      if((in_graph_filename == "") || !file_exists(in_graph_filename))
        throw except::invalid_options("I have no input file to work with, use --fn <file> to give me a scaffold graph.");
      if(out_seq_filename != "" && !file_exists(in_seq_filename))
        throw except::invalid_options("I want to output fasta sequences to '" + out_seq_filename + "' but I cannot open the sequence database '" + in_seq_filename + "' (given with --is) for reading");
      if(find_missed_pairs && !file_exists(in_seq_filename))
        throw except::invalid_options("I want to pairwise overlap contigs but I cannot open the sequence database '" + in_seq_filename + "' (given with --is) for reading");

    }
};

//! return the value of an option or the given default if the value has not been supplied
template<typename T>
const T& get_option(const po::variables_map& vm, const std::string& option_name, const T& default_value){
  if(vm.count(option_name))
    return vm[option_name].as<T>();
  else
    return default_value;
}

po::options_description initialize_options(const char* const program_name)
{
  // Declare the supported options.
  po::options_description desc((std::string)"syntax: " + program_name + " [options] --fn <scaffold graph file>\nwhere valid options are");
  desc.add_options()
          ("help", "produce help message")
          ("fn", po::value<std::string>(), "file to read scaffold graph from")
          ("is", po::value<std::string>(), "file to read contig sequences from (required for --os)")
          ("os", po::value<std::string>(), "file to output solution sequences to (fasta format) (requires --is)")
          ("og", po::value<std::string>(), "file to output solution graph to (dot format)")
          ("om", po::value<std::string>(), "file to output CPLEX model to (see cplex::exportModel()) ")
          ("c", po::value<unsigned>(), "allowed number of cycles (default: 0 if --o is not present, #objects - #paths if --o and --p are present, #objects if only --o is present)")
          ("p", po::value<unsigned>(), "allowed number of paths (default: 0 if --o is not present, #objects - #cycles if --o and --c are present, #objects if only --o is present)")
          ("o", po::value<unsigned>(), "allowed number of objects (paths or cycles) (default: #paths + #cycles)")
          ("co", po::value<unsigned>(), "cut off all edges whose weight is strictly below the argument (default: 0)")
          ("fo", "before scaffolding, scan contigs to find overlaps and insert missing non-contig ""edges accordingly (default: false)")
          ("nl", "no linearization after scaffolding")
          ("ns", "no scaffolding before linearization")
          ;
  return desc;
}

po::variables_map parse_arguments(const int argc, const char** argv, const po::options_description& desc)
{
  po::variables_map vm;
  po::store(po::parse_command_line(argc, argv, desc), vm);
  po::notify(vm);
  return vm;
}

//! process the program options given on the command line
GeneralOptions handle_options(const po::variables_map& vm, po::options_description& desc)
{
  if(vm.count("help")) { std::cout << desc << "\n"; exit(EXIT_SUCCESS); }

  GeneralOptions options;
  options.in_graph_filename  = get_option<std::string>(vm, "fn", "");
  options.out_graph_filename = get_option<std::string>(vm, "og", "");
  options.in_seq_filename    = get_option<std::string>(vm, "is", "");
  options.out_seq_filename   = get_option<std::string>(vm, "os", "");
  options.out_model_filename = get_option<std::string>(vm, "om", "");

  const bool o_present = vm.count("o");
  const bool p_present = vm.count("p");
  const bool c_present = vm.count("c");
  options.cycles          = get_option<unsigned>(vm, "c", 0);
  options.paths           = get_option<unsigned>(vm, "p", 0);
  options.objects         = get_option<unsigned>(vm, "o", options.paths + options.cycles);

  if(o_present){
    if(p_present && !c_present) options.cycles = options.objects - std::min(options.objects, options.paths);
    if(!p_present && c_present) options.paths = options.objects - std::min(options.objects, options.cycles);
    if(!p_present && !c_present) options.paths = options.cycles = options.objects;
  }

  options.cut_off         = get_option<unsigned>(vm, "co", 0);

  options.find_missed_pairs = vm.count("fo");

  options.do_linearization = !vm.count("nl");
  options.do_scaffolding = !vm.count("ns");

  try{
    options.sanity_check();
  } catch(except::invalid_options& e){
    std::cout << "Command line option error: "<<e.what() << std::endl;
    std::cout << desc << std::endl;
    exit(EXIT_FAILURE);
  }

  return options;
}


