

/** \file filters.hpp
 * collection of filters to work with boosts filtered graphs (see predicates.hpp)
 */

#ifndef FILTERS_HPP
#define FILTERS_HPP

#include "utils/utils.hpp"

namespace scaffold{ namespace filters {

  //! dummy filter to always say YES
  template<class Object>
  struct yes_filter{
    bool operator()(const Object& x) const { return true; }
  };

  //! filter to work with filtered graphs that is based on a set of vertices, passed to the constructor
  /** if invert = true is passed to the constructor, then the selection is inverted, filtering elements _NOT_ in the filter_set */
  template<class Set>
  struct contained_filter{
    const Set& filter_set;
    const bool invert;

    //! constructor
    contained_filter(const Set&  _filter_set, const bool _invert = false):
      filter_set(_filter_set), invert(_invert) {}
    
    // operator() comparing two keys
    bool operator()(const typename Set::key_type& e) const {
      return contains(filter_set, e) != invert; 
    }
  };

  //! filter to work with filtered graphs that is based on a set of VertexPairs, passed to the constructor
  /** if invert = true is passed to the constructor, then the selection is inverted, filtering elements _NOT_ in the filter_set */
  template<class Graph, class Set>
  struct VP_contained_filter{
    const Set& filter_set;
    const Graph& g;
    bool invert;

    //! constructor initalizing filter_set and invert
    VP_contained_filter(const Set& _filter_set, const Graph& _g, const bool _invert = false):
      filter_set(_filter_set), g(_g), invert(_invert) {}
    
    bool operator()(const Edge<Graph>& e) const {
      return contains(filter_set, Set::value_type(boost::source(e, g), boost::target(e, g))) != invert; 
    }
  };


  //! filter to work with filtered graphs that is based on a a mapping of vertices: filter all vertices mapped to some initial X
  /** if invert = true is passed to the constructor, then the selection is inverted, filtering elements _NOT_ in the filter_map */
  template<class Map>
  struct map_filter{
    const Map& filter_map;
    const typename Map::mapped_type& mapped_to;
    bool invert;

    //! constructor initializing the map, the element and inversion flag
    map_filter(const Map& _filter_map, const typename Map::mapped_type& _mapped_to, const bool _invert = false):
      filter_map(_filter_map), mapped_to(_mapped_to), invert(_invert) {}

    bool operator()(const typename Map::key_type& e) const {
      return (filter_map.at(e) == mapped_to) != invert;
    }
  };


}}

#endif
