
#pragma once 

#include <vector>

//#include "utils/set_queue.hpp"
#include "utils/graph_typedefs.hpp"
#include "utils/scaffold_graph.hpp"
#include "utils/ilp_common.hpp"


namespace scaffold{ 

  struct JumpEdgeProperty{
    unsigned index;
  };
  typedef boost::adjacency_list<
      boost::hash_setS, // OutEdgeList
      boost::hash_setS,     // VertexList
      boost::bidirectionalS, // (bi)directed ?
      boost::no_property, // vertex property
      JumpEdgeProperty  // edge property
    > JumpGraph;

  typedef Vertex<JumpGraph> JumpVertex;
  typedef VertexPair<JumpGraph> JumpVertexPair;
  typedef Edge<JumpGraph> JumpEdge;



  // a contig jump network is a subgraph of contigs whose total length fits between two contigs
  class ILPContigJumpNetwork
  {
    const ScaffoldGraph& sg;
    JumpGraph jump_graph;
    VTranslateMap<JumpGraph, RawScaffoldGraph> translate;
    const ScafVertexPair jumped_edge;
    const VertexPair<JumpGraph> translated_jumped_edge;
    const UncertainLength& gap_length;
    const float jump_threshold_probability;
    //! for each multiplicity, the network can be used to prove existance of one solution path from u to v
    /** we will use edge indices as keys in this map */
    std::vector<std::vector<IloNumVar> > edge_to_vars;

    //! return whether a given uncertain length fits in the gap
    bool fits_in_gap(const UncertainLength& len) const
    {
      return (len - gap_length).prob_of_picking_at_most(0) >= jump_threshold_probability;
    }

    //! add a vertex to the JumpGraph, registering its copy in the RawScaffoldGraph
    const JumpVertex add_vertex(const ScafVertex& u)
    {
      JumpVertex result = boost::add_vertex(jump_graph);
      translate.emplace(result, u);
      return result;
    }

    //! add an edge to the JumpGraph, registering its Ilo Variables in edge_to_vars
    const std::pair<JumpEdge, bool> add_edge(const JumpVertex& u_prime,
                                             const JumpVertex& v_prime,
                                             const ScafVertex& u,
                                             const ScafVertex& v)
    {
      // step 1: create the actual edge
      const auto create_result = boost::add_edge(u_prime, v_prime, jump_graph);

      if(create_result.second){
        const JumpEdge& e = create_result.first;
        jump_graph[e].index = boost::num_edges(jump_graph) - 1;
        edge_to_vars.emplace_back();
      }
      return create_result;
    }

#warning "TODO: deal with jumps u---x===y---v where y---v is a misaligned and, due to its length, causes us to not consider the jump"
    //! parse the ScaffoldGraph to construct the JumpGraph
    void construct_jump_graph()
    {
      // define a priority queue that gives the nearest vertex first (recall that STL sorts vectors component wise)
      typedef std::pair<UncertainLength, ScafVertex> WeightedVertex;
      typedef std::priority_queue<WeightedVertex, std::vector<WeightedVertex>, std::greater<WeightedVertex> > DistQueue;

      const ScafVertex& u = jumped_edge.first;
      const ScafVertex& v = jumped_edge.second;
      const JumpVertex& u_prime = translated_jumped_edge.first;
      const JumpVertex& v_prime = translated_jumped_edge.second;

      // we only need to translate back to the JumpGraph while constructing it
      // for later uses, a translate map JumpGraph->RawScaffoldGraph suffices
      VTranslateMap<RawScaffoldGraph, JumpGraph> sg_to_jg({{u, u_prime},{v, v_prime}});

      DistQueue Q;
      Q.DEEP_EMPLACE(0, u);
      ScafVertexSet seen;

      while(!Q.empty()){
        // step 1: get the minimum element of the DistQueue
        const WeightedVertex& x_pair = Q.top();
        const UncertainLength x_dist = x_pair.first;
        const ScafVertex x = x_pair.second;
        const JumpVertex x_prime = sg_to_jg.at(x);
        Q.pop();
        seen.insert(x);
        DEBUG5(std::cout << "CJG: next in queue: "<<sg[x]<<std::endl);
        
        for(auto e_it = sg.get_incident_non_matching(x); e_it; ++e_it){
          const ScafEdge& xy = *e_it;
          const ScafVertex& y = sg.target(xy);
          DEBUG5(std::cout << "CJG: checking edge "<<sg.get_edge_name(xy)<<" (len: "<<sg[xy].length<<")"<<std::endl);
          if((x!=u) || (y!=v)){
            // step 2: check if some u-y-z path can be put in the gap
            const UncertainLength xy_length = sg.get_uncertain_length(xy);
            
            if(xy_length.mean != NO_LENGTH) {
              const ScafVertex& z = sg.matched_with(y);
              const ScafEdge yz = sg.find_edge(y, z).first;
              const int yz_length = sg[yz].length;

              if(y != v){
                if(yz_length != NO_LENGTH){
                  UncertainLength z_dist = x_dist + xy_length + yz_length;
                  // the path fits in the gap if and only if fit_in_gap picks at most 0 with high probability
                  DEBUG5(std::cout << "CJG: found path "<<sg[x]<<"-->"<<sg[y]<<"-->"<<sg[z]<<" (dist "<< z_dist<<" = dist "<<x_dist<<" + len "<<xy_length + yz_length<<")"<<std::endl);
                  
                  // even if z_dist is too far away, maybe with a negative-length z-v edge, this can be remedied
                  if(!fits_in_gap(z_dist)){
                    const auto zv_edge = sg.find_edge(z, v);
                    if(zv_edge.second){
                      const UncertainLength zv_length = sg.get_uncertain_length(zv_edge.first);
                      if(zv_length.mean != NO_LENGTH)
                        z_dist += sg.get_uncertain_length(zv_edge.first);
                    }
                  }
                  DEBUG5(std::cout << "CJG: probability of jump of length "<<gap_length<<" over "<<z_dist<<": "<< (z_dist - gap_length).prob_of_picking_at_most(0)*100<<"%"<<std::endl);

                  if(fits_in_gap(z_dist)){
                    JumpVertex y_prime, z_prime;
                    // yes, u-...-x-y-z fits in the gap, so add it to the graph if it's not already there
                    // find y, and z in the JumpGraph, creating them if necessary                
                    const auto tr_bk_it_y = sg_to_jg.find(y);
                    if(tr_bk_it_y == sg_to_jg.cend()){
                      y_prime = add_vertex(y);
                      sg_to_jg.emplace_hint(tr_bk_it_y, y, y_prime);
                    } else y_prime = tr_bk_it_y->second;
                    add_edge(x_prime, y_prime, x, y);

                    // also add y-z if it's not already there
                    const auto tr_bk_it_z = sg_to_jg.find(z);
                    if(tr_bk_it_z == sg_to_jg.cend()){
                      z_prime = add_vertex(z);
                      sg_to_jg.emplace_hint(tr_bk_it_z, z, z_prime);
                    } else z_prime = tr_bk_it_z->second;
                    add_edge(y_prime, z_prime, y, z);

                    // add z to the next layer unless it's been seen before
                    if(!contains(seen, z)) Q.emplace(z_dist, z);
                  }// if u-...-x-y-z fits in the gap
                }// if y-z has a length
              } else {
              // if y = v, then we also add the edge xv, provided it fits in the gap
                DEBUG5(std::cout << "CJG: found the end of the network" << std::endl);
                const UncertainLength y_dist = x_dist + xy_length;
                if(fits_in_gap(y_dist)) add_edge(x_prime, v_prime, x, v);
              }// if xy is xv
            }// if x-y has a length
          }// if we didn't hit u or v
        }// for all out_edges of x
      }// while Q not empty
    }


    //! initialize the variables
    void initialize_variables(const IloEnv& env, const unsigned jump_edge_multiplicity)
    {
      DEBUG5(std::cout << "initializing variables in "<<jump_edge_multiplicity<<" layers..." <<std::endl);
      const EdgeName jumped_edge_name = sg.get_edge_name(jumped_edge);
      const std::string plain_name = "jmp_" + jumped_edge_name.first + "," + jumped_edge_name.second + "_";
      for(auto& evec: edge_to_vars){
        evec.reserve(jump_edge_multiplicity);
        for(unsigned i = 0; i < jump_edge_multiplicity; ++i){
          const std::string name = plain_name + std::to_string(i);
          evec.emplace_back(env, 0, 1, IloNumVar::Bool, name.c_str());
        }
      }
    }

    //! parse the JumpGraph to compute the constraints on the variables, integrating main_jump_var
    /** assumes that the variables have been initialized */
    void initialize_constraints(const IloEnv& env, IloModel& model, const IloNumVar& main_jump_var, const EdgeVarMap& times_used)
    {
      DEBUG5(std::cout << "initializing constraints..." <<std::endl);
      assert(!edge_to_vars.empty());
      const unsigned jump_edge_multiplicity = edge_to_vars.cbegin()->size();
      IloRangeArray c(env);
      // step 1: assure that the sum over all representations of an edge uv in the ScaffoldGraph does not exceed the registered jump-multiplicity
      DEBUG5(std::cout << "(a) all "<<jump_edge_multiplicity<<" representations of each edge must be represented by y ('times_used')..." <<std::endl);
      for(auto e_range = boost::edges(jump_graph); e_range.first != e_range.second; ++e_range.first){
        const JumpEdge& e = *e_range.first;
        const unsigned e_index = jump_graph[e].index;
        const JumpVertex& u = boost::source(e, jump_graph);
        const JumpVertex& v = boost::target(e, jump_graph);
        const ScafVertexPair uv_orig(translate.at(u), translate.at(v));
        
        IloExpr expr(env);
        expr += times_used.at(uv_orig);
        for(const auto& e_var: edge_to_vars[e_index]) expr -= e_var;
        c.add(expr >= 0);
      }

      // step 2: flow conservation for each level:
      // each vertex conserves the flow going out via its matching edge by incoming flow via the non-matching edges and
      // each vertex conserves the flow coming in via its matching edge by outgoing flow via the non-matching edges
      // special rules apply to translated_jumped_edge
      boost::unordered_map<JumpVertex, std::vector<IloExpr> > conserve_outgoing_matching;
      boost::unordered_map<JumpVertex, std::vector<IloExpr> > conserve_incoming_matching;
      // initialize conservation expressions: #multiplicity expressions for each vertex
      DEBUG5(std::cout << "(b) flow conservation..." <<std::endl);
      // first, create all expressions
      for(auto u_range = boost::vertices(jump_graph); u_range.first != u_range.second; ++u_range.first){
        const JumpVertex& u = *u_range.first;
        auto& com_u_exvec = conserve_outgoing_matching.DEFAULT_EMPLACE(u).first->second;
        auto& cim_u_exvec = conserve_incoming_matching.DEFAULT_EMPLACE(u).first->second;
        com_u_exvec.reserve(jump_edge_multiplicity);
        cim_u_exvec.reserve(jump_edge_multiplicity);
        for(unsigned i = 0; i < jump_edge_multiplicity; ++i){
          com_u_exvec.emplace_back(env);
          cim_u_exvec.emplace_back(env);
        }
      }
      // then, for each edge, add it to the outgoing expression of its source and the incoming expression of its target
      for(auto e_range = boost::edges(jump_graph); e_range.first != e_range.second; ++e_range.first){
        const JumpEdge& e = *e_range.first;
        const unsigned e_index = jump_graph[e].index;
        const JumpVertex& u = boost::source(e, jump_graph);
        const JumpVertex& v = boost::target(e, jump_graph);
        const bool is_matching = (sg.matched_with(translate.at(u)) == translate.at(v));

        for(unsigned level = 0; level < jump_edge_multiplicity; ++level) {
          const IloNumVar& e_at_level = edge_to_vars.at(e_index).at(level);
          if(is_matching){
            conserve_outgoing_matching[u][level] -= e_at_level;
            conserve_incoming_matching[v][level] -= e_at_level;
          } else {
            conserve_outgoing_matching[v][level] += e_at_level;
            conserve_incoming_matching[u][level] += e_at_level;
          }
        }
      }
      // add the expressions to the model
      DEBUG5(std::cout << "adding expressions to the model..." <<std::endl);
      for(auto u_range = boost::vertices(jump_graph); u_range.first != u_range.second; ++u_range.first){
        const JumpVertex& u = *u_range.first;
        // do not add constraints for the translated_jumped_edge vertices to the model
        if((u != translated_jumped_edge.first) && (u != translated_jumped_edge.second)){
          for(const auto& exp: conserve_outgoing_matching.at(u)) c.add(exp == 0);
          for(const auto& exp: conserve_incoming_matching.at(u)) c.add(exp == 0);
        }
      }

      // step 3: tie in the main_jump_var
      //sum over all outgoing from u minus incoming to u
      const JumpVertex& x = translated_jumped_edge.first;
      IloExpr main_bound(env);
      main_bound -= main_jump_var;
      for(auto e_range = boost::out_edges(x, jump_graph); e_range.first != e_range.second; ++e_range.first){
        const JumpEdge& e = *e_range.first;
        const unsigned e_index = jump_graph[e].index;
        const auto& e_vars = edge_to_vars.at(e_index);
        for(unsigned i = 0; i < jump_edge_multiplicity; ++i) main_bound += e_vars.at(i);
      }
      for(auto e_range = boost::in_edges(x, jump_graph); e_range.first != e_range.second; ++e_range.first){
        const JumpEdge& e = *e_range.first;
        const unsigned e_index = jump_graph[e].index;
        const auto& e_vars = edge_to_vars.at(e_index);
        for(unsigned i = 0; i < jump_edge_multiplicity; ++i) main_bound -= e_vars.at(i);
      }
      c.add(main_bound >= 0);

      model.add(c);
    }

  public:

    //! construct a jump network
    /** note that the vertices u & v are inserted into the jumpgraph by the initializer */
    ILPContigJumpNetwork(const ScaffoldGraph& _sg,
                         const ScafVertexPair& _jumped_edge,
                         const float threshold_probability,
                         const UncertainLength& _length):
      sg(_sg),
      jump_graph(),
      jumped_edge(_jumped_edge),
      translated_jumped_edge(add_vertex(jumped_edge.first), add_vertex(jumped_edge.second)),
      gap_length(_length),
      jump_threshold_probability(threshold_probability)
    {
      assert(gap_length.mean != NO_LENGTH);
      DEBUG4(std::cout << "constructing JumpGraph of "<<sg.get_edge_name(jumped_edge)<<" with gap "<<gap_length<<std::endl);
      // make sure the edge is not a matching edge
      assert(!sg.is_matching_edge(jumped_edge));
      construct_jump_graph();
      DEBUG4(std::cout << "constructed jump graph with "<<boost::num_edges(jump_graph)<<" edges"<<std::endl);
    }

    //! initialize the Ilo variables & constraints and use main_jump_var to hold the jump-multiplicity of jumped_edge
    void integrate_into_model(const IloEnv& env,
                              IloModel& model,
                              const unsigned jump_edge_multiplicity,
                              const IloNumVar& main_jump_var,
                              const EdgeVarMap& times_used)
    {
      if(boost::num_edges(jump_graph)){
        DEBUG5(std::cout << "integrating JumpNetwork (with "<<boost::num_edges(jump_graph)<<" edges) of "<<sg.get_edge_name(jumped_edge)<<" (len "<<sg[jumped_edge].length<<") with "<<jump_edge_multiplicity<<" layers"<<std::endl);
        DEBUG5(for(auto r = boost::edges(jump_graph); r.first != r.second; ++r.first){
            const ScafVertex& u_orig = translate.at(boost::source(*r.first, jump_graph));
            const ScafVertex& v_orig = translate.at(boost::target(*r.first, jump_graph));
            const ScafEdge e = sg.find_edge(u_orig, v_orig).first;
            std::cout << sg.get_edge_name(e)<<" (len "<<sg[e].length << ")" << std::endl;
          }
        );
        // step 1: initialize variables
        initialize_variables(env, jump_edge_multiplicity);
        // step 2: initialize constraints
        initialize_constraints(env, model, main_jump_var, times_used);
      } else DEBUG5(std::cout << "not integrating empty JumpNetwork of "<<sg.get_edge_name(jumped_edge)<<std::endl);
    }

    //! return whether the network contains no edges
    bool empty() const{
      return num_edges() == 0;
    }
    //! return the number of edges in the jump network
    unsigned num_edges() const{
      return boost::num_edges(jump_graph);
    }

  };

}// namespace

