
#pragma once

#include "utils/utils.hpp"
#include "utils/string_utils.hpp"

enum Op : byte {NOOP, DEL_LEFT, DEL_RIGHT, CHANGE, FAIL};
typedef std::vector2d<Op> Levenstein_Op_Table;

//! modified Levenstein distance of an alignment of two sequences
/** this Levenstein distance is modified to:
 * 1. absorb any suffix of s1 or s2 at no cost, as it will be used to find alignments of sub-sequences and
 *    we don't want to punish the fact that one of the sequences may be shorter.
 * 2. ignore transformations that cost at least "max_cost"
 * Note that the seeming advantage of short sequences is countered by using the AVERAGE distance for scoring.
 * NOTE: this is a modification of https://en.wikibooks.org/wiki/Algorithm_Implementation/Strings/Levenshtein_distance#C.2B.2B
 */
std::vector<unsigned> modified_levenstein_distance(const std::string& s1,
                                                   const std::string& s2,
                                                   const size_t lower_index,
                                                   const size_t upper_index,
                                                   Levenstein_Op_Table* op_table = NULL)
{
  assert(upper_index >= lower_index);
  assert(upper_index < s1.length());
  const ssize_t len1 = s1.length() - lower_index;
  const ssize_t len2 = s2.length();
  const ssize_t window = upper_index - lower_index;
  DEBUG5(std::cout << "computing shifted Levenstein distances for strings of length "<<_len1<<" and "<<_len2<<" - op table @"<<op_table<<std::endl);
  std::cout << "constructing table of size ("<<len1+1<<" x "<<len2+1<<")"<<std::endl;

  // d[x,y] = the cost for the substrings s1[x:len1] and s2[y:len2]
  std::vector2d<unsigned short> d(len1 + 1, len2 + 1);
  if(op_table) op_table->resize(len1 + 1, len2 + 1, NOOP);

  // the right and lower borders of the table are initialized depending on whether they are be reached while respecting max_cost
  d[{len1, len2}] = 0;

  // step1: init; allow eating any suffix of s1 OR s2 for free unless the max_cost is reached
  for(unsigned i = 0; i < len1; ++i) d[{i, len2}] = 0;
  for(unsigned j = 0; j < len2; ++j) d[{len1, j}] = 0;

  // step2: compute inner cells
  for(ssize_t i = len1 - 1; i >= 0; --i){
    for(ssize_t j = len2 - 1; j >= 0; --j){
      const unsigned del1_cost = (unsigned)d[{i+1, j}] + 1;
      const unsigned del2_cost = (unsigned)d[{i, j+1}] + 1;
      const unsigned change_cost = (unsigned)d[{i+1, j+1}] + (s1[i + lower_index] == s2[j] ? 0 : 1);
      if(del1_cost < del2_cost){
        if(change_cost < del1_cost){
          d[{i, j}] = change_cost;
          if(op_table) (*op_table)[{i, j}] = CHANGE;
        } else {
          d[{i, j}] = del1_cost;
          if(op_table) (*op_table)[{i, j}] = DEL_LEFT;
        }
      } else {
        if(change_cost < del2_cost){
          d[{i, j}] = change_cost;
          if(op_table) (*op_table)[{i, j}] = CHANGE;
        } else {
          d[{i, j}] = del2_cost;
          if(op_table) (*op_table)[{i, j}] = DEL_RIGHT;
        }
      }
    }
  }

  std::vector<unsigned> result(window);
  for(unsigned i = 0; i < window; ++i) result[i] = d[{i, 0}];

  return result;
}
//! a simple char consensus; This can be used if some transitions are more likely than others
char char_consensus(const char& x, const char& y)
{
  return x;
}

//! retrace an operations table for a Levenstein-distance computation, producting the implied consensus
std::string levenstein_consensus(const std::string& s1,
                                 const std::string& s2,
                                 const unsigned lower_index,
                                 const unsigned s2_offset_from_lower_index,
                                 const Levenstein_Op_Table& op_table)
{
  const unsigned s2_offset = lower_index + s2_offset_from_lower_index;
  DEBUG4(std::cout << "lower_index = "<<lower_index<<" s2_offset_from_lower_index = "<<s2_offset_from_lower_index<<" s2_offset = "<<s2_offset<<std::endl);
  std::string result(s1.substr(0, s2_offset));
  DEBUG4(std::cout << "starting with "<<result<<" (len: "<<result.length()<<")"<<std::endl);
  const size_t len1 = op_table.size().first - 1;
#ifndef NDEBUG
  const size_t len2 = op_table.size().second - 1;
#endif
  DEBUG4(std::cout << "op table size is ("<<len1<<", "<<len2<<")"<<std::endl);
  size_t i = s2_offset_from_lower_index, j = 0;
  Op last_op;

  do{
    last_op = op_table[{i, j}];
    switch(last_op){
      case DEL_LEFT:
        ++i;
        break;
      case DEL_RIGHT:
        ++j;
        break;
      case CHANGE:
        result += char_consensus(s1[lower_index + i], s2[j]);
        ++i;
        ++j;
        break;
      default:
        break;
    }
  } while(last_op != NOOP);
  
  DEBUG4(std::cout << "found NOOP at ("<<i<<", "<<j<<"); " << "after merge: result is "<<result<<" (len: "<<result.length()<<")"<<std::endl);
  assert(((i == len1) && (j <= len2)) || ((j == len2) && (i <= len1)));

  if(i == len1) result += s2.substr(j);
  // if result is still completely inside of s1, then add the rest of s1 to result
  if(lower_index + result.length() < s1.length())
    result += s1.substr(lower_index + result.length());
  
  DEBUG4(std::cout << "final result: "<<result<<" (len: "<<result.length()<<")"<<std::endl);
  return result;
}


