

#ifndef MIN_MATRIX_HPP
#define MIN_MATRIX_HPP

#include <memory>

#include "utils/utils.hpp"
#include "utils/graph_typedefs.hpp"

// set if the solution should be handled by the DP algo

#define HANDLE_SOLUTION(x)  x

namespace scaffold{
  typedef std::list<EdgeName> EdgeNameList;
  struct global_list;

  typedef std::shared_ptr<global_list> SharedGList;
  typedef std::list<SharedGList> ListGList;
  typedef std::shared_ptr<ListGList> SharedListGList;

#define NO_EDGE ((EdgeName){NONAME, NONAME})

  // a global list of edges to be shared by all the DP-tables
  struct global_list{
    EdgeName e;

    SharedGList left, right;

    global_list(const EdgeName& _e, const SharedGList& _left = nullptr): e(_e), left(_left), right(nullptr) {};
    global_list(const SharedGList& _left, const SharedGList _right): e(NO_EDGE), left(_left), right(_right) {};
  };

  // an entry for a field of (path,cycle)
  class MatrixEntry
  {
  public:
    size_t weight;
  protected:
    SharedGList edges;
  public:
    // copy constructor
    MatrixEntry(const MatrixEntry& E): weight(E.weight), edges(E.edges) {}
    // weight == SIZE_MAX will represent an undefined entry
    MatrixEntry(const size_t _weight = SIZE_MAX): weight(_weight), edges(nullptr) {}
    MatrixEntry(const EdgeName& e, const size_t _weight = SIZE_MAX): weight(_weight), edges(std::make_shared<global_list>(e)) {}
    // handling undefinedness
    bool is_undefined() const { return weight == SIZE_MAX; }
    void set_undefined() { weight = SIZE_MAX; }

    bool has_no_edges(){
      return edges == nullptr;
    }

    void clear(){
      weight = 0;
      edges.reset();
    }
        
    void append_edge(const EdgeName& e, const size_t _weight = 0)
    {
      weight += _weight;
      HANDLE_SOLUTION(
        if(edges != nullptr) {
          // if we already have some edges, add e as the new head of the global list
          edges = std::make_shared<global_list>(e, edges);
        } else {
          // if e is our first edge, just initialize our edges
          edges = std::make_shared<global_list>(e);
        }
      )
    }

    void join(const MatrixEntry& ME)
    {
      weight += ME.weight;
      HANDLE_SOLUTION(
        if(edges != nullptr){
          // if we already have some edges, create a new splitting node pointing to both our and ME's edges
          edges = std::make_shared<global_list>(edges, ME.edges);
        } else {
          // if we don't have any edges yet, then point to ME's edges
          edges = ME.edges;
        }
      )
    }

    EdgeNameList get_edges() const
    {
      if(edges != nullptr){
        EdgeNameList result;
        std::stack<SharedGList> el_stack;
        el_stack.push(edges);
        while(!el_stack.empty()){
          SharedGList top_el = el_stack.top(); el_stack.pop();
          const EdgeName& e = top_el->e;
          if(e != NO_EDGE) result.push_back(e);
          if(top_el->left != nullptr) el_stack.push(top_el->left);
          if(top_el->right != nullptr) el_stack.push(top_el->right);
        }
        return result;
      } else return EdgeNameList();
    }
   

    // assign the MatrixEntry
    void operator=(const MatrixEntry& ME)
    {
      weight = ME.weight;
      HANDLE_SOLUTION(edges = ME.edges;)
    }

    // add weight and edges of another MatrixEntry
    void operator+=(const MatrixEntry& ME)
    {
      join(ME);
    }
    MatrixEntry operator+(const MatrixEntry& ME) const
    {
      MatrixEntry result(*this);
      result += ME;
      return result;
    }

  };

  // update an entry if it is worse than the given one & return whether an update took place
  template<typename Index, typename Matrix, class Compare = std::less<size_t> >
  bool update_if_worse(Matrix& m, const Index& index, MatrixEntry* ME)
  {
    const Compare is_better;
    // get the entry at index, creating it as a copy of E if necessary
    const typename Matrix::iterator ME_i = m.find(index);
    if(ME_i == m.end()){
      m.emplace(index, ME);
      return true;
    } else {
      // if it existed before and its weight is worse than E's, then update it
      if(is_better(ME->weight, ME_i->second->weight)) {
        delete ME_i->second;
        ME_i->second = ME;
        return true;
      } else return false;
    }
  }
/*
  // update an entry if it is worse than the given one & return whether an update took place
  template<typename Index, typename Matrix, class Compare = std::less<size_t> >
  bool update_if_worse(Matrix& m, const Index& index, const SharedListGList& edges, const size_t weight)
  {
    MatrixEntry* ME = new MatrixEntry(edges, weight);
    if(!update_if_worse<Compare>(m, index, ME)){
      delete ME;
      return false;
    } else return true;
  }
  template<typename Index, typename Matrix, class Compare = std::less<size_t> >
  bool update_if_worse(Matrix& m, const Index& index, const MatrixEntry& E)
  {
    return update_if_worse<Index, Matrix, Compare>(m, index, E.edges, E.weight);
  }
  */
  // update an entry if it is worse than the sum of the given two & return whether an update took place
  template<typename Index, typename Matrix, class Compare = std::less<size_t> >
  bool update_if_worse(Matrix& m, const Index& index, const MatrixEntry& E, const MatrixEntry& cE)
  {
    const Compare is_better;
    // get the entry at index, creating it as a copy of E if necessary
    const typename Matrix::iterator ME_i = m.find(index);
    if(ME_i == m.end()){
      MatrixEntry* ME = new MatrixEntry(E);
      *ME += cE;
      m.emplace(index, ME);
      return true;
    } else {
      MatrixEntry& ME = *(ME_i->second);
      if(is_better(E.weight + cE.weight, ME.weight)){
        // if it existed before and its weight is worse than E's, then update it
        ME = E;
        ME += cE;
        return true;
      } else return false;
    }// if
  }// function



  // a matrix of paths & cycles that keeps a minimum & refuses insertion of pareto-worse elements
  template<class Compare = std::less<size_t> >
  class PCMinMatrix
  {
  public:
    typedef unordered_map<PathsAndCycles, MatrixEntry*> entries_t;

    // the bounds
    const PathsAndCycles max;
  protected:
    entries_t* entries;
    const Compare is_better;

  public:
    PathsAndCycles best_index;

    PCMinMatrix(const unsigned paths, const unsigned cycles): max(paths, cycles), entries(new entries_t()), is_better() {}
    PCMinMatrix(const PathsAndCycles& _max): max(_max), entries(new entries_t()), is_better() {}
    PCMinMatrix(const PCMinMatrix& m): 
      max(m.max),
      entries(new entries_t()),
      is_better(),
      best_index(m.best_index) 
    {
      if(!m.entries->empty()){
        const MatrixEntry* const best_ME = m.entries->at(best_index);
        size_t best_weight = best_ME->weight;
        entries->emplace(best_index, new MatrixEntry(*best_ME));
        for(auto& ME: *m.entries)
          if( !(best_index <= ME.first) || is_better(ME.second->weight, best_weight))
            entries->emplace(ME.first, new MatrixEntry(*ME.second));
      }
    }
    ~PCMinMatrix()
    {
      for(auto& ME: *entries) delete ME.second;
      delete entries; 
    }

    // writing to the storage, return whether something was written
    // NOTE: if an entry is overwritten, its memory is freed
    bool put(const PathsAndCycles& index, MatrixEntry* ME)
    {
      if(index > max) return false;

      const bool was_empty = entries->empty();
      const size_t best_weight = was_empty ? SIZE_MAX : entries->at(best_index)->weight;
      const size_t weight = ME->weight;
      // first, check if the best_index is better than what we are about to enter
      if(was_empty || !(best_index <= index) || is_better(weight, best_weight)){
        DEBUG5(std::cout << "trying to put entry at "<<index<<"..."<<std::flush);
        if(!update_if_worse<PathsAndCycles, entries_t, Compare>(*entries, index, ME)) return false;
        DEBUG5(std::cout << "success, now updating best_index ("<<best_index.p<<","<<best_index.c<<") if "<<was_empty<<" || Compare("<<weight<<","<<best_weight<<")"<<std::endl);
        // finally, update best_index, throwing out the item at best_index if necessary
        if(index != best_index){
          if(!was_empty && (index < best_index) && is_better(weight, best_weight)){
            const typename entries_t::iterator min_i = entries->find(best_index);
            assert(min_i != entries->end());
            delete min_i->second;
            entries->erase(min_i);
          }
          if(was_empty || is_better(weight, best_weight)) best_index = index;
        }
        DEBUG5(std::cout << "new best_index is ("<<best_index.p<<","<<best_index.c<<")"<<std::flush<<" with weight "<<entries->at(best_index)->weight<<std::endl);
        return true;
      } else return false;
    }
    
    bool put_empty(const PathsAndCycles& index) 
    {
      MatrixEntry* ME = new MatrixEntry(0);
      if(!put(index, ME)){
        delete ME;
        return false;
      } else return true;
    }
    // put an entry containing a single edge
    bool put(const PathsAndCycles& index, const EdgeName& e, const size_t weight)
    {
      MatrixEntry* ME = new MatrixEntry(e, weight);
      if(!put(index, ME)){
        delete ME;
        return false;
      } else return true;
    }
    // put an entry containing all edges of g
    template<class Graph>
    bool put_all(const PathsAndCycles& index, const Graph& g)
    {
      MatrixEntry* ME = new MatrixEntry(0);
      for(EdgeIterRange<Graph> er = boost::edges(g); er.first != er.second; ++er.first){
        const Edge<Graph>& e = *er.first;
        const Vertex<Graph>& u = boost::source(e, g);
        const Vertex<Graph>& v = boost::target(e, g);
        ME->append_edge(EdgeName(g[u].name, g[v].name), g[e].weight);
      }
      if(!put(index, ME)) {
        delete ME;
        return false;
      } else return true;
    }
    /*
    bool put(const unsigned paths, const unsigned cycles, const SharedEdgeList& edges, const size_t weight) 
    {
      return put(PathsAndCycles(paths, cycles), edges, weight);
    }
    bool put(const PathsAndCycles& index, const MatrixEntry& E) 
    {
      MatrixEntry* ME = new MatrixEntry(E);
      if(!put(index, ME)){
        delete ME;
        return false;
      } else return true;
    }
    bool put(const unsigned paths, const unsigned cycles, const MatrixEntry& E) 
    {
      return put(PathsAndCycles(paths, cycles), E);
    }
    */
    // put an entry at index with the sum of the given entries
    bool put(const PathsAndCycles& index, const MatrixEntry& E, const MatrixEntry& cE)
    {
      assert(index <= max);

      const size_t sum_weight = E.weight + cE.weight;
      const bool was_empty = entries->empty();
      const size_t best_weight = was_empty ? SIZE_MAX : entries->at(best_index)->weight;
      // first, check if the best_index is better than what we are about to enter
      if(was_empty || !(best_index <= index) || is_better(sum_weight, best_weight)){
        if(!update_if_worse<PathsAndCycles, entries_t, Compare>(*entries, index, E, cE)) return false;
        // finally, update best_index
        if(index != best_index){
          if(!was_empty && (index <= best_index) && is_better(sum_weight, best_weight)) entries->erase(best_index);
          if(was_empty || is_better(sum_weight, best_weight)) best_index = index;
        }
        return true;
      } else return false;
    }// function

    // add an edge of certain weight to all entries
    void add_to_all(const EdgeName& e, const size_t weight)
    {
      for(auto& ME : *entries) ME.second->append_edge(e, weight);
    }

    // reading from the storage
    const MatrixEntry& get(const PathsAndCycles& index) const 
    {
      assert(index <= max);
      return *(entries->at(index));
    }
    const MatrixEntry& get(const unsigned paths, const unsigned cycles) const 
    {
      return get({paths, cycles});
    }
    const MatrixEntry& at(const PathsAndCycles& index) const 
    {
      return get(index);
    }
    const MatrixEntry& at(const unsigned paths, const unsigned cycles) const 
    {
      return get(paths, cycles);
    }
    std::pair<entries_t::const_iterator, entries_t::const_iterator> get_entries() const
    {
      return std::pair<entries_t::const_iterator, entries_t::const_iterator>(entries->begin(), entries->end());
    }
    
    // merge a Matrix pc into our own, destroying pc in the process
    void destructive_merge(PCMinMatrix<Compare>& pc)
    {
      DEBUG6(std::cout << "destructively merging "<<pc.entries->size()<<" entries into our "<<size()<<" entries"<<std::endl);
      for(auto ME = pc.entries->begin(); ME != pc.entries->end(); ME = pc.entries->erase(ME)){
        // try to put this entry into our own matrix
        if(!put(ME->first, ME->second)){
          // if we didn't put the entry into our table, we'll free its memory & remove it from the map
          delete ME->second;
        }// if
      }// for
    }// function


    // modify a key by adding a number of paths & cycles
    bool add_to_key(const PathsAndCycles& key, const PathsAndCycles& pc){
      if(pc.p + pc.c > 0){
        const entries_t::iterator i = entries->find(key);
        assert(i != entries->end());
        const MatrixEntry* ME = i->second;
        entries->erase(i);
        return put(key + pc, ME);
      }
    }
    // modify each key, watch the pointer-juggling!
    void add_to_each_key(const PathsAndCycles& pc){
      if(pc.p + pc.c > 0){
        // update max: doesn't change, min will be updated automatically
        // update entries (write completely anew)
        entries_t* old_entries = entries;
        entries = new entries_t;
        for(auto& ME: *old_entries) put(ME.first + pc, ME.second);
        // delete old entries
        delete old_entries;
      }
    }

    // find an entry E of me and an entry SE of S such that their indices sum up to (p,c) and their weight is min among all such entries
    PathsAndCycles find_min_entry_combination_by_iteration(const PathsAndCycles& pc, const PCMinMatrix& m){
      PathsAndCycles min_pc;
      size_t best_weight = SIZE_MAX;

      for(auto E : *entries){
        const PathsAndCycles Epc = E.first;
        const entries_t::const_iterator mE = m.entries->find(pc - Epc);
        if(mE != m.entries->end()){
          if(is_better(E.second->weight + mE->second->weight, best_weight)){
            best_weight = E.second->weight + mE->second->weight;
            min_pc = Epc;
          }// if
        }// if
      }// for
      return min_pc;
    }// function

    // find the minimum entry combination by testing all pairs (p',c') <= (p,c)
    PathsAndCycles find_min_entry_combination_by_testing(const PathsAndCycles& pc, const PCMinMatrix& m){
      PathsAndCycles min_pc;
      size_t best_weight = SIZE_MAX;

      for(unsigned mp = 0; (mp <= pc.p) && (mp <= m.max.p); ++mp)
        for(unsigned mc = 0; (mc <= pc.c) && (mc <= m.max.c); ++mc){
          PathsAndCycles current(mp, mc);
          // check if both entries exist
          const entries_t::const_iterator mE = m.entries->find(current);
          PathsAndCycles diff(pc - current);
          if(mE != m.entries->end()){
            const entries_t::const_iterator E = entries->find(diff);
            if(E != entries->end())
              if(is_better(mE->second->weight + E->second->weight, best_weight)){
                best_weight = mE->second->weight + E->second->weight;
                min_pc = diff;
              } // if
          } // if
        } // for
      return min_pc;
    }// function

    // get a pair of references to solution entries such that their paths add up to p and their cycles add up to c
    PathsAndCycles find_min_entry_combination(const PathsAndCycles& pc, const PCMinMatrix& m){
      // if the number of entries is not too large yet, we iterate through the entries instead of testing all (p',c') <= (p,c)
      if(entries->size() < std::min(pc.p * pc.c, m.max.p * m.max.c))
        return find_min_entry_combination_by_iteration(pc, m);
      else
        return find_min_entry_combination_by_testing(pc, m);
    }// function

    // output a minimum-weight MatrixEntry within the solution boundaries
    const MatrixEntry& get_min_entry() const
    {
      return *(entries->at(best_index));
    }

    const size_t size() const
    {
      return entries->size();
    }

    // return the largest pc for which pc <= i holds for all indices (note that pc is not necessarily an existing index)
    const PathsAndCycles smallest_index() const
    {
      if(entries->empty()) return PathsAndCycles(0,0);
      PathsAndCycles result(UINT_MAX,UINT_MAX);
      for(auto& i : *entries){
        if(i.first.p < result.p) result.p = i.first.p;
        if(i.first.c < result.c) result.c = i.first.c;
      }
      return result;
    }

    // << operator for solutions outputting a table of weights
    friend std::ostream& operator<<(std::ostream& os, const PCMinMatrix& m){
      // write header
      os << std::endl<< "   ";
      for(unsigned p = 0; p <= m.max.p; ++p) os <<"\t" << p;
      os << "\t\tbest index = "<<m.best_index<<std::endl;
      // write table
      for(unsigned c = 0; c <= m.max.c; ++c){
        if(c < 10) os << " ";
        os << c<<":\t";
        for(unsigned p = 0; p <= m.max.p; ++p){
          try{ os << m.at(p, c).weight << "\t"; } catch(...) { os << "\t"; };
        }
        os << std::endl;
      }// for
      return os;
    } // function

  };


}// namespace

#endif

