

/** \file permissive_matching.hpp
 * implementation of a matching on the numbers 1..n (with n < 256) that allows reflexive pairs,
 * trying to be as space-efficient as possible.
 * Currently, we use an array of (compile-time) fixed size such that matched[u] = v for each pairing uv
 *
 * TODO: since this basically saves each edge twice, I suspect that this can be done more space-efficiently
 */

#ifndef PERMISSIVE_MATCHING_HPP
#define PERMISSIVE_MATCHING_HPP

#include <boost/unordered_map.hpp>

#include "utils/utils.hpp"

#define ARRAY_LENGTH MAX_TW

namespace scaffold{

  class PermissiveMatching;
  typedef PermissiveMatching BytePermutation;

  class PermissiveMatching{
  private:
    byte* matched;
  public:

    // ======== Constructors & Destructor ================
    //! construct from an unordered_map
    /** the given map does not have to be a permutation on 1..n; if it is not,
     * then all indices that are not assigned by edges are set to identity
     */
    PermissiveMatching(const unordered_map<byte, byte>* edges = NULL): matched(new byte[ARRAY_LENGTH])
    {
      if(edges){
        for(byte i = 0; i < ARRAY_LENGTH; ++i){
          const auto edge_iter = edges->find(i);
          if(edge_iter != edges->end()){
            const byte edge_target = edge_iter->second;
            matched[i] = edge_target;
            assert((edge_target > i) || (matched[edge_target] == i));
          } else matched[i] = i;
        }// for
      } else for(byte i = 0; i < ARRAY_LENGTH; ++i) matched[i] = i;
    }// constructor
    
    //! copy constructor
    PermissiveMatching(const PermissiveMatching& P): matched(new byte[ARRAY_LENGTH])
    {
      memcpy(matched, P.matched, ARRAY_LENGTH);
    }

    //! destructor
    ~PermissiveMatching()
    {
      delete[] matched;
    }// destructor


    // =========== methods ==================

    //! conversion to size_t, for hash calculation
    operator size_t() const
    {
      size_t result = 0;
      for(byte i = 0; i < ARRAY_LENGTH; ++i)
        boost::hash_combine(result, matched[i]);
      return result;
    }// function


    //! access to the array
    const byte& operator[](const byte i) const
    {
      return matched[i];
    }

    //! access to the array
    byte& operator[](const byte i)
    {
      return matched[i];
    }

    //! introduce an edge uv, unless uv is already present OR the insertion would violate the matching constraint; return whether uv was inserted
    bool introduce_edge(const byte u, const byte v)
    {
      assert(u < ARRAY_LENGTH);
      assert(v < ARRAY_LENGTH);

      if((matched[u] != u) || (matched[v] != v)) return false;
      matched[u] = v;
      matched[v] = u;
    }

    //! return whether everyone matches to itself
    bool is_identity() const
    {
      for(byte i = 0; i < ARRAY_LENGTH; ++i)
        if(matched[i] != i) return false;
      return true;
    }

    //! rename the indices according to the given BytePermutation
    /** For example, suppose our matching is
     * (0,1,3,2), meaning there is an edge between 2 & 3. Now, if we relabel using
     * (3,2,1,0), that is, rename 0 to 3, rename 1 to 2, rename 2 to 1, rename 3 to 0), then we get
     * (1,0,2,3), that is, a matching consisting of an edge between 0 and 1 (who were known as 2 & 3 before)
     */
    void permute(const BytePermutation& perm)
    {
      byte* old_matched = matched;
      DEBUG5(std::cout << "permuting "<<*this<<" with permutation "<<perm<<std::endl);
      matched = new byte[ARRAY_LENGTH];
      for(byte i = 0; i < ARRAY_LENGTH; ++i) matched[i] = i;
      for(byte i = 0; i < ARRAY_LENGTH; ++i)
        matched[perm[i]] = perm[old_matched[i]];
      delete[] old_matched;
      DEBUG5(std::cout<<"-->"<<*this<<std::endl);
    }

    //! TODO: move this out of the permissive matching class as it's unique to the DP code & this should be a more general class
    /** merge this matching with a given one and return the number of paths and cycles that are closed below this DP node
     * deg1_in_both should contain all vertices with degree 1 in both me and P -- they are used to count the number of finished paths
     */
    PathsAndCycles merge_with(const PermissiveMatching& P, const std::bitset<ARRAY_LENGTH>& deg1_in_both)
    {
      PathsAndCycles result;
      byte* old_matched = matched;
      matched = new byte[ARRAY_LENGTH];

      for(byte u = 0; u < ARRAY_LENGTH; ++u){
        byte v = u;
        byte w = u;

        if(P[u] != u){
          // propagate v
          do{
            v = P[v];
            v = old_matched[v];
          } while((v != u) && (v != P[v]) && (v != old_matched[v]));
          if(u == v){
            // if we arrived at u again, we walked in a cirlce, so register the circle
            if(u > v) result.c++;
            matched[u] = u;
            continue;
          }
        }
        if(old_matched[u] != u){
          // propagate w
          do{
            w = old_matched[w];
            w = P[w];
          } while((w != old_matched[w]) && (w != P[w]));
          // note that we cannot have arrived at u since otherwise we'd have 'continue'ed earlier
        }
        // note that, at this point, we have exactly one of P & old_matched assigning v to itself, and the same for w
        if(P[u] == u){
          if(old_matched[u] == u){
            if(deg1_in_both[u]) result.p++;
            matched[u] = u;
          } else {
            // we have a u-w path in the merge
            if(deg1_in_both[u]){
              // since P[u] == u we have a u-w path with a dangling path at u
              matched[u] = u;
              // if we also have a dangling path at w, then increase the number of paths
              if(deg1_in_both[w] && (u > w)) result.p++;
            } else {
              if(deg1_in_both[w]){
                matched[u] = u;
              } else {
                matched[u] = w;
              }
            }// if
          }// if
        } else {
          // P[u] != u, so we have a v-u path
          if(old_matched[u] == u){
            if(deg1_in_both[u]){
              // since old_matched[u] == u, we have a v-u path with a dangling path at u
              matched[u] = u;
              // if v also has a dangling path, increase the number of paths
              if(deg1_in_both[v] && (u > v)) result.p++;
            } else {
              // we have a v-u path in the merge, but v could have a dangling path
              if(deg1_in_both[v]){
                matched[u] = u;
              } else {
                matched[u] = v;
              }
            }
          } else {
            // P[u] != u and old_matched[u] != u, implying deg1_in_both[u]
            // so u is an inner vertex on some v-w path
            matched[u] = u;
          }// if
        }// if
      }// for
      delete[] old_matched;

      return result;
    }

    friend std::ostream& operator<<(std::ostream& os, const PermissiveMatching& P)
    {
      os << "(";
      for(byte i = 0; i < ARRAY_LENGTH; ++i) os << (unsigned)P.matched[i] << " ";
      return os << ")";
    }

  };



  size_t hash_value(const PermissiveMatching& P)
  {
    return (size_t)P;
  }
};


#endif

