

#ifndef TD_SUBGRAPH_HPP
#define TD_SUBGRAPH_HPP

#include "utils/filters.hpp"

namespace scaffold{
  using namespace filters;

  // a subgraph of edges introduced in a subtree of the Tree Decomposition
  template<class Graph>
  class TD_Subgraph
  {
    public:
      typedef contained_filter<VertexSet<Graph> > V_Filter;
      typedef contained_filter<EdgeSet<Graph> > E_Filter;
      typedef filtered_graph<Graph, E_Filter, V_Filter> Subgraph;
    protected:
      const Graph& super_g;
      VertexSet<Graph> introduced_vertices;
      VertexSet<Graph> forgotten_vertices;
      EdgeSet<Graph> introduced_edges;
      V_Filter vertex_filter;
      E_Filter edge_filter;
      Subgraph g;
    public:
      // constructor from a given host graph
      TD_Subgraph(const Graph& _super_g):
        super_g(_super_g),
        introduced_vertices(),
        forgotten_vertices(),
        introduced_edges(default_buckets, _super_g),
        vertex_filter(&introduced_vertices),
        edge_filter(&introduced_edges),
        g(Subgraph(super_g, edge_filter, vertex_filter))
      {}
      // destructor
      ~TD_Subgraph() {}

      // introduce a vertex
      void introduce_vertex(const Vertex<Graph>& v)
      {
        introduced_vertices.emplace(v);
      }
      // forget a vertex; note: forget edges whose both vertices have been forgotten!
      void forget_vertex(const Vertex<Subgraph>& u)
      {
        // forget edges incident to v and a forgotten vertex
        for(OEdgeIterRange<Subgraph> r = out_edges(u, g); r.first != r.second;){
          const Vertex<Subgraph>& v = target(*r.first, g);
          // see if v is forgotten
          const auto i = forgotten_vertices.find(v);
          if(i != forgotten_vertices.end()){
            // if v is forgotten, then forget the edge uv, but advance r.first before
            const OEdgeIter<Subgraph> e = r.first++;
            introduced_edges.erase(*e);
            // if the deletion of e isolated v in g, then delete v
            if(degree(v, g) == 0) {
              forgotten_vertices.erase(i); 
              introduced_vertices.erase(v);
            }// if
          } else ++r.first;
        }// for
        // if all edges of u are forgotten, delete u as well, otherwise mark it forgotten
        if(degree(u, g) == 0) introduced_vertices.erase(u);
          else forgotten_vertices.emplace(u);
      }

      // introduce an edge
      void introduce_edge(const Edge<Graph>& e)
      {
        introduced_edges.emplace(e);
      }
      
      void join(const TD_Subgraph<Graph>& h)
      {
        for(const Vertex<Graph>& v : h.introduced_vertices)
          if(!contains(introduced_vertices, v)) introduce_vertex(v);

        for(const Vertex<Graph>& v : h.forgotten_vertices)
          forgotten_vertices.emplace(v);

        for(const Edge<Graph>& e : h.introduced_edges){
          assert(!contains(introduced_edges, e));
          introduce_edge(e);
        }
      }
  };

}//namespace

#endif

